MODULE abundances
  use dimensions,only:nz,ndim
  implicit none
  ! abundances
  double precision,dimension(nz,ndim)::xa!,na
  double precision,dimension(nz)::neo,nio,nzz,avm,nbar,ye
end MODULE abundances

MODULE chemistry
  ! CHEMICAL PARAMETERS
  ! initialise mass fractions, molecular weights, and reaction rates
  use constants,only:amu
  use dimensions,only:species,nreacmax,ndim
  implicit none

  ! modular parameters
  character*3::QVALUES
  character*3::DEDUCE

  ! initial mass fractions
  double precision,dimension(ndim)::frac_init

  ! small values for the mass fractions
  double precision::threshold,tiny,threflux

  ! species identification
  ! molecular weights
  double precision,dimension(ndim)::AM
  ! names
  integer,parameter::nmend=36
  character*2,dimension(nmend),parameter::mendeleiev=(/&
       'H ',&
       'He',&
       'Li',&
       'Be',&
       'B ',&
       'C ',&
       'N ',&
       'O ',&
       'F ',&
       'Ne',&
       'Na',&
       'Mg',&
       'Al',&
       'Si',&
       'P ',&
       'S ',&
       'Cl',&
       'Ar',&
       'K ',&
       'Ca',&
       'Sc',&
       'Ti',&
       'V ',&
       'Cr',&
       'Mn',&
       'Fe',&
       'Co',&
       'Ni',&
       'Cu',&
       'Zn',&
       'Ga',&
       'Ge',&
       'As',&
       'Se',&
       'Br',&
       'Kr'/)
  character*7,dimension(0:ndim)::specie
  ! useful indices
  integer::H1,HE3,HE4,C12,N14,O16,NE20,MG24,A23M,A23D,A21D,A21M,A25D,A25M
  ! status of each specie
  character*1,dimension(0:ndim)::rstatus,istatus
  ! atomic numbers
  integer,dimension(ndim)::IZ,nat,BN
  ! primaries
  integer,dimension(species)::ispecie
  integer,dimension(0:ndim)::iprim

  ! Reaction identification
  integer nreac
  double precision,dimension(nreacmax,6)::stoech
  integer,dimension(nreacmax,5)::zreac,ireac,ipreac,areac
  double precision,dimension(nreacmax)::Qvalue
  integer,dimension(nreacmax)::kprod,kreac
  character*7,dimension(nreacmax)::rname
  character*7::rdname
  character*1,dimension(nreacmax)::rtype
  integer::iRPP,iR33,iR34,iRBE,iRBP,iRPC,iRPN,iRPO,iR3A,&
       iRAC,iRAN,iRAO,iRANE,iRCC,iRCO,iROO,iRGNE,iRGMG,&
       iRCCG,iRPNG,iup23,ium23,iup21,ium21,iup25,ium25
CONTAINS
  SUBROUTINE paramchemistry(unit)
    implicit none
    ! Read the chemical parameters
    integer unit

    read(unit,*)!'---  Chemical parameters ---'
    read(unit,*)QVALUES
    
    ! threshold on the fractional abundances to control the time step
    read(unit,*)threshold
    read(unit,*)threflux
    read(unit,*)tiny
  END SUBROUTINE paramchemistry

  SUBROUTINE readtab
    ! Read the parameters for the chemical species

    use dimensions,only:nvarh,nvar,nameofvar,idx
    implicit none
    integer         i,l
    character*14 filename
    character*2 name

    PRINT*,' '
    PRINT*,' ================================================ '
    OPEN (UNIT=15, FILE = 'chemistry-data',STATUS = 'unknown') 
    PRINT*,' ====== CHEMISTRY FILE : chemistry-data   ======= '
    PRINT*,' ================================================ '

    ! --- INDICES FOR SPECIES
    H1   =0
    HE3  =0
    HE4  =0
    C12  =0
    N14  =0
    O16  =0
    NE20 =0
    MG24 =0
    A21M =0
    A21D =0
    A23M =0
    A23D =0
    A25M =0
    A25D =0
    ! --- SPECIES :                  
    !     ---------
    print*,'Species contained in file chemistry-data :'
2110 format('name    ',1x,' Z ',1x,' N ',1x,' A ',' X_init   ',2x,' mu(uma)')
    print 2110
    SPECIE(0)='(?????)'
    DEDUCE='not'
    l=0
    iprim(0)=0
    DO  I = 1,ndim
1000   format(a2,1x,i3,1x,2(a1),1x,1pe10.3,2x,1pe15.8)
       read(15,1000) name,BN(i),rstatus(i),istatus(i),frac_init(i),AM(i)
       specie(i)=nameofspec(name,BN(i))
       IZ(i)=znumber(name)
       nat(i)=BN(i)-IZ(i)

       if (rstatus(i)=='n'.or.rstatus(i)=='E'.or.&
            rstatus(i)=='C'.or.rstatus(i)=='D') then
          l=l+1
          if (l.gt.species) then
             print*,'Sorry, you cannot specify more than ',species,' primary species.'
             print*,specie(i),' is too much...'
             STOP
          endif
          ispecie(l)=i
          iprim(i)=l
       else
          iprim(i)=0
       endif
       if (rstatus(i)=='d'.and.DEDUCE=='yes') then
          print*,'I can only deduce ONE species...'
          print*,specie(i),' is too much...'
          STOP
       endif
       if (rstatus(i)=='e'.and.istatus(i).ne.' ') then
          print*,'WARNING ! For ',specie(i),' the mode "e" prevents me to'
          print*,'know its actual abundance. Then, it won''t use it in the'
          print*,'ionisation state.'
       endif
       if (rstatus(i)=='d'.and.DEDUCE=='not') DEDUCE='yes'

       ! --- IDENTIFICATION of SPECIES
       IF (SPECIE(I).EQ.'H1     ') H1     = I
       IF (SPECIE(I).EQ.'He3    ') HE3    = I
       IF (SPECIE(I).EQ.'He4    ') HE4    = I
       IF (SPECIE(I).EQ.'C12    ') C12    = I
       IF (SPECIE(I).EQ.'N14    ') N14    = I
       IF (SPECIE(I).EQ.'O16    ') O16    = I
       IF (SPECIE(I).EQ.'Ne20   ') NE20   = I 
       IF (SPECIE(I).EQ.'Mg24   ') MG24   = I
       IF (SPECIE(I).EQ.'Ne21   ') A21M   = I
       IF (SPECIE(I).EQ.'F21    ') A21D   = I
       IF (SPECIE(I).EQ.'Na23   ') A23M   = I
       IF (SPECIE(I).EQ.'Ne23   ') A23D   = I
       IF (SPECIE(I).EQ.'Mg25   ') A25M   = I
       IF (SPECIE(I).EQ.'Na25   ') A25D   = I
       ! Print what we've just read in chemistry-data
2111   format(a7,1x,3(i3,1x),1pe10.3,2x,1pe15.8)
       print 2111,specie(i),IZ(i),nat(i),BN(i),frac_init(i),AM(i)
    END DO
    

    if (l.lt.species) then
       print*,'I did not find all the primary species in the list.'
       print*,'I have ',l,' species instead of ',species,'.'
       print*,'Please provide more, or change the parameter species in'
       print*,'module dimensions.f90 and recompile all.'
       STOP
    endif
    ! Names of primary chemical variables
    nameofvar(nvarh+1+species:nvarh+species+species)=specie(ispecie(1:species))
    if (idx==2) then
       nameofvar(nvarh+1:nvarh+species)='g'//specie(ispecie(1:species))
    endif

    PRINT*,' '
    PRINT*,' Name in the list of species, rank'
    PRINT*,' H1     ', H1     
    PRINT*,' HE3    ', HE3    
    PRINT*,' HE4    ', HE4    
    PRINT*,' C12    ', C12    
    PRINT*,' N14    ', N14    
    PRINT*,' O16    ', O16    
    PRINT*,' NE20   ', Ne20   
    PRINT*,' MG24   ', MG24   
    PRINT*,' A23M   ', A23M   
    PRINT*,' A23D   ', A23D   

  end SUBROUTINE READTAB

  subroutine NETWORK
    ! Read the chemical network
    use constants,only:CMEVMU,CL
    implicit none
    INTEGER       i, j      
    character*2,dimension(5)::name
    character*3::dum
    double precision::checkmass,checkcharge

    iRPP=0
    iR33=0
    iR34=0
    iRBE=0
    iRBP=0
    iRPC=0
    iRPN=0
    iRPO=0
    iR3A=0
    iRAC=0
    iRAN=0
    iRAO=0
    iRANE=0
    iRCC=0
    iRCO=0
    iROO=0
    iRGNE=0
    iRGMG=0
    iRCCG=0
    iRPNG=0
    iup23=0
    ium23=0
    print*
    print*,' Those are the reactions found in the file chemistry-data :'
    i=0
    rdname=''
    do while(i.lt.nreacmax.and.rdname.ne.'END')
       i=i+1
1000   format(a7,1x,a1,5(1x,F3.1,1x,a2,1x,i3,2x))
       READ (15,1000) rdname,rtype(i),(stoech(i,j),name(j),areac(i,j),j=1,5)
       if (rdname.ne.'END')  then
          rname(i)=rdname
          kprod(i)=0
          kreac(i)=0
          Qvalue(i)=0.
          checkmass=0.
          checkcharge=0.
          do j=1,5
             zreac(i,j)=znumber(name(j))
             ireac(i,j)=identify(name(j),areac(i,j))
             ipreac(i,j)=iprim(ireac(i,j))
             if (stoech(i,j).ne.0.) then
                if (j.le.2)  kreac(i)=kreac(i)+1
                if (j.gt.2) then
                   kprod(i)=kprod(i)+1
                   stoech(i,j)=-stoech(i,j)
                endif
                Qvalue(i)=Qvalue(i)+stoech(i,j)*AM(ireac(i,j))
                checkmass=checkmass+stoech(i,j)*areac(i,j)
                checkcharge=checkcharge+stoech(i,j)*zreac(i,j)
             endif
          enddo
          stoech(i,6)=checkcharge
          if (checkmass.ne.0) then
             print*,'WARNING: mass is not conserved...'
             print*,'checkmass=',checkmass
          endif
          if (checkcharge.ne.0) then
             print*,'WARNING: charge is not conserved...'
             print*,'checkcharge=',checkcharge
          endif
          Qvalue(i)=Qvalue(i)*(cl**2/cmevmu)
          if (rname(i)=='RPP    ')  iRPP=i
          if (rname(i)=='R33    ')  iR33=i
          if (rname(i)=='R34    ')  iR34=i
          if (rname(i)=='RBE    ')  iRBE=i
          if (rname(i)=='RBP    ')  iRBP=i
          if (rname(i)=='RPC    ')  iRPC=i
          if (rname(i)=='RPN    ')  iRPN=i
          if (rname(i)=='RPO    ')  iRPO=i
          if (rname(i)=='R3A    ')  iR3A=i
          if (rname(i)=='RAC    ')  iRAC=i
          if (rname(i)=='RAN    ')  iRAN=i
          if (rname(i)=='RAO    ')  iRAO=i
          if (rname(i)=='RANE   ')  iRANE=i
          if (rname(i)=='RCC    ')  iRCC=i
          if (rname(i)=='RCO    ')  iRCO=i
          if (rname(i)=='ROO    ')  iROO=i
          if (rname(i)=='RGNE   ')  iRGNE=i
          if (rname(i)=='RGMG   ')  iRGMG=i
          if (rname(i)=='RCCG   ')  iRCCG=i
          if (rname(i)=='RPNG   ')  iRPNG=i
          if (rname(i)=='UR21+  ')  iup21=i
          if (rname(i)=='UR21-  ')  ium21=i
          if (rname(i)=='UR23+  ')  iup23=i
          if (rname(i)=='UR23-  ')  ium23=i
          if (rname(i)=='UR25+  ')  iup25=i
          if (rname(i)=='UR25-  ')  ium25=i
          if (i==irbe) zreac(i,2)=0
          if (i==irccg.or.i==irpng) then
             zreac(i,1)=0
             zreac(i,2)=0
          endif
          if (i==ir3a) zreac(i,1)=4
          ! Fudge the Q-values to mimic Eggleton code
          if (QVALUES=='egg') then
             if (iran.ne.0) Qvalue(iran)=4.415
             if (ircc.ne.0) Qvalue(ircc)=4.621
          endif
          ! Account for the reaction read :
1001      format(a7,(1x,F3.1,1x,a7),' + ',(1x,F3.1,1x,a7),'-->',&
               (1x,F3.1,1x,a7),' + ',(1x,F3.1,1x,a7),' + ',(1x,F3.1,1x,a7),&
               ' Q=',1pe10.3,' MeV')
          if (rdname.ne.'END') then
             print 1001,rname(i),(abs(stoech(i,j)),specie(ireac(i,j)),j=1,5),&
                  Qvalue(i)
          endif

       endif
    enddo
    read(15,1000) dum
    if (dum.ne.'END'.and.i.gt.nreacmax) then
       print*,'There are more reactions than allowed by the available memory.'
       STOP
    endif
    ! Total number of reactions
    nreac=i-1
    print*,' I found nreac=',nreac,' reactions in total.'
    print*,'RPP=',iRPP
    print*,'R33=',iR33
    print*,'R34=',iR34
    print*,'RBE=',iRBE
    print*,'RBP=',iRBP
    print*,'RPC=',iRPC
    print*,'RPN=',iRPN
    print*,'RPO=',iRPO
    print*,'R3A=',iR3A
    print*,'RAC=',iRAC
    print*,'RAN=',iRAN
    print*,'RAO=',iRAO
    print*,'RANE=',iRANE
    print*,'RCC=',iRCC
    print*,'RCO=',iRCO
    print*,'ROO=',iROO
    print*,'RGNE=',iRGNE
    print*,'RGMG=',iRGMG
    print*,'RCCG=',iRCCG
    print*,'RPNG=',iRPNG
    print*,'up23=',iup23
    print*,'um23=',ium23

  end subroutine NETWORK

  SUBROUTINE dimchemistry
    ! Give their proper dimension to the chemical parameters
    integer j,i
    ! reaction rates (depending on type)
  end SUBROUTINE dimchemistry

  character*7 function nameofspec(name,A)
    ! Build the name of a specie
    integer::A
    character*2::name

    nameofspec=trim(name)//trim(number(A))
    nameofspec=nameofspec//repeat(' ',7-len_trim(nameofspec))
    return
  end function nameofspec

  character*3 function number(i)
    ! Build the string 'i' from number i
    integer::i
    integer::n,u,b
    character*1,dimension(0:9)::digit=(/'0','1','2','3','4','5','6','7','8','9'/)
    b=INT(alog10(float(i)))
    n=i
    number=''
    do while (b.ge.0)
       u=n/10**b
       n=n-u*10**b
       if (b.le.2.and.n.lt.i) number=TRIM(number)//digit(u)
       b=b-1
    enddo
  end function number

  integer function znumber(name)
    ! Get the atomic number from a name
    character*2::name
    integer i
    znumber=0
    do i=1,nmend
       if (name==mendeleiev(i)) znumber=i
    enddo
    return
  end function

  integer function identify(name,a)
    ! Identify a specie from its name and mass number
    character*2::name
    integer::a,i
    identify=0
    if (a==0) return
    do i=1,ndim
       if (nameofspec(name,a)==specie(i)) identify=i
    enddo
    return
  end function

end MODULE chemistry
