MODULE constants       
  ! PHYSICAL CONSTANTS AND UNIT SYSTEM

  implicit none

  ! pi
  double precision::pimath

  ! useful constants for stellar evolution
  double precision::CPI, PI4, CLN10, CA, CB, CL, CD, CG, CR, CT, CEVB 
  double precision::CEN, CPL, CMEVMU, CSECYR, TSUNYR 
  double precision::AME, PLANCK, BOLTZM, ECHAR, EVOLT, CHE, LAMC, LAMC3, CRHO
  double precision::MSUN,RSUN,LSUN
  double precision::AMU
  double precision::BNEL

CONTAINS

  SUBROUTINE base
    use dimensions
    implicit none
    integer i,j

    !  pi is physics independant...
    pimath=acos(-1.D0)

    ! Constants usefull for stellar evolution 

   ! mathematical constants
    CPI = 4*ATAN(1.D0)
    PI4 = 4*CPI
    CLN10 = LOG(10.0D0)
    ! physical constants in cgs units, copied from JCD's code
    AMU = 1.6605402D-24
    AME = 9.1093897D-28
    BNEL=AME/AMU
    CL = 2.99792458D10
    PLANCK = 6.6260755D-27
    BOLTZM = 1.380658D-16
    ECHAR = 4.8032068D-10
    EVOLT = 1.60217733D-12
    CSECYR = 3.155692597D7
    CG = 6.672D-8
    ! derived physical constants
    CA = BOLTZM/(CL*PLANCK)
    CA = 8*CPI**5*BOLTZM*CA**3/15.
    CMEVMU = 1.0D6*EVOLT/AMU
    CEVB = EVOLT/BOLTZM
    CR = BOLTZM/AMU
    CHE = AME*CL**2
    CT = BOLTZM/CHE
    LAMC = PLANCK/(AME*CL)
    LAMC3 = LAMC**3
    CRHO = 8*CPI/LAMC3
    CB = CRHO*CHE
    CD = CRHO*AMU
    CEN = PLANCK/SQRT(2*CPI*AMU*BOLTZM)
    CEN = CEN**3/AMU
    CPL = SQRT(PI4/(AMU*BOLTZM**3))*ECHAR**3
    ! solar mass, luminosity, radius, age
    MSUN = 1.9891D0
    LSUN = 3.844D0
    RSUN = 0.69598D0
    TSUNYR = 4.57D9
  end SUBROUTINE base

end MODULE constants
