MODULE dimensions       
  ! DIMENSIONS OF THE ARRAYS AND VARIABLES IDENDTIFICATION
  ! PRIMARIES AND TIME STEP.

  implicit none

  ! Verbose for debug
  integer::verbose

  ! number of zones
  integer,parameter::nz=199!700

  ! number of variables
  ! structure variables
  integer,parameter::nvarh=7  ! lnf,lnT,lnm,lnr,L,dm,nsmooth
  ! composition variables 
  !(you need to ajust these numbers according to the 'chemistry-data' list)
  integer,parameter::species=5!5!8!5!2          ! number of species solved for
  integer,parameter::ndim=11!11!10!14!17!14     ! total number of species
  ! total number of variables
  integer,parameter::idx=2    ! If chemical gradients are solved for
  integer,parameter::nvar=nvarh+species*idx

  ! Total number of equations
  integer,parameter::neq=nvar*nz

  ! Maximum number of chemical reactions
  integer,parameter::nreacmax=30

  ! matrix structure
  integer,parameter::m1=3*nvar!2*nvar-5!+1!-4          ! subdiagonals
  integer,parameter::m2=3*nvar!2*nvar-5!+1!-4          ! superdiagonals
  integer,parameter::ndiag=m1+1+m2      ! number of diagonals
  integer,parameter::idiag=m1+1         ! central diagonal position

  ! identification of the variables:
  integer,parameter:: ilnf = 1 
  integer,parameter:: ilnT = 2
  integer,parameter:: ilnm = 3 
  integer,parameter:: ilnr = 4
  integer,parameter:: iL   = 5
  integer,parameter:: idm  = 6
  integer,parameter:: ins  = 7
  ! Names of variables
  character*7,dimension(nvar)::nameofvar
  ! Which variables are >0 or >=0
  logical,dimension(nvar)::defpos,varpos
  ! Which ones are defined on interfaces
  logical,dimension(nvar)::interf

  ! Vector of the primaries: current, previous, increment, previous increment
  double precision,dimension(neq)::x,xold,dx,dxold

  ! Time: current, timestep, old timestep, previous old timestep
  double precision::time,dt,dtold,dtsave

  ! Structure of the vector of the primaries
  integer,dimension(neq)        ::zone,variable

CONTAINS

  SUBROUTINE identify
    implicit none
    integer iz,jvar,i
    ! description of the structure in zones and variables of 
    ! the solution vector
    ! iequation --> izone,ivariable
    do iz=1,nz+1
       do jvar=1,nvar
          i=(iz-1)*nvar+jvar
          if (i.le.neq) then
             zone(i)=iz
             variable(i)=jvar
          endif
       enddo
    enddo

    ! Names of the structure variables
    nameofvar(1:nvarh)=(/&
         'lnf    ',&
         'lnT    ',&
         'lnm    ',&
         'lnr    ',&
         'L      ',&
         'dm     ',&
         'nt     '/)
    ! The remainder of this array is initialized in chemistry$readtab
    ! with the names of the chemical species found in the network.

    ! Positive and non zero variables
    defpos=.false.

    ! Positive variables 
    varpos=defpos
    if (idx==2) then
       varpos(nvarh+species+1:nvarh+species+species)=.true. ! Abundances
    else
       varpos(nvarh+1:nvarh+species)=.true.
    endif

    ! Variables computed at interface
    interf=.false.
    interf(iL)=.true.
    interf(ilnr)=.true.
    interf(ilnm)=.true.
    if (idx==2) then
       interf(nvarh+1:nvarh+species)=.true. ! Chemical gradients
    endif
  end SUBROUTINE identify

end MODULE dimensions

