! Originally, it was routine equns1.f from STAR program.
! I do not think there is one single line in common now..
!
! This module takes the local variables of modules
! 'variables' and 'secondaries' and use them to build the equations
! of stellar evolution which mix variables between different zones
! or interfaces.
!
MODULE equations
  use dimensions,only:nz,species,neq,time,verbose,dt
  use parameters,only:TRUEL,tcentre,MIXSCHEME,STAGGERED,&
       imuidn,tcool,taccretion,NOFLUX
  use secondaries
  use variables
  use maths,only:ps1,ps2,ps3,ps,pps,lps,argsh,cbrt
  implicit none
  logical::noprint=.false.,computequ=.false.
  logical::relax=.false.

  ! convective boundaries
  type boundary
     integer::iold,iz,ir,ic
     double precision::m,m0,mt,mt0,dm
     double precision::xm,xp
     double precision::dlnt
     logical::top,advancing
  end type boundary
  type(boundary),dimension(nz)::cb,cbsave,cbold
  logical,dimension(0:nz+2)::cv,cvold,cvsave
  logical,dimension(nz)::isolated
  integer::ncb,ncbsave,ncbold=0
  integer,dimension(nz)::icb
  logical::noflagcb=.false.,savemt=.true.
 
  ! Old profiles
  double precision,dimension(nz,2)::lnToldrc,lnToldsave
  double precision,dimension(nz,ndim,2)::naoldrc,naoldsave

  ! Convective criterion
  double precision,dimension(nz)::galpha,gbeta,gl
  double precision,dimension(nz,ndim)::gphi

  ! Convective time scale
  double precision,dimension(nz)::tconv
  ! Burning time scale
  double precision,dimension(nz)::tburn,tburn2,tburnold,tburnsave

  ! Gradients
  double precision,dimension(nz)::gradeN,gradt
  double precision,dimension(nz)::grade,gradTa
  double precision,dimension(nz,ndim)::gradN,gradtN
  double precision,dimension(nz)::gradesave,gradtasave
  double precision,dimension(nz,ndim)::gradNsave
  double precision,dimension(nz)::gradeold,gradtaold
  double precision,dimension(nz,ndim)::gradNold

  
  ! to printings
  double precision,dimension(nz)::gradr,gradl,gradc,grads
  double precision,dimension(nz)::factor1,factor2

  double precision,dimension(nz)::equP,equr,equT,equL,equm,equdm,equsmooth
  double precision,dimension(nz,species)::equX
  double precision,dimension(nz,species)::equdX

  double precision,dimension(nz)::dr3k,dr3kold

  double precision,dimension(nz)::rhowant,Twant

  double precision,dimension(nz)::muidN,muedne

  ! moving mesh
  double precision,dimension(nz)::n,nold
  double precision,dimension(nz-2)::ntilde,ntildeold
  double precision::taugrille
  double precision,dimension(nz)::dmt,dmt2
  double precision,save,dimension(nz)::mts
  double precision,dimension(nz-1)::qk
  double precision,dimension(nz)::res,ires

  ! driving mode
  character*3 MODE

  ! resolution
  double precision,dimension(nz)::nhat

  ! Composition equation terms
  double precision,dimension(nz)::DtN,diffN,advN,RN,normX,fluxN
  double precision,dimension(nz,species)::diffus,DtNlag
  double precision,dimension(nZ)::Xadv

  ! Luminosity equation terms
  double precision,dimension(nz)::normL
  double precision,dimension(nz)::TadS,DtE
  ! Resolution function arguments
  integer,parameter::narg=3+species
  double precision,dimension(nz,narg)::arg

  ! EQUATIONS LOC
  ! Convective state parameters and secondaries wanted in printings
  double precision,dimension(nz)::dglg,dg,lg,dtg
  double precision,dimension(nz)::Ftyp,Ltypold,Ltyp,Ltypsave,dldpold,dldpsave
  double precision,dimension(nz,species)::Ftypold,Ftypsave
  double precision,dimension(nz+1)::Ltot
  double precision,dimension(nz+1,species)::Fconv,Fconvold,Fconvsave
  double precision,dimension(nz)::Ftop,Fbottom
  double precision,dimension(nz)::Lconv,Wconv,pconv,Lrad,md,conv,vcud3
  double precision,dimension(nz)::chic,chiT,chiM
  double precision,dimension(nz)::Hp,lm
  double precision,dimension(nz)::gradt1,gradt2
  double precision,dimension(nz)::dlnp,dlnT,dxlnp,dxlnT,dxenergy,denergy
  double precision,dimension(nz)::vc,vcc
  double precision,dimension(nz)::dcubic,u0,u0m,u1,uD
  double precision,dimension(nz,ndim)::gamman
  double precision,dimension(nz,species)::dN,dxN
  double precision,dimension(nz)::qm
  double precision,dimension(nz)::mp
  double precision,dimension(nz)::dmold0


  ! Number of zones across which to keep chemistry uniform at surface
  integer,parameter::isurf=5
contains

  SUBROUTINE EQUNS1
    use constants,only:Lsun,pi4,cg,csecyr,msun
    !!    use chemistry
    use chemistry,only:istatus,rstatus,ispecie,specie,frac_init,&
         iprim,H1,Ne20,He4,threshold,threflux,A23D,A23M,C12,O16,Ne20,bn,iup23,izat=>iz
    use nuclear,only:ex,enx,rate
    use EOS,only:p
    use dimensions,only:dt,dtold,idx
    use parameters,only:RESOLUTION,MESH,taugref,stiffness,MIXING,&
         alphagrid,PARTCONV,&
         ifix,nbs,DOSAVE
    implicit none
    integer i,j,iz,jc
    double precision::dmconv,factt,xsurf

    if (verbose.ge.2) print*,'  equns1'

    if (MODE.eq.'rho') then
       ! Mass density is given.
       equP=rho/rhowant-1.
       equr=lnr-lnrold
       equT=lnT-log(Twant)
       equL=L-Lold
       equm=lnm-lnmold
       equdm=dm-dmold
       equsmooth=nsmooth-nsmoothold
       EquX=frac-fracold
       EqudX=gfrac-gfracold
       return
    endif
    if (MODE.eq.'i=0') then
       ! Get density for zero net heating
       equP=rho-rhowant
       equT=Netheat
       equr=lnr-lnrold
       equL=L-Lold
       equm=lnm-lnmold
       equdm=dm-dmold
       equsmooth=nsmooth-nsmoothold
       EquX=frac-fracold
       EqudX=gfrac-gfracold
       return
    endif

    ! volume of each cell
    dr3k(1:nz-1)=(r(1:nz-1)-r(2:nz))&
         *(r(1:nz-1)*r(1:nz-1)+r(1:nz-1)*r(2:nz)+r(2:nz)*r(2:nz))
    dr3k(nz)=r(nz)**3
    dr3kold(1:nz-1)=(rold(1:nz-1)-rold(2:nz))&
         *(rold(1:nz-1)*rold(1:nz-1)+rold(1:nz-1)*rold(2:nz)+rold(2:nz)*rold(2:nz))
    dr3kold(nz)=rold(nz)**3

    if (savemt.or.DOSAVE.ne.'yes') mts=mt
    if (NOFLUX.eq.'yes'.or.relax) then
       mts=0.
       mt=0.
       dmold0=dm
    else
       dmold0=dmold
    endif

    ! intercell mass
    dmt(1)=dm(1)
    dmt(2:nz)=.5*(dm(1:nz-1)+dm(2:nz))
    dmt=MAX(1.d-100,dmt)

    ! Convective theory
    call convection

    ! mesh spacing
    n=1.d0/dm
    nold=1.d0/dmold
    if (MESH.eq.'old'.or.MESH.eq.'OLD') then
       ntilde=n(2:nz-1)
       ntildeold=nold(2:nz-1)
    else
       ntilde=n(2:nz-1)-stiffness*(1.d0+stiffness)*(n(1:nz-2)-2.*n(2:nz-1)+n(3:nz))
       ntildeold=nold(2:nz-1)-stiffness*(1.d0+stiffness)&
            *(nold(1:nz-2)-2.*nold(2:nz-1)+nold(3:nz))
    endif

    ! Grid adaptive time
    taugrille=taugref*tgold
    if (taugref.lt.0.) taugrille=-taugref*csecyr

    ! Time smoothed grid point density
    nhat=nsmooth+taugrille*invdt*(nsmoothnew-nsmoothold)


    ! Scales for luminosity and particle flow
    Ltyp=(surf*rho*1.d-11*cp*T*ch)/hp
    Ftyp=surf*rho*1.d-11*wm(u0)*vc/(vc+wm(u0))

    ! mesh resolution
    qk=q(1:nz-1)-q(2:nz)
    if (RESOLUTION=='old') then
       ! STAR like resolution
       res(1:nz-1)=qk(1:nz-1)*n(1:nz-1)
       res(nz)=bc4*n(nz)
    else
       ! Dorfi&Drury (84) type of resolution
       arg=0.
       arg(:,1)=1.d0
       factt=1.+0.*0.5
       if (lnpold(nz)-lnpold(1).ne.0.)&
            arg(:,2)=mold(1)/(lnpold(nz)-lnpold(1)+&
            factt*(maxval(lntold)-minval(lntold)))*float(nbs)/(float(nz-nbs))&
            *wp(pm)/p*(1.+factt*abs(dlnT/dlnP))
       arg(:,narg-species+1)=0.
       do j=1,species
          ! with dxN
          !W!             arg(:,narg-species+1)=MAX(arg(:,narg-species+1),&
          !W!                  abs(grad(log(alphagrid*(1.d-8+MAXVAL(fracold(:,j)))+frac(:,j)))))          
          arg(:,narg-species+1)=MAX(arg(:,narg-species+1),&
               abs(grad(gfrac(:,j)))/sqrt(gfrac(:,j)**2+((surf*rho*1.d-11*ch)/hp)**2)*m/(.1*Msun+m))
       enddo
       arg(:,narg-species+1)=.1*arg(:,narg-species+1)*&
            4.!/(20.*log(10.))*log((2.+pold(nz))/(1.+pold(1)))
       if (alphagrid.ne.0.) then
          arg(:,3)=alphagrid*arg(:,3)
          arg(:,4)=alphagrid*arg(:,4)
       endif
       res=1.d0
       do i=1,2
          res=res+arg(:,i)**2
       enddo
       if (alphagrid.ne.0.) then
          do i=3,4
             res=res+arg(:,i)**2
          enddo
       endif
       res=sqrt(res)
    endif
    ! mass equation
    mp(1:nz-1)=m(2:nz)
    mp(nz)=0.
    equm=m-(mp+wc(dm))

    ! mesh smoothing equation
    equsmooth(1)=nsmooth(1)/n(1)-1.d0
    equsmooth(2:nz-1)=nsmooth(2:nz-1)/ntilde-1.d0    
    equsmooth(nz)=nsmooth(nz)/n(nz)-1.d0
    if (MESH=='lag'.or.MESH=='old'.or.MESH=='OLD') equsmooth=nsmooth/n-1.d0

    ! mesh equation
    equdm(1)=equm(1)
    equm(1)=bc1    
    equdm(2:nz)=(nhat(2:nz)*ires(2:nz)-nhat(1:nz-1)*ires(1:nz-1))
    equdm(2:nz)=(nhat(2:nz)/res(2:nz)-nhat(1:nz-1)/res(1:nz-1))&
         *1.!/MAXVAL((1.d0+nsmoothold)/(1.d0+resold))
    if (RESOLUTION.ne.'old') then
       equdm(nz)=dm(nz)/dmold(nz)-1.d0
       if(tcentre.ne.0.) then
          equdm(nz)=dm(nz)/dmold(nz)-(1.d0+dt/tcentre/csecyr)
       endif
    endif
    if (MESH=='old') then
       equdm(2:nz-1)=(qk(2:nz-1)/qk(1:nz-2))-1.d0
       equdm(nz)=qk(nz-1)/bc4-1.d0
    endif
    if (MESH=='OLD') then
       equdm(2:nz)=dm(1:nz-1)*qm(1:nz-1)/(dm(2:nz)*qm(2:nz))-1.d0
    endif
    if (MESH=='lag') then
       equdm=dm/dmold-1.d0
    endif
    ! keep the mesh fixed on last ifix zones or first -ifix zones
    if (abs(ifix).gt.nz) then
       print*,'|ifix|>nz'
       STOP
    endif
    if (ifix.ge.1) then
       equm(nz-ifix:nz)=m(nz-ifix:nz)/mold(nz-ifix:nz)-1.d0
       equdm(nz-ifix:nz)=dm(nz-ifix:nz)/dmold(nz-ifix:nz)-1.d0
    else
       if (ifix.le.-1) then
          equdm(1:-ifix)=dm(1:-ifix)/dmold(1:-ifix)-1.d0
       endif
    endif

    ! Physics equations

    ! Radius equation
    equr=(wc(dm*(1./rho))-pi4*dr3k/3.)*rhoold/mold

    ! Pressure equation
    equP(1)=1.d0-bc2/(1.0D11*CG*.5*(m(1)+m(2))/(.5*(r(1)+r(2)))**2)
!    equP(1)=1.d0-bc2/(1.0D11*CG*m(1)/r(1)**2)
    equP(2:nz)=(lnp(1:nz-1)-lnp(2:nz))-dlnp(2:nz)

    ! Temperature equation
    equT(1)=1.d0-bc3*.5*(Ltot(1)+Ltot(2))/(.5*(r(1)+r(2)))**2
!    equT(1)=1.d0-bc3*Ltot(1)/r(1)**2
    equT(2:nz)=(lnT(1:nz-1)-lnT(2:nz))-dlnT(2:nz)

    ! Luminosity equation
    ! advection term
    selectcase (truel)
    case('TdS')
       TadS=adv(dxenergy)
    case('egg')
       TadS=adv(wm(exp(lnT)*cp*dtg)*dxlnp)
    case default
       print *,'Only TRUEL=''TdS'' or ''egg'' are implemented... sorry !'
       STOP
    endselect
    ! Lagrange time derivative of energy
    DtE=invdt*cp*T*(wa(lnt-oldrc(lnToldrc))-grada*(lnp-lnpold))*dmold0&
         +wa(Tst2)*dm-Wconv*dm-Tads
    muidN=0.
    do i=1,ndim
       muidN=muidN+(mui(:,i)-.5*mue*bn(i))*invdt*wa(na(:,i)-oldrc(naoldrc(:,i,:)))*dmold0
    enddo
    DtE=DtE+imuidN*muidN
    ! Total luminosity
    equL=DtE+Ltot(1:nz)-Ltot(2:nz+1)
    ! Flux at convective interfaces due to chemistry discontinuous
    if (PARTCONV.eq.'yes') then
       ! Correction at CB 
       if (imuidN.ne.0.) then
          do j=1,ncb
             if (cb(j)%advancing) then
                iz=cb(j)%iz
                if (iz.gt.isurf) then
                   do i=1,ndim
                      equL(iz)=equL(iz)-(mui(iz,i)-.5*mue(iz)*bn(i))*cb(j)%mt0*&
                           (na(iz-1,i)-na(iz+1,i))*imuidN
                   enddo
                endif
             endif
          enddo
       endif
    endif
    ! Normalise
    normL=0.
    if (dtold.ne.0.) normL=max(normL,abs(Uold/dtold*dmold))
    normL=max(normL,Ltypold)
    normL=max(abs(Lold),normL)
    normL(1:nz-1)=max(abs(Lold(2:nz)),normL(1:nz-1))
    !    normL=max(abs(tst-tads),normL)
    if (all(normL.ne.0.))equL=equL/normL

    ! Composition equations
    if (dt.eq.0.) then
       equx=fracnew-fracold
       return
    endif
    diffus=0.
    do i=1,species
       if (rstatus(ispecie(i)).eq.'E') then
          equX(:,i)=xt2(:,i)/MAX(fracold(:,i),threshold)*dtold
       else
          if (rstatus(ispecie(i)).eq.'C') then
             equX(:,i)=(fracnew(:,i)-fracold(:,i))&
                  /MAX(fracold(:,i),threshold,threshold)
          else
             ! Fluxes at top and bottom of the zone
             Ftop=   Fconv(1:nz  ,i)
             Fbottom=Fconv(2:nz+1,i)
             ! Advection 
             Xadv=bn(ispecie(i))*adv(dxN(:,i))      
             ! Lagrange derivative
             DtNlag(:,i)=invdt*wa(na(:,ispecie(i))-oldrc(naoldrc(:,ispecie(i),:)))&
                   *bn(ispecie(i))*dmold0+wa(xt2(:,i))*dm-Xadv
             ! Correction from moving discontinuities
             if (PARTCONV.eq.'yes') then
                ! Correction at CB
                do j=1,ncb
                   if (cb(j)%advancing) then
                      iz=cb(j)%iz
                      if (iz.gt.isurf) &
                           Ftop(iz)=Ftop(iz)-cb(j)%mt0*(frac(iz-1,i)-frac(iz+1,i))
                   endif
                enddo
             endif
             ! Diffusion term
             diffus(:,i)=Ftop-Fbottom
             ! Final equation
             equx(:,i)=DtNlag(:,i)+diffus(:,i)
             ! Normalise
             normX=0.
             if (dtold.ne.0.)&
                  normX=max(normX,ABS(dmold*MAX(threshold,fracold(:,i)))/dtold)
             where (normX.ne.0.) equX(:,i)=equX(:,i)/normX

             ! Printing stuff
             !!             if (i.eq.iprim(H1)) then
             DtN=xt1(:,i)*dmold0-Xadv
             diffN=diffus(:,i)
             RN=xt2(:,i)*dm
             advN=Xadv
             FluxN=Fconv(1:nz,i)
             !!             endif
             if (verbose.le.-4) noprint=.true.
             if (i.le.2.and..not.noprint) then
                print*,specie(ispecie(i))
                print 2
                do iz=1,nz,10
1                  format(i6,2x,40(1pe10.3,1x))
2                  format(' zone  ',2x,&
                        'equX       ',&
                        'normX      ',&
                        'DtN        ',&
                        'diffN      ',&
                        'RN         ',&
                        'advN       ',&
                        'FluxN      ',&
                        'Ftypold    ')
                   if (normx(iz).ne.0.) then
                      print 1,iz,equx(iz,i)/normx(iz),&
                           normX(iz),&
                           DtN(iz)/normX(iz),&
                           diffN(iz)/normX(iz),&
                           RN(iz)/normX(iz),&
                           advN(iz)/normX(iz),&
                           FluxN(iz),&
                           Ftypold(iz,i)
                   else
                      print *,iz,'normX==0',Ftypold(iz,i)
                   endif
                enddo
                print 2
             endif
             if (i == 2.and. .not.noprint) then
                read(*,*)
             endif
          endif
       endif
    enddo
    ! Composition uniform in first isurf zones
    equx(1:isurf,:)=frac(1:isurf,:)-frac(2:isurf+1,:)

    ! composition gradient equations:
    if (idx==2) then
       if (MIXSCHEME=='egg') then 
          ! Eggleton like, but probably obsolete
          equdx=gfrac-Fconv(1:nz,:)
       else
          do j=1,species
             i=ispecie(j)
             equdx(2:nz,j)=bn(i)*(Na(1:nz-1,i)-Na(2:nz,i)-dN(2:nz,j))
          enddo
          equdx=equdx/MAX(threshold,fracold)
       endif
    endif
    ! Gradient extrapolated for radiatie interface next to an advancing CB.
    if (PARTCONV.eq.'yes') then
       ! Correction at CB
       do i=1,ncb
          if (cb(i)%advancing) then
             iz=cb(i)%iz
             if (cb(i)%top) then
                equdx(iz,:)=gfrac(iz,:)-gfrac(iz-1,:)
             else
                if (iz.lt.nz-1) then
                   equdx(iz+1,:)=gfrac(iz+1,:)-gfrac(iz+2,:)
                else
                   equdx(iz+1,:)=gfrac(iz+1,:) 
                endif
             endif
          endif
       enddo
    endif
    ! first few zones mixed
    equdx(1:isurf+1,:)=gfrac(1:isurf+1,:)
        
  END SUBROUTINE EQUNS1

  SUBROUTINE savesecold
    ! Save old secondaries : probably more than half of these
    ! are not used anymore some cleaning would be good..
    use variables,only:t2nd
    use dimensions,only:dtold,dtsave
    t2nd=.true.
    dtsave=dtold
    if (verbose.ge.1) print*,' savesecold'
    rhoisave=rhoiold
    rhosave=rhoold
    usave=uold
    TdtSsave=TdtSold
    lnpsave=lnpold
    ssave=sold
    ressave=resold
    mtsave=mtold
    nesave=neold
    muesave=mueold
    muisave=muiold
    nasave=naold
    diffsave=diffold
    dgsave=dgold
    Ltotsave=Ltotold
    convsave=convold
    netheatsave=netheatold
    dglgsave=dglgold
    lgsave=lgold
    tkhsave=tkhold
    Fconvsave=Fconvold
    tcsave=tcold
    gradnsave=gradnold
    gradesave=gradeold
    gradTasave=gradTaold
    Ltypsave=Ltypold
    Ftypsave=Ftypold
    dtxsave=dtxold
    tgsave=tgold
    lnToldsave=LnToldrc
    naoldsave=naoldrc
    cvsave=cvold
    cbsave=cbold
    ncbsave=ncbold
    tburnsave=tburnold
  end SUBROUTINE savesecold

  SUBROUTINE recoversecold
    ! Recover old secondaries
    use variables,only:t2nd
    if (verbose.ge.1) print*,' recoversecold'
!    t2nd=.false.
    TdtSold=TdtSsave
    rhoiold=rhoisave
    rhoold=rhosave
    uold=usave
    lnpold=lnpsave
    sold=ssave
    resold=ressave
    mtold=mtsave
    neold=nesave
    mueold=muesave
    muiold=muisave
    naold=nasave
    diffold=diffsave
    dgold=dgsave
    Ltotold=Ltotsave
    convold=convsave
    netheatold=netheatsave
    dglgold=dglgsave
    lgold=lgsave
    tkhold=tkhsave
    Fconvold=Fconvsave
    tcold=tcsave
    gradnold=gradnsave
    gradeold=gradesave
    gradTaold=gradTasave
    Ltypold=Ltypsave
    Ftypold=FTypsave
    dldpold=dldpsave
    dtxold=dtxsave
    tgold=tgsave
    lnToldrc=lnToldsave
    naoldrc=naoldsave
    cvold =cvsave
    cbold =cbsave
    ncbold=ncbsave
    tburnold=tburnsave
  end SUBROUTINE recoversecold

  SUBROUTINE setsecold
    ! Set old secondaries
    use dimensions,only:dt,nreacmax,species
    use parameters,only:rmg
    use constants,only:CG
    use EOS,only:ne,energy,s
    use nuclear, only:rate,qcc
    use chemistry,only:rtype
    use maths,only:smax
    implicit none
    integer::i,j,nj
    integer::iz1,iz2
    double precision::TTc
    double precision::tchem
    double precision::dlntsum
    if (verbose.ge.1) print*,' setsecold'

    ! Before old values are changed, compute change in lnT within
    ! each convective region. This is to get an estimate of the
    ! timescale on which each boundary moves.
    iz1=1
    do i=1,ncb
       if (cb(i)%top) iz1=cb(i)%iz+1
       if (.not.cb(i)%top.or.i==ncb) then
          if (i.eq.ncb.and.cb(i)%top) then
             iz2=nz
          else
             iz2=cb(i)%iz-1
          endif
          if (iz2.ge.iz1) then
             ! Average change of lnT over convective region
             dlntsum=0.
             nj=0
             do j=iz1,iz2
                if(cvold(j).and.cvold(min(j+1,nz))) then
                   dlntsum=dlntsum+abs(lnt(j)-lntold(j))
                   nj=nj+1
                endif
             enddo
             if (nj.ge.1) dlntsum=dlntsum/float(nj)
          else
             dlntsum=0.
          endif
          cb(i)%dlnt=dlntsum
          cb(max(1,i-1))%dlnt=dlntsum
       endif
    enddo

    ! Compute old values
    rhoiold=1.d0/rho
    rhoold=rho
    uold=energy
    lnpold=lnp
    sold=s
    Tdtsold=Tdts
    resold=res
    mtold=mt
    neold=ne
    mueold=mue
    muiold=mui
    naold=na
    diffold=diffus
    dgold=dg
    Ltotold=Ltot
    convold=conv
    netheatold=netheat
    dglgold=dglg    
    lgold=lg
    Fconvold=Fconv
    dtxold=dtx
    do i=1,ndim
       gradnold(:,i)=1.d0!UD!/(1.d0+vc/wm(ud))!gradn/gradtn
    enddo
    gradeold=grade!/gradr
    gradTaold=gradTa!/gradr
    Ltypold=Ltyp!(surf*rho*cp*T*1.d-11*ch)/hp
    do i=1,species
       Ftypold(:,i)=Ftyp!(surf*rho*1.d-11*chid)/hp
    enddo
    
    ! Grid time is the min of few of the following times
    ! Thermal diffusion time
    tdold=0.
    do i=1,nz
       tdold=tdold+(dm(i)/rho(i)/surf(i)*1d11)**2/ch(i)
    enddo
    tgold=tdold
    ! Kelvin-Helmholtz time
    tKHold=CG*Mold(1)**2/Lold(1)/Rold(1)*1.e22
! K-H is not good for cooling WD.. better thermal diffusion time
!    tgold=tKHold
    ! Nuclear times
    if (ANY(rate.ne.0.)) then
       tchem=MIN(tgold,1.d0/MAXVAL(abs(rate)))
    else
       tchem=tgold
    endif
    tcold=tchem
    ! Accretion time
    if (rmg.ne.0.) tgold=MIN(tgold,1.e-2/abs(rmg))
    
    ! Old convective boundaries
    cvold=cv
    cbold=cb
    ncbold=ncb

    ! Extrapolate radiative and convective profiles
    call extraprall

    ! Burning time scale
    tburn=0.
    where (qcc.ne.0.) tburn=cp*T/qcc
    tburn2=0.
    do i=1,nz
       if (tburn(i).ne.0..and.tburnold(i).ne.0..and.tburn(i).ne.tburnold(i)) then
          tburn2(i)=dtlnT(i)/(1./tburn(i)-1./tburnold(i))
       endif
    enddo
    tburnold=tburn


  end SUBROUTINE setsecold

  subroutine extraprall
    implicit none
    integer j
    ! Extrapolate radiative and convective profiles
    lnToldrc(:,1)=extrapro(lnT,gradT,1.d0)
    lnToldrc(:,2)=lnT!extrapro(lnT,gradT,-1.d0)
    do j=1,ndim
       naoldrc(:,j,1)=extrapro(na(:,j),gradN(:,j),1.d0)
       naoldrc(:,j,2)=Na(:,j)!extrapro(na(:,j),gradN(:,j),-1.d0)
    enddo
  end subroutine extraprall

  function extrapro(x,gx,eps) result(y)
    ! This routine extrapolates the profile of x 
    ! thanks to its gradient gx from convective (eps>0)
    ! or radiatie (eps<0) regions outward.
    use parameters,only:PARTCONV
    implicit none
    double precision,dimension(nz)::x,y,gx
    double precision::eps
    double precision,dimension(nz)::xb,gxb,dltlnp
    integer,dimension(nz)::ip,dir
    logical,dimension(nz)::alive,record
    integer::ibot,nbot,i,j
    logical,dimension(nz)::isok
    integer::ipass
    integer,parameter::npass=3

    ! Copy values
    y=x
    if (PARTCONV.ne.'yes') return

    ! Convective state
    isok=(gradc*eps).gt.0.
    
    ! Create bots
    ibot=0
    alive=.false.
    record=.false.
    do i=1,nz-1
       if (isok(i).and.(.not.isok(i+1)).and.i.gt.1) then
          !          if (i-2.ge.1.and.isok(i-1).and.isok(i-2)) then
          ibot=ibot+1
          dir(ibot)=1
          ip(ibot)=i-1
          xb(ibot)=x(i-1)
          gxb(ibot)=gx(i-1)
          alive(ibot)=.true.
          if (i-2.ge.1.and.isok(i-1).and.isok(i-2)) then
             record(ibot)=.true.
          endif
       endif
       if (isok(i+1).and.(.not.isok(i))) then
          !          if (i+3.le.nz.and.isok(i+2).and.isok(i+3)) then
          ibot=ibot+1
          dir(ibot)=-1
          ip(ibot)=i+1
          xb(ibot)=x(i+1)
          gxb(ibot)=gx(min(nz,i+2))
          alive(ibot)=.true.
          if (i+3.le.nz.and.isok(i+2).and.isok(i+3)) then
             record(ibot)=.true.
          endif
       endif
    enddo
    nbot=ibot

    ipass=1
    do while(any(alive).and.ipass.le.npass)
       ! Move bots
       do i=1,nbot
          if (alive(i)) then
             do j=1,nbot
                ! Kill bots if collision
                if (ip(j).eq.ip(i)+dir(i).and.alive(j).and.alive(i)) then
                   alive(j)=.false.
                   alive(i)=.false.
                endif
             enddo
             if (alive(i)) ip(i)=ip(i)+dir(i)
          endif
       enddo
       ! Kill bot if out of range
       where(alive.and.ip.gt.nz) alive=.false.
       where(alive.and.ip.lt.1) alive=.false.
!!$       ! Kill bot if it crosses an OK interface
!!$       where(alive.and.dir.eq.1 .and.isok(ip))   alive=.false.
!!$       where(alive.and.dir.eq.-1.and.isok(ip+1)) alive=.false.
       ! Get dlnp increment
       where(alive.and.dir.eq. 1) 
          dltlnp=-dlnp(ip)
       endwhere
       where(alive.and.dir.eq.-1) 
          dltlnp=dlnp(ip+1)
       endwhere
       ! Update their value
       where(alive) xb=xb+dltlnp*gxb
       ! copy it in y
       where(alive.and.record) y(ip)=xb       
       ipass=ipass+1
    enddo
  end function extrapro

  function oldrc(f) Result(frc)
    ! Old value may be extrapolated from the old profile
    ! sometimes..
    implicit none
    double precision,dimension(nz,2)::f    
    double precision,dimension(nz)::frc
    !WARNING! Extrapolated values may still not correspond to the same
    ! convective region: need to check...
    ! Need extrapolation only when new zone has a convective interface
    ! and old zone was fully radiative.
    where((cv(1:nz).or.cv(2:nz+1)).and..not.(cvold(2:nz+1).or.cvold(2:nz+1))) 
       frc=f(:,1)
    elsewhere
       frc=f(:,2)
    endwhere
  end function oldrc




  function adv(df) Result(fa)
    ! Computes advection term
    use maths,only:ps
    use parameters,only:VANLEER
    implicit none
    double precision,dimension(nz)::df,fa
    ! locals
    double precision,dimension(nz)::mtl,mtls
    double precision,dimension(nz)::dfl,dfr,dfv

    mtl(1:nz-1)=mt(2:nz)
    mtl(nz)=0. ! Center has always same mass (0)
    mtls(1:nz-1)=mts(2:nz)
    mtls(nz)=0. ! Center has always same mass (0)
        if (STAGGERED=='yes') then
       dfr=df*dmt
    else
       dfr=df*dm
    endif
    dfl(1:nz-1)=-dfr(2:nz)
    dfl(nz)=0.  
    where (mts.ge.0. .and. mtls.le.0.)
       fa=( (dfr-dfl)*(mt)+dfl*invdt*(dm-dmold0) )
    elsewhere
       fa=dfr*ps(mt)+dfl*ps(-mtl)
    endwhere
  end function adv

  subroutine flagcb
    ! Flags which zone are convective boundaries (CB),
    ! interpolate their position and link them to CBs
    ! of the previous time step to get their motion.
    use constants,only:Msun,csecyr,pi4
    use maths,only:squadric
    use parameters,only:DOSAVE
    implicit none
    integer::i,im,i0,ip,jm
    integer::k,dk
    logical::outofrange,goon
    double precision::xmc,gp,g0,g1
    double precision::rp,rm
    double precision::target,dmt0,oldmt0
    if (verbose.ge.3) print*,'   flagcb'

    ! Convective state
    cv(1:nz)=gradc.gt.0.
    cv(nz+1)=cv(nz)

    ! Loop on zones to find new convective boundaries
    icb=0
    ncb=0
    isolated=.false.
    do i=2,nz-1
       if ((cv(i-1).and.cv(i).and..not.cv(i+1))&
            .or.(.not.cv(i).and.cv(i+1).and.cv(i+2)))&
            then
          ncb=ncb+1
          cb(ncb)%iz=i
          if (cv(i)) then
             ! Bottom of a convective region
             cb(ncb)%top=.false.
             ip=i
             i0=i+1
             im=min(i+2,nz)
             jm=i+1
             cb(ncb)%ic=i
          else
             ! Top of a convective region
             cb(ncb)%top=.true.
             ip=i+1
             i0=i
             im=i-1
             jm=i-1
             cb(ncb)%ic=i+1
          endif
          cb(ncb)%ir=jm

          ! Gradient of the convective criterium
          gp=(gradc(i0)-gradc(im))*0.5*dm(i)/dm(jm)
          if (gp.le.0.) gp=0.5*(gradc(ip)-gradc(i0))
          ! Relative position of the convective boundary in the zone
          xmc=-gradc(i0)/gp          
          ! Prevent xmc from going too far out of the zone
          if (xmc.gt.1.) xmc=1.+(xmc-1.)/(1.+sqrt(xmc-1.))**2 
          xmc=0.5*xmc

          ! Weights for function wa()
          if (cb(ncb)%top) then
             cb(ncb)%xm=1.-xmc
          else
             cb(ncb)%xm=xmc
          endif
          cb(ncb)%xp=1.-cb(ncb)%xm
          ! Convective mass within zone
          cb(ncb)%dm=(1.-xmc)*dm(i)
          if (cb(ncb)%dm.lt.0.) then
             print*,'ncb',ncb,' xmc=',xmc
             print*,'Negative dm convective !!'
             stop
          endif
          ! Mass coordinate of convective boundary
          if (cb(ncb)%top) then
             cb(ncb)%m=m(cb(ncb)%ic)+cb(ncb)%dm
          else
             cb(ncb)%m=m(cb(ncb)%ic)-cb(ncb)%dm
          endif
          ! Pointer to the boundary
          icb(i)=ncb
       endif
    enddo

    ! Link to boundaries at former time step
    cb%iold=0
    do i=1,ncbold
       ! Take a former convective boundary
       k=cbold(i)%ic
       ! First, go within the old convective region until we meet 
       ! the new convective region
       if (cbold(i)%top) then
          dk=1
          if (.not. noflagcb.and.verbose.ge.-6)print*,'Old top'
       else
          dk=-1
          if (.not. noflagcb.and.verbose.ge.-6)print*,'Old bottom'
       endif
       outofrange=.false.
       if (.not.cvold(k))  print*,'Ha non !'
       do while(cvold(k).and..not.cv(k).and..not.outofrange)
          if (.not. noflagcb.and.verbose.ge.-6)print*,'Going within:',k,'->',k+dk
          k=k+dk
          outofrange=(k.lt.1.or.k.gt.nz+1)
       enddo
       if (outofrange.and.verbose.ge.-6) print*,'out of range 1'
       if (.not.outofrange.and.cv(k)) then
          ! Then go back within the new convective region until 
          ! the edge is reached
          do while(cv(k).and..not.outofrange)
             if (.not. noflagcb.and.verbose.ge.-6)print*,'Going out',k,'->',k-dk
             k=k-dk
             outofrange=(k.lt.1.or.k.gt.nz+1)
          enddo
          if (outofrange.and.verbose.ge.-6) print*,'out of range 2'
          if (.not.outofrange) then
             ! If this is a clean convective boundary, link it
             if (k.gt.1) then
                if (icb(k-1).gt.0..and..not.cbold(i)%top)then
                   cb(icb(k-1))%iold=i
                   if (.not. noflagcb.and.verbose.ge.-6)print*,'Bottom'
                endif
             endif
             if (k.lt.nz+1) then
                if (icb(k).gt.0..and.cbold(i)%top) then
                   cb(icb(k))%iold=i
                   if (.not. noflagcb.and.verbose.ge.-6)print*,'Top'
                endif
             endif
          endif
       endif
    enddo

    ! Compute boundaries speed
    cb%mt=0.
    cb%mt0=0.
    cb%m0=cb%m
    cb%advancing=.true.
    do i=1,ncb
       if (cb(i)%iold.gt.0) then 
          ! Velocity of the real CB (can be spurious if bad convergence)
          cb(i)%mt=invdt*(cb(i)%m-cbold(cb(i)%iold)%m)
          ! Smoothed velocity
          oldmt0=cbold(cb(i)%iold)%mt0
          if (oldmt0==0.) then
             cb(i)%mt0=cb(i)%mt
          else
             target=invdt*(cb(i)%m-cbold(cb(i)%iold)%m0)
! Work in progress...
!!$             dmt0=MAX(0.2,4.*13.*cbold(cb(i)%iold)%dlnt)*&
!!$                  (abs(oldmt0)+0.2*abs(target-oldmt0))
             dmt0=Msun*invdt*(20.*cbold(cb(i)%iold)%dlnt)**2+0.1*abs(target-oldmt0)
             if ((abs(target-oldmt0)+dmt0).ne.0.) &
                  cb(i)%mt0=oldmt0+dmt0*(target-oldmt0)/(abs(target-oldmt0)+dmt0)
!             cb(i)%mt0=oldmt0+4.*log(10.)/log(2.)*cbold(cb(i)%iold)%dlnt&
!                  *(target-oldmt0)
             if (verbose.ge.-6.and..not.noflagcb.and.i.ge.2) then
                print*,'-----'
                print*,'target=',target/(Msun/csecyr)
                print*,'oldmt0=',oldmt0/(Msun/csecyr)
                print*,'m=',cb(i)%m/Msun
                print*,'m0old=',cbold(cb(i)%iold)%m0/Msun
                PRINT*,'DLNT=',cbold(cb(i)%iold)%dlnt
                print*,'dmt0=',dmt0/(Msun/csecyr)
                print*,'newmt0=',cb(i)%mt0/(Msun/csecyr)
                print*,'-----'
             endif
          endif
          ! New position of the virtual CB 
          cb(i)%m0=max(0.,cbold(cb(i)%iold)%m0+dt*cb(i)%mt0)
          ! Receding region has no flux at boundary
          if ((cb(i)%top      .and.cb(i)%mt0.lt.0.).or.&
               (.not.cb(i)%top.and.cb(i)%mt0.gt.0.)) then 
             cb(i)%advancing=.false.
          endif
       endif
    enddo
    if (.not.noflagcb) then
    do i=1,ncb     
       if (verbose.ge.-6) then
          print*,i,&
               'mc=',cb(i)%m/Msun,&
               'mc0=',cb(i)%m0/Msun,&
               ' dmc=',cb(i)%dm/dm(cb(i)%iz),&
               ' mt=',cb(i)%mt/Msun*csecyr,&
               ' mt0=',cb(i)%mt0/Msun*csecyr,&
               ' icb=',cb(i)%iz
          print*,'   xm',cb(i)%xm,'xp',cb(i)%xp
          print*,'   iold',cb(i)%iold
          if (cb(i)%iold.gt.0) print*,' izold',cbold(cb(i)%iold)%iz
          print*,'top',cb(i)%top
       endif
    enddo
    endif
  end subroutine flagcb

  subroutine convection
    ! Computes convective gradients anf fluxes
    use dimensions,only:dtold,dt
    use secondaries,only:p,cP,T,delta,rni,ch,grada,na
    use chemistry,only:am,specie,ispecie,bn,iprim,C12,ircc,iz
    use constants,only:csecyr,Msun,pi4,cl,cg,Rsun
    use nuclear,only:rate
    use parameters,only:workc,alpham,rcd,UDIFF,UCHEM,MIXING,&
         USEDGold,MIXSCHEME,SEMICONV,&
         imuidn,imuphi
    use maths,only:lcubic,cubic,argsh,discr
    implicit none
    double precision,dimension(nz)::vf,wfudge
    double precision,parameter::eps=1.d-30
    double precision,dimension(nz)::alphaP,alphapp
    !MIXIN
    double precision,dimension(nz)::psrho,aa,bb,cc
    integer i,j
    ! convective thermal diffusion parameter :
    double precision,parameter::betaMLT=4.5!,chid=7.
    ! fudge coeff
    double precision,dimension(nz)::coeff,dlnp1
    ! Convective stuff centred in cells
    double precision,dimension(nz)::u1c
    double precision,dimension(nz)::lgtemp
    double precision,dimension(nz)::sg
    double precision,dimension(nz)::surf2
    double precision,dimension(nz)::pnum
    double precision,dimension(nz)::phi,mu,muphi
    double precision,dimension(nz)::aconv,bconv
    double precision,dimension(nz)::gradcs,gradss
    
!    double precision,dimension(2*nz)::dg2,lg2,dlnp2,denergy2,dlnT2,vc2,fac2
    if (verbose.ge.2) print*,'  convection'

    psrho=p/rho

    ! Mixing length
    HP = Psrho*wp(1.d0/grav)
    Lm=alpham*Hp!min(wp(r)*1.d11,alpham*HP)
    alphap=lm/(2.*Hp)
    alphapp=.5*alpham

    ! Diffusive velocities
    u0=2.*ch/lm
    ud=2.*chid/lm

    ud=0.!UD!

    ! Prepare variables for convective criterion
    gl=L
    galpha=psrho/(rho*cP*T*ch*pi4*CG)
    gbeta=-grada/galpha
    gphi=0.
    gamman=0.
    do i=1,ndim
       gphi(:,i)=-rni(:,i)/delta/galpha
       gamman(:,i)=Ftot(:,i)/bn(i)*m
    enddo

    ! Convective criterion
    ! Radiative gradient
    gradr=gl/m*wm(galpha)
    ! Schwarzshild criterion
    grads=gradr-wm(grada)
    ! Chemical part of the Ledoux criterion
    gradl=0.
    do i=1,ndim
       if (UCHEM.eq.'yes')gradl=gradl+wm(gphi(:,i))*gammaN(:,i)
    enddo
    gradl=gradl*wm(galpha)/m
    ! Ledoux criterion
    gradc=grads+gradl

    ! Convective boundaries
    ! Remove galpha coeff
    galpha=1.d0
    ! set convective boundaries
     if (.not.noflagcb) then
    call flagcb
     endif
    !Recover gammaN
    do i=1,ndim
       gamman(:,i)=Ftot(:,i)/bn(i)
    enddo    

    ! Effective sound speed 
    u1=alphap*sqrt(psrho)

    ! Convective velocity
    vc=wm(u1)*cbrt(wm(delta*u0/u1)*MAX(0.,gradc))&
         *1.d0!*max(0.,gradc)/(1.d0+max(0.,gradc))
    if (SEMICONV.eq.'yes') then
       ! Semi-convection
       where(gradc.lt.0. .and. grads.gt.0.)
          vc=wm(u0)*gradl/gradc
       endwhere
!!$    ! Salt fingers
!!$       where(gradc.lt.0. .and. gradl.gt.0.)
!!$          vc=-wm(u0)*gradl/gradc
!!$       endwhere
    endif

    ! Growth rate
    tconv=0.
    where (vc.gt.0.) 
       tconv=lm/vc
    endwhere
    ! Semi-convection
    where(gradc.lt.0. .and. grads.gt.0.)
       tconv=lm/(wm(u0)*gradl/gradc)
    endwhere
!!$    ! Salt fingers
!!$    where(gradc.lt.0. .and. gradl.gt.0.)
!!$       tconv=lm/(-wm(u0)*gradl/gradc)
!!$    endwhere
    where(tconv.gt.0.) 
       aconv=exp(-dt/tconv)
    elsewhere
       aconv=0.
    endwhere
    aconv=0.
    bconv=1.d0-aconv
   
    ! Pressure
    dxlnp=pm*wm(1.d0/p)
    dlnp=dmt*dxlnp

    ! u0 on interfaces
    u0m=1./wm(1./u0)
    ! (harmonic mean is better behaved for convective boundaries)

    where(isolated) vc=0.
    factor1=1.d0/(vc/u0m+1.d0)
    factor2=vc/(vc+u0m)

    ! Chemistry
    gradeN=0.
    do i=1,ndim
       mu=wm((mui(:,i)-.5*bn(i)*mue)/(cp*T))
       muphi=imuphi*imuidn*wm(delta*(mui(:,i)-.5*bn(i)*mue)/(cp*T*rni(:,i)))+1.d0
       gradN(:,i)=1.d0/(1.d0+muphi*vc/u0m)*gammaN(:,i)  !SC!     
       where(isolated) gradN(:,i)=gammaN(:,i)
       gradeN=gradeN+imuidn*mu*gradN(:,i)
    enddo

    ! Temperature
    gradTa=factor1*grads
    if (UCHEM.eq.'yes') gradTa=gradTa-factor2*gradeN
    ! Energy
    grade=gradTa+ gradeN

!!$    ! Time-dependent gradients
!!$    gradTa=(aconv*gradTaold+bconv*gradTa/gradr)*gradr
!!$    grade =(aconv*gradeold +bconv*grade/gradr)*gradr

    ! Chemistry
    do j=1,species
       i=ispecie(j)
       muphi=imuidn*imuphi*wm(delta*(mui(:,i)-.5*bn(i)*mue)/(cp*T*rni(:,i)))+1.d0
       dxN(:,j)=gradN(:,i)*dxlnp
       dN(:,j)=dxN(:,j)*dmt
       Fconv(1:nz,j)=1.d-11*alphapp*surf*wm(rho)*u0m*&
               vc/(muphi*vc+u0m)*&
               gfrac(:,j)
       where (isolated) Fconv(1:nz,j)=0.
    enddo
    Fconv(nz+1,:)=0.

    ! Temperature
    gradt=wm(grada)+gradTa
    dxlnT=(wm(grada)+gradTa)*dxlnp
    dlnT=dxlnT*dmt

    ! Energy
    dxenergy=wm(cp*T)*grade*dxlnp

!!$    ! Convective work (To be updated, and recompute dlnrho, dlnt, dlnf..)
    Wconv=0.
    if (Workc.eq.'yes') then
       do j=species-1,species
          Wconv=Wconv+wp(gradm(sni(:,ispecie(j)))*Fconv(1:nz,j)/bn(ispecie(j)))
       enddo
    endif

    ! Luminosities
    Ltot(1:nz)=L
    Ltot(nz+1)=0.

    ! Convection parameter (for grid and time-step controls)
    conv=1.d0!d-2*argsh(gradc)!1.!*.1*argsh(1.e2*(dg-lg))

    ! Resolution function (STAR like)
    QM = VPP*dxlnp + VTT*dxlnt + VMM + VRR*2./(pi4*wm(rho)*r) 

    if (USEDGOLD=='yes') then       
       lg=lgtemp
    endif

    if (verbose.ge.2) print*,'  end convection'
  end subroutine convection

  ! Many different ways of weighting a function across 2 zones or interfaces

  function wt(f) Result(fw)
    implicit none
    double precision,dimension(nz)::f
    double precision,dimension(2:nz)::fw
    fw=(dm(1:nz-1)*f(1:nz-1)+dm(2:nz)*f(2:nz))/(dm(1:nz-1)+dm(2:nz))
  end function wt

  function wtp(f) Result(fw)
    implicit none
    double precision,dimension(nz)::f
    double precision,dimension(nz)::fw
    fw(1)=f(1)
    fw(2:nz)=.5*(f(2:nz)+f(1:nz-1))
!!    fw(nz)=f(nz)
  end function wtp

  function wtm(f) Result(fw)
    implicit none
    double precision,dimension(nz)::f
    double precision,dimension(nz)::fw
    fw(1:nz-1)=(dm(1:nz-1)*f(1:nz-1)+dm(2:nz)*f(2:nz))/(dm(1:nz-1)+dm(2:nz))
    fw(nz)=f(nz)
  end function wtm

  function wp(f) Result(fc)
    double precision,dimension(nz)::f,fc
    if (STAGGERED=='yes') then
       fc(1:nz-1)=.5*(f(1:nz-1)+f(2:nz))
!!$    fc(1:nz-1)=sqrt(f(1:nz-1)*f(2:nz))
       fc(nz)=f(nz)
    else
       fc=f
    endif
  end function wp

  function wm(f) Result(fc)
    double precision,dimension(nz)::f,fc
    fc(2:nz)=(dm(2:nz)*f(1:nz-1)+dm(1:nz-1)*f(2:nz))&
         /(dm(1:nz-1)+dm(2:nz))
    fc(1)=f(1)-0*(f(2)-f(1))*dm(1)/(dm(1)+dm(2))
  end function wm
    
  function wa(f) Result(fc)
    ! Partially convective weighting
    use parameters,only:PARTCONV
    double precision,dimension(nz)::f
    double precision,dimension(nz)::fc
    double precision::dmc,xmc
    double precision::fm,fp
    integer::i,j
    fc=f
    if (PARTCONV.ne.'yes') return
    do i=1,ncb
       j=cb(i)%iz
       fc(j)=cb(i)%xm*f(j-1)+cb(i)%xp*f(j+1)
    enddo
  end function wa
  
  function wd(f) Result(fc)
    double precision,dimension(nz)::f
    double precision,dimension(nz)::fc
    double precision::xm,xp
    integer::i,j
    fc=f*dmt
    do i=1,ncb
       j=cb(i)%iz
       xm=cb(i)%xm
       if (xm.le.0.5) then         
          fc(j)=  (0.5+xm)*f(j)*0.5*dm(j-1)+(0.5-xm)*f(j+1)*0.5*dm(j)
       else
          xp=cb(i)%xp
          fc(j+1)=(0.5-xp)*f(j)*0.5*dm(j)  +(0.5+xp)*f(j+1)*0.5*dm(j+1)
       endif
    enddo
  end function wd

  function wc(f) Result(fc)
    double precision,dimension(nz)::f,fc
    if (STAGGERED=='yes') then
       fc=f
    else
       fc(1:nz-1)=.5*(f(1:nz-1)+f(2:nz))
       fc(nz)=1.5*f(nz)-.5*f(nz-1)
    endif
  end function wc

  function wcc(f) Result(fc)
    double precision,dimension(nz)::f,fc

    fc(1:nz-1)=.5*(f(1:nz-1)+f(2:nz))
    fc(nz)=1.5*f(nz)-.5*f(nz-1)
  end function wcc

  
  ! Gradient
  function grad(f) Result(df)
    ! Computes gradient of an interface based function
    double precision,dimension(nz)::f
    double precision,dimension(1:nz)::df
    if (STAGGERED=='yes') then
       df(1:nz-1)=(f(1:nz-1)-f(2:nz))/dm(1:nz-1)
       df(nz)=df(nz-1)
    else
       df=gradm(f)
    endif
  end function grad

  function gradm(f) Result(df)
    ! Computes gradient of centre of zone based function
    double precision,dimension(nz)::f
    double precision,dimension(nz)::df
    df(2:nz)=(f(1:nz-1)-f(2:nz))*2./(dm(1:nz-1)+dm(2:nz))
    df(1)=0.
  end function gradm

  ! Full theory of convection (not used, to be updated)
  function solvetot(c1,c2,c3,a0,b0,ad) Result(x)
    use maths,only:lcbrt
    implicit none
    double precision,dimension(nz)::c1,c2,c3,a0,b0,ad
    double precision,dimension(nz)::a,b,c
    double precision,dimension(nz)::x
    double precision::f,fx,dx
    double precision,dimension(nz),save::xs=0.
    double precision,parameter::tol=1.d-15
    integer::i,j
    if (Verbose.ge.3) print*,'   solvetot'

    do i=1,nz
       if (c1(i)+c3(i).gt.0.) then
          x(i)=0.
       else
          ! First guess
          if (xs(i).ne.0.) then
             x(i)=xs(i)
          else
             if (c1(i).lt.0.) then
                x(i)=lcbrt(-c1(i)/b0(i))
             else
                x(i)=sqrt(-c3(i)-c1(i))
             endif
          endif
          dx=2.*tol*x(i)
          j=0
          do while(dx.gt.tol*x(i))           
             CALL fdf(c1(i),c2(i),c3(i),a0(i),b0(i),ad(i),x(i),f,fx)
             dx=-f/fx
             if (x(i)+dx.lt.0.) dx=-.5*x(i)
             x(i)=x(i)+dx
             j=j+1
             if (verbose.ge.5) print*,'j',j,' x=',x(i),' dx/x=',dx/x(i)
          enddo
          if (verbose.ge.4) print*,'jTOT=',j
       endif
    enddo
    xs=x
  end function solvetot

  subroutine fdf(c1,c2,c3,a0,b0,ad,x,f,fx)
    implicit none
    double precision::c1,c2,c3,a0,b0,ad,x
    double precision::f,fx
    double precision::a,b,c,ax,bx,cx
    if (Verbose.ge.4) print*,'    fdf'

    if (x.eq.0.) then
       f=c3
       fx=c1/a0-ad*c3
    else
       c=1.d0/(1.d0+ad*x)
       cx=-ad*c*c
       b=1.d0+a0/x+b0*x
       bx=-a0/(x*x)+b0
       a=c1+c2*b0*x*c
       ax=c2*b0*(c+x*cx)
       c=c3*c
       cx=c3*cx
       f=x*x+a/b+c
       fx=2.*x+ax/b-a*bx/(b*b)+cx
    endif
  end subroutine fdf
end MODULE equations

