MODULE maths
  use dimensions,only:verbose
CONTAINS


  SUBROUTINE NUMREC_SOLVE(A,X,B,n,INFO)
    implicit none
    ! Numrec, p.40
    integer::n
    double precision,dimension(n,n)::a
    double precision,dimension(n)::b,x
    integer,dimension(n)::indx
    integer::info
    double precision::d
    
    call ludcmp(a,n,n,indx,d)
    x=b
    call lubksb(a,n,n,indx,x)
  end SUBROUTINE NUMREC_SOLVE

  SUBROUTINE ludcmp(a,n,np,indx,d)
    integer n,np,indx(n)
    double precision::d
    double precision,dimension(np,np)::a
    double precision,parameter::tiny=1.d-60
    integer i,imax,j,k
    double precision::aamax,dum,sum
    double precision,dimension(n)::vv

    d=1.
    do i=1,n
       aamax=0.
       do j=1,n
          if (abs(a(i,j)).gt.aamax) aamax=abs(a(i,j))
       enddo
       if (aamax.eq.0.) pause 'singular matrix in ludcmp'
       vv(i)=1./aamax
    enddo

    do j=1,n
       do i=1,j-1
          sum=a(i,j)
          do k=1,i-1
             sum=sum-a(i,k)*a(k,j)
          enddo
          a(i,j)=sum
       enddo
       aamax=0.
       do i=j,n
          sum=a(i,j)
          do k=1,j-1
             sum=sum-a(i,k)*a(k,j)
          enddo
          a(i,j)=sum
          dum=vv(i)*abs(sum)
          if (dum.ge.aamax) then
             imax=i
             aamax=dum
          endif
       enddo
       if (j.ne.imax) then
          do k=1,n
             dum=a(imax,k)
             a(imax,k)=a(j,k)
             a(j,k)=dum
          enddo
          d=-d
          vv(imax)=vv(j)
       endif
       indx(j)=imax
       if (a(j,j).eq.0.) a(j,j)=tiny
       if (j.ne.n) then
          dum=1./a(j,j)
          do i=j+1,n
             a(i,j)=a(i,j)*dum
          enddo
       endif
    enddo
    return
  end SUBROUTINE ludcmp

  SUBROUTINE lubksb(a,n,np,indx,b)
    integer::n,np
    integer,dimension(n)::indx
    double precision,dimension(np,np)::a
    double precision,dimension(np)::b
    integer i,ii,j,ll
    double precision sum

    ii=0
    do i=1,n
       ll=indx(i)
       sum=b(ll)
       b(ll)=b(i)
       if (ii.ne.0) then
          do j=ii,i-1
             sum=sum-a(i,j)*b(j)
          enddo
       else if (sum.ne.0.) then
          ii=i
       endif
       b(i)=sum
    enddo
    do i=n,1,-1
       sum=b(i)
       do j=i+1,n
          sum=sum-a(i,j)*b(j)
       enddo
       b(i)=sum/a(i,i)
    enddo
    return
  end SUBROUTINE lubksb

  ! =========================================
  SUBROUTINE SQUARE_SOLVE(A,XX,B,n,INFO,invert)
    ! =======================================
    implicit none
    integer::n
    logical::invert
    character(1)                :: FACT,TRANS,EQUED
    integer                     :: nrhs,ldab,ldafb,ldb,ldx,lda
    double precision,dimension(n,n)  :: A
    double precision,allocatable,save,dimension(:,:):: AFB
    integer,save,allocatable,dimension(:)       :: IPIV
    double precision,save,allocatable,dimension(:)       :: Rs,Cs
    double precision,dimension(n)::XX,B
    double precision                        :: Rcond
    double precision,dimension(n)::FERR,BERR
    double precision,    dimension(4*n)     :: WORK
    integer, dimension(n)     :: IWORK
    integer                     :: INFO

    nrhs=1
    lda=n
    ldafb=n
    ldb=n
    ldx=n
    if (invert) then
       if (allocated(Afb)) then
          deallocate(afb,ipiv,rs,cs)
       endif
       allocate(AFB(n,n),IPIV(n),Rs(n),Cs(n))
       FACT = 'E'
    else
       FACT='F'
    endif
    TRANS= 'N'
!!$    call DGESVX( FACT, TRANS, n, NRHS, A, LDA, AFB, LDAFB, &
!!$         IPIV, EQUED, Rs, Cs, B, LDB, XX, LDX, RCOND, FERR, BERR, &
!!$         WORK, IWORK, INFO )
    if (INFO.ne.0) then
       print*,'Solve matrix INFO=',INFO
       print*,'EQUED=',EQUED
    endif

  end SUBROUTINE 

  SUBROUTINE SOLVE_sparse(A,x,b,n,ifail)
    use dimensions,only:m1,m2,ndiag
    implicit none
    integer::n


    double precision,dimension(ndiag,n):: A
    double precision,dimension(n)::b,x
    double precision, allocatable,dimension(:)::Asparse,work
    integer, allocatable, dimension(:)::irow,icol
    integer::nnz,m,maxitn,itn,lwork
    integer,dimension(2*n+1)::iwork
    character*8::METHOD
    character*1::PRECON
    double precision::omega,tol,rnorm
    integer::ifail
    double precision::coeff

    integer::i,j,k,mm

!    x=b/A(1+m1,:)
    ! Set up the sparse matrix format
    ! Number of non-zero elements
    k=0
    do i=1,n
       do mm=1,ndiag
          j=i-(1+m1)+mm
!          j=m1+1+i-mm
          if (j.ge.1.and.j.le.n) then
             if (A(mm,j).ne.0.)  k=k+1
          endif
       enddo
    enddo
    print*,'Non-zero elts k=',k
!    print*,' other=',SUM(1,mask=A.ne.0.,DIM=2)
    allocate(Asparse(k),irow(k),icol(k))
    nnz=k
    k=0
    do i=1,n
       do mm=1,ndiag!,1,-1
          j=i-(1+m1)+mm
!          j=m1+1+i-mm
          if (j.ge.1.and.j.le.n) then
             if (A(mm,j).ne.0.) then
                k=k+1
                irow(k)=i
                icol(k)=j
                Asparse(k)=A(mm,j)
             endif
          endif
       enddo
    enddo

!    coeff=max(maxval(abs(asparse)),maxval(abs(b)))
!    asparse=asparse/coeff
!    b=b/coeff

    ! Set up parameters
    method='BICGSTAB'!!'RGMRES' ! 'RGMRES' ! 'CGS' ,
    precon='J' ! 'N' 'J' 'S'
    omega=1.5
    tol=0.!1.d-12!1.e-14
    maxitn=100

    ! Allocate work memory
    if (method=='RGMRES') then
       M=25
       lwork=2*(4*n+m*(m+n+4)+n+1)
    endif
    if (method=='BICGSTAB') then
      m=5
      lwork=2*(2*n*(m+2)+m*(m+2)+n+2*n)
    endif
    if (method=='CGS') then
       lwork=2*(8*n+n)
    endif
    allocate(work(lwork))

    ifail=0
    x=1.
!    CALL F11def(method,precon,n,nnz,Asparse,irow,icol,omega,b,m,tol,&
!         maxitn,x,rnorm,itn,work,lwork,iwork,ifail)
    print*,itn,' iterations rnorm=',rnorm
    deallocate(Asparse,irow,icol,work)
 !    b=b*coeff
  end SUBROUTINE SOLVE_SPARSE
  

  SUBROUTINE trim_band(m1t,m2t,A)
    use dimensions,only:neq,m1,m2
    implicit none
    double precision,dimension(m1+1+m2,neq)::A
    integer::m1t,m2t
    integer::mm,i,j
    m2t=m2
    do while(all(A(m2+1-m2t,:)==0.))
       m2t=m2t-1
    enddo
    m1t=m1
    do while(all(A(m2+1+m1t,:)==0.))
       m1t=m1t-1
    enddo
  end SUBROUTINE trim_band

  ! =========================================
  SUBROUTINE SOLVE(A,XX,B,neq,INFO,invert)
    ! =========================================
    use dimensions,only:m1,m2
    implicit none
    integer::neq
    logical::invert
    character(1)                :: FACT,TRANS,EQUED
    integer,save                     :: nrhs,ldab,ldafb,ldb,ldx,lda
    double precision,dimension(m1+1+m2,neq)  :: A,A0
    double precision,save,allocatable,dimension(:,:)::afb,At
    integer,save,allocatable,dimension(:)       :: IPIV
    double precision,save,allocatable,dimension(:)       :: Rs,Cs
    double precision,dimension(neq)::XX,B,B0
    double precision                        :: Rcond,FERR,BERR
    double precision,    dimension(3*neq)     :: WORK
    integer, dimension(3*neq)     :: IWORK
    integer                     :: INFO
    double precision,dimension(neq)::SA
    integer i,j,mm
    integer,save::m1t,m2t
    integer::m1ts,m2ts

    if (invert) then
       if (allocated(AT))deallocate(At,afb,rs,cs,ipiv)
       ! Trims the band matrix from zero diagonals (accelerates inversion).
       m1ts=m1t
       m2ts=m2t
       call trim_band(m1t,m2t,A)
       if (verbose.ge.-3) then
          if (m1t.ne.m1ts.or.m2ts.ne.m2t) then
             print*,'Trimming changing !'
             print*,'m1t=',m1t,' m2t=',m2t
             print*,'m1ts=',m1ts,' m2ts=',m2ts
          endif
       endif
       allocate(At(m1t+m2t+1,neq),rs(neq),cs(neq),ipiv(neq))
       do i=1,neq
          At(1:m2t+1,i)=A(m2+1-m2t:m2+1,i)
          At(2+m2t:1+m2t+m1t,i)=A(m2+2:m2+1+m1t,i)
       enddo
       if (verbose.ge.-1) print*,'m1+m2+1=',m1+m2+1,' m1t+m2t+1=',m1t+m2t+1

       nrhs=1
       lda=m1t+1+m2t
       ldafb=2*m1t+m2t+1
       allocate(afb(ldafb,neq))
       ldb=neq
       ldx=neq
       FACT = 'E'
!!$       !  conditionnement de A:
!!$       do i = 1,neq
!!$          SA(i)  = 0.
!!$          do mm = 1,lda
!!$             j=i-(m1+1)+mm
!!$             if (j.ge.1.and.j.le.neq) SA(i) = SA(i) + abs(A(lda+1-mm,j))
!!$          end do
!!$          if (SA(i).eq.0.)SA(i) = 1.
!!$          do mm = 1,lda
!!$             j=i-(m1+1)+mm
!!$             if (j.ge.1.and.j.le.neq) A(lda+1-mm,j)=A(lda+1-mm,j)/SA(i)
!!$          end do
!!$       end do
!!$       ! conditionnement  de B:
!!$       B=B/SA
    else
       FACT='F'
    endif
    TRANS= 'N'
!!$    call DGBSVX( FACT, TRANS, neq, m1t, m2t, NRHS, At, LDA, AFB, LDAFB, &
!!$         IPIV, EQUED, Rs, Cs, B, LDB, XX, LDX, RCOND, FERR, BERR, &
!!$         WORK, IWORK, INFO )
!!$  if (INFO.ne.0) then
!!$     print*,'Solve matrix INFO=',INFO
!!$     print*,'EQUED=',EQUED
!!$  endif
  end SUBROUTINE 

  ! ====================================================
  SUBROUTINE SOLVE_JPC(At,X,B0,n,m1,m2,A,A0,A_L,index,SA,xold,INFO,invert)
    ! ====================================================
    use dimensions,only:nameofvar,variable,zone
    logical::invert
    integer                     :: n,m1,m2,icmax
    double precision, dimension(m1+1+m2,n)  :: At
    double precision,    dimension(n)       :: X,B0,B,xold,x0
    integer                     :: INFO
    double precision::A  (1:n,1:m1+m2+1)
    double precision::A0 (1:n,1:m1+m2+1)
    double precision::A_L(1:n,1:m1)
    integer,dimension(n)::index
    double precision,dimension(n)::SA

    integer ndiag,idiag
    integer np,mp,mpl

    integer jj,i,j,kk
    double precision erreur
    double precision,parameter::precision=1.e-16

    double precision,dimension(n)::v0,v,v1
    integer,parameter::nlam=50
    double precision,dimension(n)::lambda

    double precision,dimension(n)::dxold
    if (verbose.ge.2) print*,'  SOLVE_JPC'

    dxold=0.

    ndiag=m1+m2+1
    idiag=m1+1
    np=n
    mp=m1+m2+1
    mpl=m1
    

    if (invert) then
       !  transpose la matrice
       A=0.
       do i = 1,n
          do mm  = 1,ndiag
             j=i-idiag+mm
             if (j.ge.1.and.j.le.n) A(i,mm) = At(ndiag+1-mm,j)
          end do
       end do

       !  conditionnement et stockage de A
       do i = 1,n
          SA(i)  = 0.
!!$          do jj = 1,ndiag
!!$             SA(i) = max(SA(i) , abs(A(i,jj)))
!!$          end do
!!$          SA(I)=max(SA(i),1.)
          do jj = 1,ndiag
             SA(i) = SA(i) + abs(A(i,jj))
          end do
          if (SA(i).eq.0.) SA(i) = 1.
          do jj = 1,ndiag
             A(i,jj) = A(i,jj)/SA(i)
          end do
       end do
       A0=A
    endif

    ! conditionnement et stockage de B:
    B=B0
    B=B/SA
    X=B

    if (invert) then
       !  inversion du systme:
       ! --------------------------------------------------
       CALL LU_decomposition(A,n,m1,m2,np,mp,A_L,mpl,index)
       ! --------------------------------------------------
    endif

   ! rsolution:
    ! ----------------------------------------------------
    CALL LU_substitution (A,n,m1,m2,np,mp,A_L,mpl,index,X)
    ! ----------------------------------------------------

!!$    erreur=1.
!!$    do while(erreur.ne.0.)
!!$       read(*,*)erreur
!!$       a0(:,m1+1)=a0(:,m1+1)+erreur
!!$       CALL bspectrum(a0,n,m1,m2,lambda)
!!$       a0(:,m1+1)=a0(:,m1+1)-erreur
!!$    enddo

!!    CALL check(a0,a,a_L,n,m1,m2)

    !  amlioration de la solution
    kk = 0
    erreur = 1.
    
!!$     ! Before calling iteration, estimate its convergence :
!!$    do i=1,nlam
!!$       CALL RANDOM_NUMBER(v)
!!$       if (i.le.2) v=x
!!$       if (i.eq.2) v=v/norm2(v,xold)
!!$       v0=v
!!$       CALL bande_multiplication(a0,n,m1,m2,np,mp,v0,v1)
!!$       CALL LU_substitution(a,n,m1,m2,np,mp,a_l,mpl,index,v1)
!!$       v1=v-v1
!!$!       lambda(i)=norm2(v1,xold)
!!$       lambda(i)=norm2(v1,xold)/norm2(v,xold)
!!$    enddo
!!$    print*,'Lam(x)=',lambda(1),'Maxlam=',MAXVAL(lambda(1:nlam)),' Norm(x)',norm2(x,xold)


    do WHILE (erreur.gt.precision.and.kk.lt.100.and.erreur.lt.1.d100)
       kk = kk+1
       !    ------------------------------------------------------------------
       CALL ITERATION (n,m1,m2,np,mp,mpl,A0,A,A_L,index,B,X,erreur,xold,icmax,dxold)
       !    ------------------------------------------------------------------
      if (MOD(kk,10).eq.0) then
         print 2,kk,nameofvar(variable(icmax)),zone(icmax)&
              ,erreur,xold(icmax)

!!$         print*,'Min(spA)=',MINVAL(ABS(A(:,1))),&
!!$              'Cond(A)=',MINVAL(ABS(A(:,1)))/MAXVAL(ABS(A(:,1)))
    
      endif

   end do
!!$   print 2,kk,nameofvar(variable(icmax)),zone(icmax)&
!!$        ,erreur,xold(icmax)
2  format(i5,' iterations LU : ',a7,'(',i3,')=',1pe10.3,' xold=',1pe12.3)

!!$   read(*,*)
  end SUBROUTINE 

  SUBROUTINE check(a0,a,a_l,n,m1,m2)
    integer::n,m1,m2
    double precision,dimension(n,m1+1+m2)::a0,a
    double precision,dimension(n,m1)::a_l
    double precision,dimension(n,n)::As
    double precision,dimension(n)::s
    integer::i,j,k
    
    As=0.
    do i=1,n
    enddo
    CALL spectrum(As,n,s)
  end SUBROUTINE check
  
  SUBROUTINE bspectrum(A,n,m1,m2,s)
    integer n,m1,m2
    double precision,dimension(n,m1+m2+1)::A
    double precision,dimension(n)::s
    double precision,dimension(n,n)::As
    integer i,j,k,mm
    double precision u,vt
    integer::lwork
    double precision,allocatable,dimension(:)::work
    integer info

    lwork=10*n
    allocate(work(lwork))
    As=0.
    do i=1,n
       do  mm=1,m1+m2+1
          j=i-1+m1+mm
          if (j.ge.1.and.j.le.n) As(i,j)=A(i,mm)
       enddo
    enddo
!!$    CALL DGESVD('N','N',n,n,As,n,s,u,1,vt,1,work,lwork,info)
    if (info>0) then
       print*,'DGESVD did not converge :',info,' diagonals only did converge.'
    endif
    if (info<0) then
       print*,'DGESVD : argument ',-info,' has an illegal value.'
    endif
    deallocate (work)
    do i=n-50,n
       print*,'i=',i,' l=',s(i)
    enddo
    print*,'Maxlam=',MAXVAL(s)
    print*,'Minlam=',MINVAL(s)
    print*,'Minlam0=',MINVAL(ABS(s),MASK=(s.ne.0.))
    print*,'Cond=',MINVAL(ABS(s),MASK=(s.ne.0.))/MAXVAL(s)
    print*,'Nzero=',COUNT(s==0.)
  end SUBROUTINE bspectrum
  SUBROUTINE spectrum(As,n,s)
    integer n
    double precision,dimension(n,n)::As
    double precision,dimension(n)::s
    double precision u,vt
    integer::lwork
    double precision,allocatable,dimension(:)::work
    integer info

    lwork=10*n
    allocate(work(lwork))
!!$    CALL DGESVD('N','N',n,n,As,n,s,u,1,vt,1,work,lwork,info)
    if (info>0) then
       print*,'DGESVD did not converge :',info,' diagonals only did converge.'
    endif
    if (info<0) then
       print*,'DGESVD : argument ',-info,' has an illegal value.'
    endif
    deallocate (work)
!!$    do i=1,n
!!$       print*,'i=',i,' l=',s(i)
!!$    enddo
    print*,'Maxlam=',MAXVAL(s)
    print*,'Minlam=',MINVAL(s)
    print*,'Minlam0=',MINVAL(ABS(s),MASK=(s.ne.0.))
    print*,'Cond=',MINVAL(ABS(s),MASK=(s.ne.0.))/MAXVAL(s)
    print*,'Nzero=',COUNT(s==0.)
  end SUBROUTINE spectrum
  ! =========================================================================
  subroutine ITERATION (n,m1,m2,np,mp,mpl,A,a_U,a_L,index,B,x,erreur,xold,icmax,dxold)
    ! =========================================================================
    implicit none
    ! --- Les arguments de la routines :
    integer:: m1,m2,mp,mpl,n,np
    integer,dimension(n)::index
    double precision,dimension(np,mp) ::   A,a_U
    double precision,dimension(np,mpl)::a_L
    double precision,dimension(n)::B,x
    ! --- Les variables internes a la routine :
    integer:: i,icmax
    double precision::erreur
    double precision,dimension(n)::dx,dxold
    double precision,dimension(n)::xold,dxscaled
    ! Convergence
    double precision,dimension(n)::v0,v,v1
    integer,parameter::nlam=50
    double precision,dimension(n)::lambda

    if (ALL(dxold.eq.0.)) then
       ! first iteration
       ! On calcule le residu :
       call BANDE_MULTIPLICATION(A,np,m1,m2,np,mp,x,dx)
       dx=b-dx
       ! On deduit la correction
       call LU_SUBSTITUTION(a_U,n,m1,m2,np,mp,a_L,mpl,index,dx)
    else
       call BANDE_MULTIPLICATION(A,np,m1,m2,np,mp,dxold,dx)
       call LU_SUBSTITUTION(a_U,n,m1,m2,np,mp,a_L,mpl,index,dx)
       dx=dxold-dx
    endif
    ! Sauve la correction
    dxold=dx
    ! Applique la correction
    x=x+dx
    ! Test de convergence
    dxscaled=ABS(dx/xold)
    erreur=MAXVAL(dxscaled)
    icmax=MAXLOC(dxscaled,DIM=1)
!!$    print*,'erreur=',erreur
!!$    call BANDE_MULTIPLICATION(A,np,m1,m2,np,mp,x,dx)
!!$    dx=b-dx
!!$    print*,'|Ax-b|=',norm2(dx,xold)

!!$    ! Evaluates convergence of the process :
!!$    do i=1,nlam
!!$       CALL RANDOM_NUMBER(v)
!!$       if (i.le.2) v=dx
!!$       if (i.eq.2) v=v/norm2(dx,xold)
!!$       if (i.gt.2) v=v*norm2(dx,xold)/norm2(v,xold)
!!$       v0=v
!!$       CALL bande_multiplication(a,n,m1,m2,np,mp,v0,v1)
!!$       CALL LU_substitution(a_u,n,m1,m2,np,mp,a_l,mpl,index,v1)
!!$       v1=v-v1
!!$!       lambda(i)=norm2(v1,xold)
!!$       lambda(i)=norm2(v1,xold)/norm2(v,xold)
!!$    enddo
!!$    print*,'Lam(dx)=',lambda(1),'Maxlam=',MAXVAL(lambda(1:nlam)),' Norm(dx)',norm2(dx,xold)

    RETURN
  END subroutine 

  double precision FUNCTION norm2(f,xold)
    use dimensions,only:neq
    implicit none
    double precision,dimension(neq)::f,xold
    integer i
    norm2=MAXVAL(ABS(f/xold))
    return
    norm2=0.
    do i=1,neq
       norm2=norm2+f(i)*f(i)
    enddo
    norm2=sqrt(norm2)
  end FUNCTION norm2

  !     ====================================================
  SUBROUTINE bande_multiplication(a,n,m1,m2,np,mp,x,b)
    !     ====================================================
    ! --- Calcul du produit matriciel b = Ax ; A est une matrice bande-diagonale de
    !     dimension n x n ayant m1 elements sous-diagonaux et m2 elements super-diagonaux.
    !     En d'autres termes, A(i,j) = 0 si (j > i + m2) ou si (i > j + m1).
    !     Les vecteurs d'entree x et de sortie b sont stockes dans les tableaux x(1:n) et b(1:n).
    !     Le tableau a(1:n,1:m1+m2+1) contient la matrice A sous la forme suivante :
    !       1. les elements       diagonaux sont dans a(1:n,     m1+1   )
    !       2. les elements  sous-diagonaux sont dans a(j:n,   1:m1     ) avec j > 1
    !       3. les elements super-diagonaux sont dans a(1:j,m1+2:m1+m2+1) avec j < n
    !     np et mp sont les dimensions physiques du tableau a : np  n ; mp  m1 + 1 + m2.

    IMPLICIT NONE
    ! --- Les arguments de la routines :
    integer:: m1,m2,mp,n,np
    double precision,dimension(np,mp)::a
    double precision,dimension(n)::b,x
    ! --- Les variables internes a la routine :
    integer:: i,j,k

    DO i = 1,n
       b(i) = 0.
       k    = i - m1 - 1
       DO j = MAX(1,1 - k),MIN(m1 + m2 + 1,n - k)
          b(i) = b(i) + a(i,j)*x(j+k)
       END DO
    END DO
    RETURN
  END SUBROUTINE 

  !     =========================================================
  SUBROUTINE LU_decomposition(a,n,m1,m2,np,mp,al,mpl,index)
    !     =========================================================
    ! --- Soit A une matrice bande-diagonale de dimension n x n ayant m1 elements sous-diagonaux
    !     et m2 elements super-diagonaux, stockee sous forme compact dans le tableau a(1:n,1:m1+m2+1).
    !     Cette routine calcule la decomposition LU d'une permutation ligne a ligne de A.
    !     La matrice triangulaire superieure remplace a, la matrice triangulaire inferieure est stockee
    !     dans le tableau al(1:n,1:m1). Le tableau index(1:n) contient la permutation des lignes
    !     effectuee par pivot partiel.
    !     np, mp et mpl sont les dimensions physiques des tableaux a et al.
    !     On doit avoir np  n ; mp  m1 + 1 + m2 ; mpl  m1.
    !     Cette routine doit etre utilisee avec 'bande_LU_substitution' pour resoudre un systeme
    !     lineaire bande-diagonal.

    IMPLICIT NONE
    ! --- Les arguments de la routines :
    integer:: m1,m2,mp,mpl,n,np
    integer,dimension(n)::index
    double precision,dimension(np,mp)::a
    double precision,dimension(np,mpl)::al
    ! --- Les variables internes a la routine :
    integer:: i,j,k,l,mm
    double precision:: dum

    mm  =  m1 + m2 + 1
    IF ((mm .gt. mp).OR.(m1 .gt. mpl).OR.(n .gt. np))&
         STOP 'Probleme d''arguments dans la decomposition LU.'
    l  =  m1
    DO i = 1,m1
       DO j = m1+2-i,mm
          a(i,j-l) = a(i,j)
       END DO
       l = l - 1
       DO j = mm-l,mm
          a(i,j) = 0.
       END DO
    END DO
    l = m1
    DO k = 1,n
       dum = a(k,1)
       i = k
       IF(l .lt. n) l = l + 1
       DO j = k+1,l
          IF (ABS(a(j,1)) .gt. ABS(dum)) THEN
             dum = a(j,1)
             i = j
          END IF
       END DO
       index(k) = i
       IF(dum .eq. 0.) then 
          a(k,1) = 1.D-30
          print*,'DING !'
!!          read(*,*)
       ENDIF
       IF(i.NE.k)THEN
          DO j = 1,mm
             dum    = a(k,j)
             a(k,j) = a(i,j)
             a(i,j) = dum
          END DO
       END IF
       DO i = k+1,l
          dum = a(i,1)/a(k,1)
          al(k,i-k) = dum
          DO j = 2,mm
             a(i,j-1) = a(i,j) - dum*a(k,j)
          END DO
          a(i,mm) = 0.
       END DO
    END DO
    RETURN
  END SUBROUTINE 

  !     ==========================================================
  SUBROUTINE LU_substitution(a,n,m1,m2,np,mp,al,mpl,index,b)
    !     ==========================================================
    ! --- Ayant les tableaux a,al et index calcules par la routine 'bande_LU_decomposition',
    !     et connaissant un vecteur b(1:n), cette routine resoud le systeme lineaire bande
    !     diagonal Ax = b. Le vecteur solution x est stockee dans le tableau b. Les autres
    !     entrees ne sont pas modIFiees et peuvent etre utilisees pour des appels successIFs
    !     a cette routine.
    !     np, mp et mpl sont les dimensions physiques des tableaux a et al.
    !     On doit avoir np  n ; mp  m1 + 1 + m2 ; mpl  m1.

    IMPLICIT NONE
    ! --- Les arguments de la routines :
    integer:: m1,m2,mp,mpl,n,np
    integer,dimension(n)::index
    double precision,dimension(np,mp)::a
    double precision,dimension(np,mpl)::al
    double precision,dimension(n)::b
    ! --- Les variables internes a la routine :
    integer:: i,k,l,mm
    double precision::    dum

    mm = m1 + m2 + 1
    IF ((mm .gt. mp).OR.(m1 .gt. mpl).OR.(n .gt. np)) &
         STOP   'Probleme d''arguments dans la substitution LU.'
    l = m1
    DO k = 1,n
       i = index(k)
       IF (i.NE.k) THEN
          dum  = b(k)
          b(k) = b(i)
          b(i) = dum
       END IF
       IF (l .lt. n) l = l + 1
       DO i = k+1,l
          b(i) = b(i) - al(k,i-k)*b(k)
       END DO
    END DO
    l = 1
    DO i = n,1,-1
       dum = b(i)
       DO k = 2,l
          dum = dum - a(i,k)*b(k+i-1)
       END DO
       b(i) = dum/a(i,1)
       IF (l .lt. mm) l = l + 1
    END DO
    RETURN
  END SUBROUTINE 


  ! Exponentielles avec seuil:

  FUNCTION Plsinh(f) Result(ef)
    double precision::f
    double precision::ef
    if (verbose.ge.5) print*,'     sinh'
    if (abs(f).le.100.)then
       ef=sinh(f)
    else
       ef=.5*(plexp(f)-plexp(-f))
    endif
    return
  end FUNCTION Plsinh

  FUNCTION Plexp(f) Result(ef)
    double precision::f
    double precision::ef


    if(f.lt.-100.)then
       ef=0.
    else
       ef=exp(f)
    endif
    return
  end FUNCTION Plexp

  FUNCTION Pexp(f) Result(ef)
    use dimensions,only:nz
    double precision,dimension(nz)::f
    double precision,dimension(nz)::ef


    where(f.lt.-100.)
       ef=0.
    elsewhere
       ef=exp(f)
    endwhere
    return
  end FUNCTION Pexp

  FUNCTION decexp(f) Result(ef)
    use dimensions,only:nz
    double precision,dimension(nz)::f
    double precision,dimension(nz)::ef

    where(f.lt.-100.)
       ef=0.
       elsewhere
       ef=10.**(f)
    endwhere
    return
  end FUNCTION decexp

  ! Usefull Functions
  
  ! " Chemical log "
  elemental double precision function logc(x,th)
    use chemistry,only:threshold
    implicit none
    double precision,intent(IN),optional::th
    double precision,intent(IN)::x
    if (present(th)) then
       logc=log(x/th+1.)/log(2.)
    else
       logc=log(x/threshold+1.)/log(2.)
    endif
  end function 

  elemental double precision function dlogc(x,th)
    use chemistry,only:threshold
    implicit none
    double precision,intent(IN),optional::th
    double precision,intent(IN)::x
    if (present(th)) then
       dlogc=1./log(2.)/(x+th)
    else
       dlogc=1./log(2.)/(x+threshold)
    endif
  end function 

  ! inverse of log(x+1)/log(2)
  elemental double precision function expc(x)
    implicit none
    double precision,intent(IN)::x
    if (x.lt.-100.) then
       expc=-1.
    else
       expc=2.**x-1.
    end if
  end function

  ! argsh
  elemental double precision function argsh(x)
  double precision,intent(IN)::x
    if (x.ge.0.) then
       argsh=log(sqrt(x*x+1.)+x)
    else
       argsh=-log(sqrt(x*x+1.)-x)
    endif
  end function

  elemental double precision function dargsh(x)
  double precision,intent(IN)::x
    dargsh=1./sqrt(x*x+1.)
  end function

!!$  function argsh(x) result(f)
!!$    double precision,dimension(:)::x
!!$    double precision,dimension(:),allocatable::f
!!$    integer::n
!!$    n=size(x)
!!$    allocate(f(n))
!!$    f=log(sqrt(x*x+1.)+x)
!!$    deallocate(f)
!!$  end function argsh
!!$  function dargsh(x) result(f)
!!$    double precision,dimension(:)::x
!!$    double precision,dimension(:),allocatable::f
!!$    integer::n
!!$    n=size(x)
!!$    allocate(f(n))
!!$    f=1./sqrt(x*x+1.)
!!$    deallocate(f)
!!$  end function dargsh

  ! Cubic root
  function cbrt(vx)Result(f)
    use dimensions,only:nz
    double precision,dimension(nz)::vx,f
    f=vx**(1./3.)!EXP(LOG(VX)/3.0D0)
  end function cbrt

  elemental double precision function pps(x)
  double precision,intent(IN)::x
    if (x.gt.0.) then 
       pps=x
    else
       pps=0.
    endif
  end function

  function ps(vx)Result(f)
    use dimensions,only:nz
    double precision,dimension(nz)::vx,f
    f = 0.5D0*(VX+ABS(VX))
  end function ps

  function ps1(vx) Result(f)
    use dimensions,only:nz
    double precision,dimension(nz-1)::vx,f
    f = 0.5D0*(VX+ABS(VX))
  end function ps1
  function ps2(vx) Result(f)
    use dimensions,only:nz
    double precision,dimension(nz-2)::vx,f
    f = 0.5D0*(VX+ABS(VX))
  end function ps2
  function ps3(vx) Result(f)
    use dimensions,only:nz
    double precision,dimension(nz-3)::vx,f
    f = 0.5D0*(VX+ABS(VX))
  end function ps3

  function lcbrt(vx)Result(f)
    double precision vx
    double precision f
    f=EXP(LOG(VX)/3.0D0)
    return
  end function lcbrt

  function lps(vx)Result(f)
    double precision vx
    double precision f
    f = 0.5D0*(VX+ABS(VX))
    return
  end function lps

  function rlobe(vx)Result(f)
    double precision::vx,f
    f = 0.49D0*VX*VX/(0.6D0*VX*VX+LOG(1.0D0+VX))
  end function rlobe

  function fxp(vx)Result(f)
    use dimensions,only:nz
    double precision,dimension(nz)::vx,f
    f = EXP(MAX(-50.0D0,MIN(50.0D0,VX)))
  end function fxp

  SUBROUTINE SPLINE(N,X,F)
    !
    ! Calculate the coefficients of a 1-D cubic spline:
    !
    ! Forsythe, Malcolm, Moler, Computer Methods for Mathematical
    ! Computations, Prentice-Hall, 1977, p.76
    !
    implicit none
    double precision,dimension(:)::x
    double precision,dimension(:,:)::f
    integer::n
    !locals
    integer i,ib
    double precision t

    F(2,N) = 0D0
    F(3,N) = 0D0
    F(4,N) = 0D0

    ! Set up a tridiagonal system for A*y=B where y(i) are the second
    ! derivatives at the knots.
    ! f(2,i) are the diagonal elements of A
    ! f(4,i) are the off-diagonal elements of A
    ! f(3,i) are the B elements/3, and will become c/3 upon solution
    !
    F(4,1) = X(2)-X(1)
    F(3,2) = (F(1,2) - F(1,1))/F(4,1)
    DO I = 2,N-1
       F(4,I) = X(I+1) - X(I)
       F(2,I) = 2D0*(F(4,I-1) + F(4,I))
       F(3,I+1) = (F(1,I+1) - F(1,I))/F(4,I)
       F(3,I) = F(3,I+1) - F(3,I)
    ENDDO
    !
    ! Boundaries.
    !
    F(2,2) = F(4,1) + 2D0*F(4,2)
    F(3,2) = F(3,2)*F(4,2)/(F(4,1) + F(4,2))
    !
    F(2,N-1) = 2D0*F(4,N-2) + F(4,N-1)
    F(3,N-1) = F(3,N-1)*F(4,N-2)/(F(4,N-1) + F(4,N-2))
    !
    ! Forward elimination.
    !
    T = F(4,2)/F(2,2)
    F(2,3) = F(2,3) - T*(F(4,2) - F(4,1))
    F(3,3) = F(3,3) - T*F(3,2)
    DO I = 4,N-2
       T = F(4,I-1)/F(2,I-1)
       F(2,I) = F(2,I)-T*F(4,I-1)
       F(3,I) = F(3,I)-T*F(3,I-1)
    ENDDO
    T = (F(4,N-1) - F(4,N-1))/F(2,N-2)
    F(2,N-1) = F(2,N-1) - T*F(4,N-2)
    F(3,N-1) = F(3,N-1) - T*F(3,N-2)
    !
    ! Back substitution.
    !
    F(3,N-1) = F(3,N-1)/F(2,N-1)
    DO IB = 1,N-4
       I = N-1-IB
       F(3,I) = (F(3,I) - F(4,I)*F(3,I+1))/F(2,I)
    ENDDO
    F(3,2) = (F(3,2) - (F(4,2) - F(4,1))*F(3,3))/F(2,2)
    !
    ! Reset d array to step size.
    !
    F(4,1) = X(2) - X(1)
    F(4,N-1) = X(N) - X(N-1)
    !
    ! Set f(3,1) for not-a-knot.
    !
    F(3,1) = (F(3,2)*(F(4,1) + F(4,2)) - F(3,3)*F(4,1))/F(4,2)
    F(3,N) = F(3,N-1) + (F(3,N-1) - F(3,N-2))*F(4,N-1)/F(4,N-2)
    !
    ! Compute the polynomial coefficients.
    !
    DO I = 1,N-1
       F(2,I) = (F(1,I+1) - F(1,I))/F(4,I) - F(4,I)*(F(3,I+1) + 2.0D0*F(3,I))
       F(4,I) = (F(3,I+1) - F(3,I))/F(4,I)
       F(3,I) = 3D0*F(3,I)
       F(4,I) = F(4,I)
    ENDDO
    RETURN
  END SUBROUTINE SPLINE
  
  double precision function comp(a,b)
    double precision::a,b
    if (a==0..and.b==0.) then
       comp=0.
       return
    endif
    comp=(a-b)/(ABS(a)+ABS(b))
  end function comp

  function squadric(a,b,c) Result (x)
    ! Solves for the smallest positive root of ax^2+bx+c=0,
    ! or the biggest if both are negative
    implicit none
    double precision::a,b,c
    double precision::q,d
    double precision::x

    if (abs(a).lt.1.d-150) then
       if (b.ne.0.) then
          x=-c/b
          return
       else
          x=0.
          return
       endif
    endif
    d=b*b-4.*a*c
    if (d.lt.0) then
       x=0.
    else
!       q=-.5*(b+sign(1.,b)*sqrt(d))
       if (b.ge.0.)  then
          q=-.5*(b+sqrt(d))
       else
          q=-.5*(b-sqrt(d))
       endif
       if (a.eq.0..and.q.eq.0.) then
          x=0.
          return
       endif
       if (a.eq.0.) then
          x=c/q
          return
       endif
       if (q.eq.0.) then
          x=q/a
          return
       endif
       x=MIN(q/a,c/q)
       if (x.lt.0.) then
          x=MAX(q/a,c/q)
       endif
!       x=MAX(q,b/q)
    endif
  end function squadric

  function lquadric(a,b) Result (x)
    ! Solves for the biggest root of x^2+ax+b=0
    implicit none
    double precision::a,b
    double precision::q,d
    double precision::x

    d=a*a-4.*b
    if (d.lt.0) then
       x=0.
    else
       q=-.5*(a+sign(1.,a)*sqrt(d))
!!$       if (a.ge.0.)  then
!!$          q=-.5*(a+sqrt(d))
!!$       else
!!$          q=-.5*(a-sqrt(d))
!!$       endif
       x=MAX(q,b/q)
    endif
  end function lquadric

  function quadric(a,b) Result (x)
    ! Solves x^2+ax+b=0
    use dimensions,only:nz
    implicit none
    double precision,dimension(nz)::a,b
    double precision,dimension(nz)::q,d
    double precision,dimension(nz)::x
    integer i

    d=a*a-4.*b
    do i=1,nz
       if (d(i).lt.0) then
          x(i)=0.
       else
          q(i)=-.5*(a(i)+sign(1.,a(i))*sqrt(d(i)))
          if (a(i).ge.0.)  then
             q(i)=-.5*(a(i)+sqrt(d(i)))
          else
             q(i)=-.5*(a(i)-sqrt(d(i)))
          endif
          x(i)=MAX(q(i),b(i)/q(i))
       endif
    enddo
!!$    q=-.5*(a+sign(a,1.)*sqrt(d))
!!$    where(a.ge.0.) 
!!$       q=-.5*(a+sqrt(d))
!!$    elsewhere
!!$       q=-.5*(a-sqrt(d))
!!$    endwhere
!!$    x=MAX(q,b/q)
  end function quadric

  function discr(a,b,c) Result(x)
    ! Discriminant of the cubic x^3+ax^2+bx+c=0
    use dimensions,only:nz
    use parameters,only:
    implicit none
    double precision::pi23
    double precision,dimension(nz)::a,b,c
    double precision,dimension(nz)::q,r
    double precision,dimension(nz)::x
    integer i

    Q=(a*a-3.*b)/9.
    R=(2.*a*a*a-9.*a*b+27.*c)/54.
    x=r*r-q*q*q
  end function discr

  function cubic (a,b,c) Result (x)
    ! Solves x^3+ax^2+bx+c=0
    use dimensions,only:nz
    use parameters,only:
    implicit none
    double precision::pi23
    double precision,dimension(nz)::a,b,c
    double precision,dimension(nz)::q,r,aa,bb,theta,test,y
    double precision,dimension(nz)::x
    integer i
    if (verbose.ge.3) print*,'   cubic'

    pi23=2./3.*acos(-1.)


    Q=(a*a-3.*b)/9.
    R=(2.*a*a*a-9.*a*b+27.*c)/54.
    test=r*r-q*q*q
    

    do i=1,nz
       if (c(i).eq.0.) then
          x(i)=lquadric(a(i),b(i))
       else
          if (test(i).lt.0.) then         
             ! Take the greatest of the three double precision roots
             theta(i)=acos(R(i)/sqrt(Q(i)*Q(i)*Q(i)))/3.
             x(i)=-2.*sqrt(Q(i))*MIN(cos(theta(i)),cos(theta(i)+pi23),cos(theta(i)-pi23))-a(i)/3.
          else
!!$          print*,' But that means two stream has no solution...'
             ! Take the only double precision root
             if (r(i).gt.0.) then
                aa(i)=-lcbrt(abs(R(i))+sqrt(test(i)))
             else
                aa(i)=lcbrt(abs(R(i))+sqrt(test(i)))
             endif
             if (aa(i)==0.) then
                bb(i)=0. 
             else
                bb(i)=q(i)/aa(i)
             end if
             x(i)=(aa(i)+bb(i))-a(i)/3.
!!$             if (LUMINOSITY.eq.'rad') then
!!$                ! Semi-convection
!!$                ! In this case, take the real part into account
!!$                x(i)=MAX(aa(i)+bb(i),-.5*(aa(i)+bb(i)))-a(i)/3.
!!$             endif
!!$          print*,'Real solution :',x(i)
!!$          print*,'Real part of complex roots :',-.5*(aa(i)+bb(i))-a(i)/3.
!!$          print*,'Im part of complex roots : +/-',sqrt(3.)*.5*(aa(i)-bb(i))
!!$          stop
          endif
       endif
    end do
    return

    if (ANY(test.lt.0.)) then
       where(test.lt.0.)
          ! Take the greatest of the three double precision roots
          theta=acos(R/sqrt(Q*Q*Q))/3.
          x=-2.*sqrt(Q)*MIN(cos(theta),cos(theta+pi23),cos(theta-pi23))-a/3.
       endwhere
    endif
    if (ALL(test.lt.0.))  return

    print*,'OK up to here...'
    where (test.ge.0.)
       ! Take the only double precision root
       where(r.gt.0.) 
          aa=-cbrt(abs(R)+sqrt(test))
       elsewhere
          aa=cbrt(abs(R)+sqrt(test))
       endwhere
       where (aa==0.) 
          bb=0. 
       elsewhere 
          bb=q/aa
       endwhere
       x=(aa+bb)-a/3.
    endwhere
  end function cubic

  function lcubic (a,b,c) Result (x)
    ! Solves x^3+ax^2+bx+c=0
    use parameters,only:
    implicit none
    double precision::pi23
    double precision::a,b,c
    double precision::q,r,aa,bb,theta,test,y
    double precision::x
    if (verbose.ge.3) print*,'   cubic'

    pi23=2./3.*acos(-1.)


    Q=(a*a-3.*b)/9.
    R=(2.*a*a*a-9.*a*b+27.*c)/54.
    test=r*r-q*q*q


    if (c.eq.0.) then
       x=lquadric(a,b)
    else
       if (test.lt.0.) then         
          ! Take the greatest of the three double precision roots
          theta=acos(R/sqrt(Q*Q*Q))/3.
          x=-2.*sqrt(Q)*MIN(cos(theta),cos(theta+pi23),cos(theta-pi23))-a/3.
       else
!!$          print*,' But that means two stream has no solution...'
          ! Take the only double precision root
          if (r.gt.0.) then
             aa=-lcbrt(abs(R)+sqrt(test))
          else
             aa=lcbrt(abs(R)+sqrt(test))
          endif
          if (aa==0.) then
             bb=0. 
          else
             bb=q/aa
          end if
          x=(aa+bb)-a/3.
!!$          if (LUMINOSITY.eq.'rad') then
!!$             ! Semi-convection
!!$             ! In this case, take the real part into account
!!$             x=MAX(aa+bb,-.5*(aa+bb))-a/3.
!!$          endif
!!$          print*,'Real solution :',x
!!$          print*,'Real part of complex roots :',-.5*(aa+bb)-a/3.
!!$          print*,'Im part of complex roots : +/-',sqrt(3.)*.5*(aa-bb)
!!$          stop
       endif
    endif
    return
  end function lcubic
  
  subroutine smooth(f)
    use dimensions,only:nz
    use parameters,only:stiffness
    implicit none
    integer::i,j
    double precision::sum,coeff,mcoeff
    double precision,dimension(nz)::f,fs,df
    logical::test
    if (verbose.ge.3) print*,'   smooth'

    j=0
    fs=f
!!$    mcoeff=log(1.+1./(stiffness*(1.+stiffness)))
!!$    coeff=MAXVAL(abs(log(f(1:nz-1))-log(f(2:nz))))
!!$    print*,'min(f)',minval(f)
!!$    fs=exp(log(f)*min(.99*mcoeff/coeff,1.))
    coeff=MAX(MAXVAL(fs(1:nz-1)/fs(2:nz)),&
         1./MINVAL(fs(1:nz-1)/fs(2:nz)))
    print *,'smooth step ',j, 'maxrel=',coeff,' mcoeff=',exp(mcoeff)
    test=.false.
    do i=2,nz-1
       if (1./fs(i)-stiffness*(stiffness+1.)*(1./fs(i+1)-2.*1./fs(i)+1./fs(i-1)).le.0.)&
            test=.true.
    enddo
    do while(test)
       j=j+1
       print*,'min(fs)',minval(fs)
       do i=1,nz-1
          fs(i)=sqrt(fs(i)*fs(i+1))
       enddo
       fs(nz)=fs(nz-1)

       test=.false.
       do i=2,nz-1
          if (1./fs(i)-stiffness*(stiffness+1.)*(1./fs(i+1)-2.*1./fs(i)+1./fs(i-1)).le.0.)&
               test=.true.
       enddo
          
       coeff=MAX(MAXVAL(fs(1:nz-1)/fs(2:nz)),&
            1./MINVAL(fs(1:nz-1)/fs(2:nz)))
       print *,'smooth step ',j, 'maxrel=',coeff
    enddo
    f=fs
  end subroutine smooth

  real Function c1c(a,r1,r2,r3)
    double precision::a,r1,r2,r3    
    c1c=a/(r1-r2)/(r1-r3)
  end Function c1c

  real Function c2c(a,r1,r2,r3)
    double precision::a,r1,r2,r3    
    c2c=-a*(r2+r3)/(r1-r2)/(r1-r3)
  end Function c2c

  real function c1(a1,a2,a3,r1,r2,r3)
    double precision::a1,a2,a3,r1,r2,r3
    c1=c1c(a1,r1,r2,r3)+c1c(a2,r2,r3,r1)+c1c(a3,r3,r1,r2)
  end  function c1

  real function c2(a1,a2,a3,r1,r2,r3)
    double precision::a1,a2,a3,r1,r2,r3
    c2=c2c(a1,r1,r2,r3)+c2c(a2,r2,r3,r1)+c2c(a3,r3,r1,r2)
  end  function c2

  real function max3(a1,a2,a3,r1,r2,r3)
    double precision::a1,a2,a3,r1,r2,r3
    max3=-.5*c2(a1,a2,a3,r1,r2,r3)/c1(a1,a2,a3,r1,r2,r3)
  end function max3

  real function smax(f,r)
    use dimensions,only:nz
    implicit none
    double precision,dimension(nz)::f,r
    integer i
    
!!$    print*,'Test max3. Those numbers should be zero :'
!!$    print*,max3(-1.,0.,-1.,-1.,0.,1.)
!!$    print*,max3(-.25,0.,-1.,-.5,0.,1.)
!!$    print*,max3(-.25,-0.01,-1.,-.5,.1,1.)
!!$    read(*,*)
    i=maxloc(f,dim=1)
    if (i==1.or.i==nz) then
       smax=r(i)
       return
    endif
    smax=max3(f(i-1),f(i),f(i+1),r(i-1),r(i),r(i+1))
  end function smax
    
end MODULE maths
