!=======================================================================
!
!
!                          FLASH_THE_TORTOISE
!
!   An unoptimised stellar evolution code that will (hopefully) 
!   take you through the C-flash. Maybe through He-flash as well ?
!   PL 15/11/2005
!
!
!=======================================================================
PROGRAM newton
  ! Main program. Uses a Newton-Raphson method to find the solution 'x'
  ! of equations 'equ'.

  ! The scope of these variables could be reduced to increase the readability
  ! of the code (by redistributing each of them to its relevant subroutine). 
  ! But it is more practical to keep them here for debugging
  ! and code development purposes
  use dimensions
  use constants,only:base,cmevmu,cl,Msun,Lsun,CLN10,pi4,csecyr
  use parameters,only:Masslim,jp,Tlim,mcentre,UCHEM,MESH
  use chemistry,only:threshold,iprim,rstatus,H1,He4,frac_init,tiny,a23d,a23m,specie,ispecie,iup23,bn
  use maths
  use nuclear,only:rate
  use abundances,only:xa
  use secondaries
  use variables,only:lnT,M,L,Lold,mold,resold,mt,r,dm,Ltotold,netheatold,dmold,rold,gfracold,fracnew,Uold
  use printings,only:IMP,nstep,nimp,FOLLOW,paramprintings,tphoto,nphoto,&
       photostep,vartop,nstep_init,&
       trace,photofilename,output,trace_init,initphoto,&
       outputhydro,numero
  use equations,only:setsecold,savesecold,recoversecold,dg,equns1,Ltotold,&
       lg,res,Ltot,tkhold,relax
  implicit none

  ! Names of equations
  character*7::nameofequ(nvar)

  ! Control parameters
  character*3::RESTART
  character*3::JPC,oldJPC,CHECKL,CHECKG
  character*3::DERIVATIVE
  integer::najust
  integer::nbmass=1
  integer:: nphoto_init
  double precision::dt_init
  double precision::dt1,dt2
  double precision::dtmax
  double precision::varrel,amax,asum 
  double precision::tolx,tolequ,tolcheck
  double precision::epsder
  integer::itermax
  integer::itermod,itervec
  double precision::tstop
  integer::kmax

  ! matrix
  double precision,dimension(neq)        ::EQU
  double precision,allocatable,dimension(:,:)::Asquare,A
  double precision::maxequ,tolsqr
  logical::correctsqr,stopsqr
  logical::singular,nonconv
  double precision,dimension(m1+m2+1,neq)  ::Curv,Asave,compare
  double precision,dimension(neq)          ::B0,B,Bsave,compareB,dxsave,xsave,xsave2
  integer                      ::ident,identequ,info
  integer pskip ! minimum interval of the comb
  ! matrix save for SOLVE_JPC
  double precision::A_jpc(1:neq,1:m1+m2+1)
  double precision::A0 (1:neq,1:m1+m2+1)
  double precision::A_L(1:neq,1:m1)
  integer,dimension(neq)::index
  double precision,dimension(neq)::SA
  logical toto,allnum

  ! Remesh info
  logical::remesh=.false.

  ! counters
  integer            i,ii,ij,j,iz,k,jvar,kchim,kk,iiter,mm,ivar
  integer            niter,niter2

  ! Newton-Raphson convergence
  double precision::massfrac,dmassfrac
  integer::itry
  logical::clear,invert
  double precision::dh
  double precision,dimension(nvar)::xtypical
  double precision,dimension(neq)::xscale,one,xder,savex
  double precision,dimension(nz+1)::saveLtot
  logical,dimension(nvar)::choice,choicequ,choicet
  double precision,dimension(neq)::varelative
  logical,dimension(neq):: testi
  logical::testnr,advanced,restartnr,reduce
  double precision  ::    checkcoeff,  coeff
  logical     out
  integer ngoback
  logical toomuch
  double precision::dt_want
  double precision::errcontrol,errpr,equpr
  double precision,dimension(neq)::errel
  !!  double precision,dimension(nvar)::errel
  double precision,dimension(nvar)::normequ
  double precision nemax
  integer inemax

  ! Previous time step
  double precision::dtoldpr
  double precision,dimension(neq)::dxoldpr,xoldpr

  ! kodak
  logical clicclac

  ! Identify the structure of the variables
  CALL identify

  ! Setup physical constants in cgs units
  CALL base

  ! Read the physical, chemical and control parameters
  CALL readlist

  ! Initialize x and dxold vectors
  CALL init

  ! Save previous and anteprevious(=previous at first) step
  xold=x-dxold
  xoldpr=xold
  dxoldpr=dxold
  dtoldpr=dtold

  ! initialize the "pictures" counter
  CALL initphoto

  ! Initialize the TRACE file if needed'
  if (FOLLOW.eq.'new')   CALL TRACE_init

  !     ====================  TIME STEPS LOOP  ====================  
  !     k is a passive index on time steps
  k=1
  ! nstep is from the beginning of the computation
  nstep=k+nstep_init
  ! Count the number of iterations with a matrix inversion
  niter=0

  ! set up old times
  dtsave=dt
  dt=dtold
  time=time-dt

  ! -- Compute and Save old secondaries --
  dx=0.
  x=xold
  CALL setequ
  CALL setsecold

  ! ----- First print ---------
  ! set equations
  dx=dxold
  x=xold+dxold
  CALL setequ
  CALL imp(0)
  ! -- End of first print --

  ! Recover current times
  time=time+dt
  dt=dtsave
  dtsave=dtold

  itervec=0

  DO WHILE (k.le.KMAX)
     if (verbose.ge.-10) print*,''
     ngoback=0
     
     ! former current x becomes the old one
     xold=x
     ! save old secondaries
     CALL savesecold
     ! update old secondaries
     CALL setsecold

     ! In case of non convergence, go back HERE
1111 continue

     ! Print time step informations
     if (k.eq.1) then
        print 199,k,nstep,dt/csecyr,time/csecyr
     else
        print 200,k,nstep,dt/csecyr,time/csecyr,float(niter)/float(k-1)
     endif
199  FORMAT(1x,i5,'st step (',i5,')',2x,'dt = ',1pe10.3,' year  t=',1pe10.3,' year')
200  FORMAT(1x,i5,'th step (',i5,')',2x,'dt = ',1pe10.3,' year  t=',1pe10.3,' year  <niter>=',1pe10.3)

     ! Global typical values
     do j=1,nvar
        xtypical(j)=MAX(1.,norm(xold,j,0))
     enddo
     one=1.

     ! Advanced solution as a first guess
     if (jp==0) dxold=0.
     dx=dxold
     x=xold+dx
     if (ALL(dxold.eq.0.))then
        print*,'1st guess not advanced'
        advanced=.false.
     else
        advanced=.true.
     endif

     ! In case of too much variation during a time step, go back HERE
2222 continue

     ! Check that the advancing step does not go out of definition domain
     call xdomain

     ! Update variables and equations
     CALL setequ

     ! N-R controls
     errcontrol=0.
     testnr=.true.
     checkcoeff=1.+tolcheck
     coeff=1.
     restartnr=.false.
     reduce=.false.
     CALL vtypical
     CALL sizequ
     nonconv=.false.
     ! ##################  Newton-Raphson loop ##################
     ! iiter is active counter of N-R convergence.
     iiter = 1
     do while (.not.nonconv.and.testnr)
        if (restartnr) restartnr=.false.

        ! Compute typical values for derivatives
        CALL vtypical

        ! Local derivative scale
        xder=xscale+abs(xold)
        xder=epsder*xder

        ! initialize equations
        B = - EQU 
        Bsave=B

        ! fill matrix
        invert=(itervec.eq.0)
        if (invert) then
           niter=niter+1
           pskip=2
           CALL setmatrix
        endif
        ! save matrix
        Asave=A

        ! debug (to uncomment in case you need it)
!!$        print*,'Type 1 to save matrix.'
!!$        read(*,*) i
!!$        if (i==1) CALL DEBUG  

        ! invert the system A.B=B0
        CALL invertsys
        ! Correct B to respect definition domain
        CALL Bdomain
        
        ! Compute size of correction in x for control of convergence purposes
        CALL sizeB

        ! ajust Newton-Raphson shoot 
        call ajust

        ! print results of the current correction
        if (verbose.ge.-10) then
           print 401,iiter,errel(ident),mt(zone(ident))*csecyr/Msun,&
                coeff,nameofvar(variable(ident)),zone(ident),xscale(ident)
401        FORMAT('iter=',1x,i2,' dx/x=',1pe10.3,2x,'mt=',1pe10.3,2x'fl:',e10.3,2x,a7,' (',i4,')',' xs=',1pe10.3)
        endif

        ! Save previous solution
        dxsave=dx
        xsave=x

        !  new improved solution
        if ((checkcoeff.gt.tolcheck.and.coeff.gt..1).or.k.le.1.) then
           dx=dx+coeff*xscale*B
           x=x+coeff*xscale*B
        endif

        ! Update secondaries and equations
        call xdomain
        call setequ

        ! Print variables around pbs if requested
        if (itermax.lt.0.or.verbose.ge.-3) CALL imp(zone(ident))

        ! Errors on equations
        CALL sizequ

        ! abort correction if matrix not updated and it goes wrong
        if ((tolsqr.gt.equpr.or.checkcoeff.lt.tolcheck.or.coeff.lt..1)&
             .and.itervec.ne.0) then
           coeff=1.
           checkcoeff=1.+tolcheck
           itervec=itermod-1
           B=0.
           x=xsave
           dx=dxsave
           call setequ 
        endif

        ! Increase counters for number of matrix inversions (iiter) and number
        ! of corrections since last matrix inversion (itervec)
        itervec=itervec+1
        if (itervec==itermod) itervec=0
        if (itervec==0) iiter=iiter+1

        ! Keep on iterating if testnr
        testnr=errcontrol.gt.tolx.or.stopsqr.or.correctsqr
        if (tolequ.ne.0.) testnr=testnr.or.tolsqr.gt.tolequ
        ! Unless something goes wrong (non convergence)
        nonconv=(testnr.and.iiter.gt.abs(itermax)).or.singular&
             .or.checkcoeff.le.tolcheck.or.(coeff.le.0.1.and.k.gt.1)
     END DO ! while (iiter)

     ! ######################### End of Newton-Raphson ##########################

     ! If the  N-R solver did not converge, decrease dt and do the last one again...
     if (nonconv.and.tolx.ne.0.) then
        itervec=0
        if (ngoback==0) then
           k=k-1
           nstep=nstep-1           
           xold=xoldpr
           time=time-dtold
           dt=dtold
           ! recover old secondaries of previous time step 
           ! (ie: recover anteprevious step)
           CALL recoversecold
        endif
        dt=dt*0.8           
        xold=xoldpr
        dxold=dxoldpr
        dtold=dtoldpr
        PRINT 405,k,dt/csecyr,errcontrol,tolx
405     FORMAT (1x,'* non convergence at step ',i5,2x,'next dt = ',1pe10.3,2x,'year(s) err x : ',e10.3,' > ',e8.1)
        ngoback = ngoback+1
        if (ngoback.gt.50) then
           print*,'Computation diverged... '
           stop
        end if
        go to 1111
     end if

     ! If it did converge, 
     ! compute relative increment through this time step, and deduce new time step
     ! Variables are setup to be used in TIMESTEP
     xsave2=x
     x=xold
     CALL vtypical
     x=xsave2
     !---------------------------
     ! Determine what we want for the next timestep
     call TIMESTEP(toomuch,dt_want)
     !---------------------------

     ! If the relative increment is too high, do it again with a smaller time step
     if (toomuch) then
        dx=dt_want/dt*(x-xold)
        x=xold+dx
        dt=dt_want
        PRINT 406,k,dt_want/csecyr
406     FORMAT (1x,'* TOO MUCH at step ',i5,' ! new time step : dt= ',1pe10.3,' year(s)')
        goto 2222
     endif

     ! If not, we are pleased with this x :

     ! Stop tests
     CALL stoptests

     ! Then, time may pass...
     time=time+dt
     dxold=dx
     if (clear.or.jp==0) dxold=0.
     dtold=dt
     dt=dt_want

     ! Save old values in anteprevious timestep
     xoldpr=xold
     dxoldpr=dxold
     dtoldpr=dtold

     ! Printings and files save
     CALL allprints


     ! Drive the computation for ignition 
     if (restart.eq.'i=0'.or.restart.eq.'rho') CALL ignition

     ! At last, time steps counters are incremented
     nstep=nstep+1       
     k = k+1

  END DO
  ! ======================= End of LOOP ON TIME ==========================


CONTAINS

  ! ================ Read and initialise parameters ================
  SUBROUTINE readlist
    use chemistry,only:paramchemistry,network,readtab
    use parameters,only:paramphysics

    integer unit
    integer j
    if (verbose.ge.0) print*,'readlist'
    print*,' READING INPUT PARAMETERS...'
    unit=13
    OPEN(unit=unit,file='inlist',status='unknown')
    print*,'Read control'
    CALL paramcontrol(unit)
    print*,'Read printings'
    CALL paramprintings(unit)
    print*,'Read physics'
    CALL paramphysics(unit)
    print*,'Read chemistry'
    CALL paramchemistry(unit)
    CLOSE(unit)

    ! read files for chemistry (chemistry-data)
    CALL readtab
    CALL network
    print*,'READING COMPLETED'

  end SUBROUTINE readlist

  SUBROUTINE paramcontrol(unit)
    integer unit

    read(unit,*)!'--- Control parameters ---'
    ! restart
    read(unit,*) RESTART
    ! Time step informations
    read(unit,*) dt_init 
    dt_init=dt_init*csecyr
    read(unit,*) dtmax   
    dtmax=dtmax*csecyr
    read(unit,*) dt1
    read(unit,*) dt2
    read(unit,*) varrel
    read(unit,*) asum
    read(unit,*) amax
    ! N-R parameters
    read(unit,*) JPC
    read(unit,*) CHECKL
    read(unit,*) CHECKG
    read(unit,*) DERIVATIVE
    read(unit,*) epsder
    read(unit,*) tolx
    read(unit,*) tolequ
    read(unit,*) tolcheck
    read(unit,*) itermax
    read(unit,*) itermod
    read(unit,*) najust
    ! stop parameters
    read(unit,*) tstop 
    tstop=tstop*csecyr
    read(unit,*) kmax
  end SUBROUTINE paramcontrol


  ! ================ Variables initialisation =================

  ! ------------------
  SUBROUTINE init
    ! ------------------
    ! initialise x and dxold with 
    ! a STAR model or a previous converged model read in the 'output' file
    use parameters,only:TM,JIN,MESH,STAGGERED,nbs,stiffness,mcentre
    use constants,only:Lsun,pi4,Rsun
    use chemistry
    use secondaries,only:na,rho,p,T
    use variables,only:mold,dm,frac,Tdts,primaries,setsecondaries
    use EOS,only:psi,en
    use nuclear,only:ex
    use equations,only:wcc,MODE,rhowant,ntilde,ntildeold,n,nold,dmt,dxlnp,dg
    use maths,only:argsh,dargsh
    implicit none
    ! interpolation variables
    integer::izread,ijr,ij,k0
    double precision,dimension(nz)::minter,dminter
    double precision,allocatable,dimension(:)::mread,dmread,rread,rhoread,mkread,r2read,r2kread
    double precision::arg,betarg,xlnp,xlnp0,xdlnp,dm0,Acst,totmass,xm,xm0,xm1
    double precision::mp,alpha,incstiff
    ! other variables
    double precision q2,rhomin,rhomax,rtable
    double precision,dimension(22,500)::h,dh
    double precision,parameter::eps=1.d-30
    integer nmesh
    integer idum
    double precision dum,xswap
    double precision xj,yj
    ! counters
    integer     i,j,k,jj,iold
    ! tables to read 
    double precision,dimension(:),allocatable::xread,dxread,lnpread
    double precision,dimension(:),allocatable::xread2,dxread2
    integer::nzread,nvarread
    integer::iread
    double precision::zread,thread,coeff
    double precision,dimension(nz)::mk,r2k,r2,r3
    double precision::xsurf
    if (verbose.ge.0) print*,'init'
    ! RESTART:
    ! 'non' --> STAR model
    ! 'yes' --> restart from 'output' file model
    ! 'chg' --> idem, but a with a new dt and no advanced solution
    ! 't=0' --> like 'chg' but reset time to zero
    nphoto=0
    MODE=RESTART
    if (restart=='non') then
       ! Read initial model characteristics in STAR format
99005  FORMAT (1X, 1P, 2E14.6, E17.9, 3E14.6,0P, 4I5)
       READ  (25, 99005) dum, dum, dum, dum, dum, dum, nmesh, idum, idum, idum
       ! Read the initial model   
       print*,'Read initial model'
       print *,'nmesh',nmesh
       nzread=nmesh
       nvarread=nvar
       allocate(xread(nvarread*nzread),dxread(nvarread*nzread))
       print *,'JIN',JIN
       DO  K = 1, nmesh
99002     FORMAT (1P, 12E15.8, 0P)
          READ (25, 99002) (H(J,K), J=1, JIN)
       END DO

       xread=0.
       dxread=0.
       do i = 1,nzread
          do j=1,species
             if (frac_init(ispecie(j)).ne.0.)&
                  xread(nvar*(i-1)+nvarh+species+j)=frac_init(ispecie(j))
          enddo
       enddo
       ! transfer variables into xread
       do i = 1,nzread
          j=min(i+1,nzread)
          ! Interface variables
          xj=1.
          yj=0.
          xread(nvar*(i-1)+ilnr)=xj*h(7,i)+yj*h(7,j)
          xread(nvar*(i-1)+ilnm)=xj*h(4,i)+yj*h(4,j)
          xread(nvar*(i-1)+il)=xj*h(8,i)+yj*h(8,j)
          ! Center of zones variables         
          xj=0.5
          yj=0.5
          xread(nvar*(i-1)+ilnt)=xj*h(2,i)+yj*h(2,j)
          xread(nvar*(i-1)+ilnf)=xj*h(1,i)+yj*h(1,j)
          if (iprim(H1).ne.0.) &
               xread(nvar*(i-1)+nvarh+species+iprim(H1))  =xj*h(5,i)+yj*h(5,j)
          if (iprim(HE4).ne.0.)&
               xread(nvar*(i-1)+nvarh+species+iprim(HE4)) =xj*h(9,i)+yj*h(9,j)
          if (iprim(C12).ne.0.)&
               xread(nvar*(i-1)+nvarh+species+iprim(C12)) =xj*h(10,i)+yj*h(10,j)
          if (iprim(O16).ne.0.)&
               xread(nvar*(i-1)+nvarh+species+iprim(O16)) =xj*h(3,i)+yj*h(3,j)
          if (iprim(NE20).ne.0.)&
               xread(nvar*(i-1)+nvarh+species+iprim(NE20))=xj*h(11,i)+yj*h(11,j)
       enddo
       ! initial time and time step
       time=0.
       dtold=dt_init
       dt=dt_init
       nstep_init=0
    else
       ! Read 'output' file
       OPEN   (80,FILE = 'output',STATUS = 'unknown')
       read(80,*) nvarread
       read(80,*) nzread
       allocate(xread(nvarread*nzread),dxread(nvarread*nzread))
       read(80,*) xread
       read(80,*) dxread
       read(80,*) time
       read(80,*) dtold
       read(80,*) dt
       read(80,*) nstep_init
       read(80,*) nphoto
       read(80,*,end=33) thread
33     CLOSE  (80)
    endif
    if (nzread.ne.nz.or.restart.eq.'non') then
       ! Number of zones is different in the input file and in the code
       if (nvarread.ne.nvar) then
          ! If number of variables is also different in input file and code
          ! then complain
          print*,'nz=',nz,' <> nzread=',nzread
          print*,'nvar=',nvar,' <> nvarread=',nvarread
          print*,'Two at a time is too much for me...'
          STOP
       endif
       ! If nz<>nzread, then remesh 
       remesh=.true.
       ! This assumes a shape sqrt[1+a*(dlnp/dlnm)^2] for
       ! the resolution function
       ! First we get lnp and m from xread by calling setequ as many
       ! times as needed (REM: this is strongly wasting time.. but
       ! that's only once at the beginning, so I don't think we care)
       allocate(lnpread(nzread),mread(nzread),dmread(nzread),rread(nzread),rhoread(nzread))
       do izread=1,nzread
          do j=1,nvar
             x(j:neq:nvar)=xread((izread-1)*nvar+j)
          enddo
          xold=x
          dx=0.
          call primaries
          call setsecondaries
          lnpread(izread)=lnp(1)
          mread(izread)=m(1)
          dmread(izread)=dm(1)
          rread(izread)=r(1)
          rhoread(izread)=rho(1)
          if (verbose.ge.0) print*,'izr=',izread,' lnp=',lnpread(izread),&
               ' m=',mread(izread)/Msun,&
               ' Msun dm=',dmread(izread)/Msun,' Msun'
       enddo
       if (restart.eq.'non') then
          ! STAR does not save the intershell mass: we need to recover it.
          ! Difference of two adjacent mass is not accurate enough
          ! Better to do it from radii.
          allocate(r2read(nzread),r2kread(nzread),mkread(nzread))
          ! Save total mass
          totmass=mread(1)
          ! Recover difference of radii squared
          r2read=rread*rread
          r2kread(nzread)=.5*r2read(nzread-1)
          do i=nzread-1,1,-1
             r2kread(i)=-r2kread(i+1)+2.*(rread(i)-rread(i+1))*(rread(i)+rread(i+1))
             r2kread(i)=(rread(i)-rread(i+1))*(rread(i)+rread(i+1))
          enddo
          ! Recover mk
          mkread=.5*pi4*rread*rhoread*r2kread
          ! Recover dm
          do i=1,nzread-1
             dmread(i)=.5*(mkread(i)+mkread(i+1))
          enddo
          dmread(nzread)=1.5*mkread(nzread)-0.5*mkread(nzread-1)
          ! Recover m
          mread(nzread)=dmread(nzread)
          do i=nzread-1,1,-1
             mread(i)=mread(i+1)+dmread(i)
          enddo
          ! Normalise to get back on total mass
          totmass=totmass/mread(1)
          dmread=dmread*totmass
          mread=mread*totmass
          do i=1,nzread-1
             mread(i+1)=mread(i)-dmread(i)
          enddo
          dmread(nzread)=mread(nzread)
          deallocate(r2read,r2kread,mkread)
       endif
       deallocate(rhoread,rread)
       ! Central zone mass
       xm0=mread(nzread)
       ! Normalisation parameter
       betarg=float(nbs)/float(nz-nbs)*mread(1)/(lnpread(nzread)-lnpread(1))
       if (verbose.ge.0) print*,'betarg=',betarg
       ! Find transition point for the resolution
       izread=1
       arg=2.
       do while(arg>1.)
          arg=betarg*(lnpread(izread+1)-lnpread(izread))/dmread(izread)
          izread=izread+1
       enddo
       if (verbose.ge.0) print*,'Transition at izread=',izread
       ! Other normalisation parameter
       Acst=float(nz-nbs)/(xm0-mread(izread-1))
       if (verbose.ge.0) print*,'Acst=',Acst
       i=1
       xlnp0=lnpread(1)-.5*(lnpread(2)-lnpread(1))
       dm0=dmread(1)-.5*(dmread(2)-dmread(1))
       ! Then, we compute dminter for outermost meshpoints 
       arg=2.
       izread=1
       do while(arg>1.)
          xlnp=xlnp0+(float(i)-0.5)/betarg/abs(Acst)
          if (verbose.ge.0) print*,'xlnp=',xlnp
          do while (lnpread(izread)<xlnp) 
             izread=izread+1
          enddo
          if (izread.gt.nzread) then 
             print*,'izread=',izread,'>nzread=',nzread
             stop
          endif
          xlnp=xlnp0+float(i)/betarg/abs(Acst)
          if (izread==1) then
             xdlnp=(xlnp-xlnp0)/(lnpread(izread)-xlnp0)
             dminter(i)=dm0*(1.-xdlnp)+dmread(izread)*xdlnp
          else
             xdlnp=(xlnp-lnpread(izread-1))/(lnpread(izread)-lnpread(izread-1))
             dminter(i)=dmread(izread-1)*(1.-xdlnp)+dmread(izread)*xdlnp
          endif
          arg=betarg*(lnpread(izread+1)-lnpread(izread))/dmread(izread)
          if (verbose.ge.0) print*,'i=',i,' izread=',izread,' arg=',arg,'dm=',&
               dminter(i)/Msun
          i=i+1
       enddo
       ! Compute dminter for innermost meshpoints
       ! better than uniform, do some smoothing already:
       mp=-1./Acst
       incstiff=1.2 ! increase stiffness by 20% to get smoother grid
       alpha=(1./incstiff+stiffness)/stiffness
       k0=int(log(mp/xm0)/log(alpha))
       xm1=(xm0*stiffness*incstiff+mp*float(nz-i))/(float(nz-k0-i)+stiffness*incstiff)
       dminter(i:nz-k0-1)=xm1
       do i=nz,nz-k0,-1
          dminter(i)=xm0*alpha**(nz-i)
       enddo
       ! Print dminter
       do i=1,nz
          print*,i,'dm=',dminter(i)/Msun
       enddo
       ! Smooth dminter according to mesh stiffness
       call smooth(dminter)
       ! Print dminter
       do i=1,nz
          print*,i,'dmsmoothed=',dminter(i)/Msun
       enddo
       ! Rescale it to get  total mass right
       totmass=SUM(dminter)
       dminter=dminter*mread(1)/totmass
       ! Save total mass, because we are going to compute masses from the surface
       totmass=mread(1)
       ! Compute mass at interfaces, from surface
       minter(1)=0.!mread(1)
       do i=2,nz-1
          minter(i)=minter(i-1)-dminter(i-1)
       enddo
       minter(nz)=dminter(nz)-mread(1)
       ! Recompute mread for consistency with dmread (a precision matter)
       mread(1)=0.
       do i=2,nzread-1
          mread(i)=mread(i-1)-dmread(i-1)
       enddo
       mread(nzread)=dmread(nzread)-totmass
       ! Interpolate interface variables
       izread=1
       do i=1,nz
          if (verbose.ge.0) print*,'i',i,' m=',minter(i)/Msun
          if (verbose.ge.0) print*,'izread',izread,'mread=',mread(izread)/Msun
          do while(mread(izread)>=minter(i).and.izread<nzread)
             izread=izread+1
          enddo
          if (verbose.ge.0) print*,'izread',izread,'mread=',mread(izread)/Msun
          xm=-(minter(i)-mread(izread-1))/dmread(izread-1)
          do j=1,nvar
             if (interf(j).and.j.ne.ilnm) then
                ij=(i-1)*nvar+j
                ijr=(izread-1)*nvar+j
                x(ij)=xread(ijr-nvar)*(1.-xm)+xread(ijr)*xm
             endif
          enddo
       enddo
       ! Interpolate all other variables
       izread=1
       do i=1,nz
          do while(mread(izread)+.5*dmread(izread)>=minter(i)+.5*dminter(i)&
               .and.izread<nzread)
             izread=izread+1
          enddo
          if (izread==1) then
             xm=-(minter(i)+.5*dminter(i)-mread(izread)-.5*dmread(izread))*&
                  2./(dmread(izread)+dmread(izread+1))
          else
             xm=-(minter(i)+.5*dminter(i)-mread(izread-1)-.5*dmread(izread-1))*&
                  2./(dmread(izread-1)+dmread(izread))
          endif
          do j=1,nvar
             if (.not.interf(j).and.j.ne.idm) then
                ij=(i-1)*nvar+j
                if (izread==1) then
                   ijr=nvar+j
                else
                   ijr=(izread-1)*nvar+j
                endif
                x(ij)=xread(ijr-nvar)*(1.-xm)+xread(ijr)*xm
             endif
          enddo
       enddo
       ! Deallocate tables
       deallocate(lnpread,mread,dmread)
       ! Recover masses from centre
       minter=minter+totmass
       ! Put m and dm in x vector
       do i=1,nz
          x((i-1)*nvar+ilnm)=LOG(minter(i))
          x((i-1)*nvar+idm)=LOG(dminter(i))            
          if (verbose.ge.0.) print*,i,'m=',minter(i)/Msun,' dm=',dminter(i)/Msun
       enddo
       ! Compute ntilde
       call setequ
       ! put ntilde in x vector
       j=ins
       print*,'n(1)',n(1)
       x(j)=LOG(n(1))
       do i=2,nz-1
          print*,'i',i,' ntilde=',ntilde(i-1)
          x((i-1)*nvar+j)=LOG(ntilde(i-1))
       enddo
       print*,'n(nz)',n(nz)
       x(neq-nvar+j)=LOG(n(nz))
       ! Update dxold and xold
       dxold=0.
       xold=x-dxold
       ! End of remesh
       if (restart.eq.'i=0'.or.restart.eq.'rho') then 
          deallocate(xread,dxread)
          ! avoid the call setequ in case of table computation
          CALL ignition
          xold=x-dxold
          return
       endif
       call setequ
    else
       ! If nzread=nz, check if nvarread==nvar
       if (nvarread.ne.nvar) then
          if (nvarread.lt.nvar) then
             print*,'nzread=',nzread
             print*,'nvarread=',nvarread,' < nvar=',nvar
             ! Add  variables 
             ! Since the code cannot guess which variables it needs, you need
             ! to implement this part according to what you want.
             ! If nothing is said, it assumes that only chemistry is changed
             ! and it will put new variables to their frac_init value read
             ! in the 'chemistry-data' file
             x=0.
             dxold=0.
             do i=nz,1,-1
                j=0
                do jj=1,nvarread
                   j=j+1
                   if (idx==2) then
                      if (2*(jj-nvarh-1)==nvarread-nvarh) j=nvarh+species+1
                   endif
                   x((i-1)*nvar+j)=xread((i-1)*nvarread+jj)
                   dxold((i-1)*nvar+j)=dxread((i-1)*nvarread+jj)
                enddo
                if (idx==2) then
                   do j=1,(nvar-nvarread)/2
                      x((i-1)*nvar+nvarh+2*species+1-j)=&
                           frac_init(ndim+1-j)
                   enddo
                endif
             enddo
             dx=x-xold
             CALL setequ
          else
             print*,'nzread=',nzread
             print*,'nvarread=',nvarread,' > nvar=',nvar
             do i=nz,1,-1
                do j=1,nvarh
                   x((i-1)*nvar+j)=xread((i-1)*nvarread+j)
                   dxold((i-1)*nvar+j)=dxread((i-1)*nvarread+j)
                enddo
                do j=1,species
                   ij=(i-1)*nvar+nvarh+j
                   x(ij+species)=xread((i-1)*nvarread+species+nvarh+j)
                   dxold(ij+species)=dxread((i-1)*nvarread+species+nvarh+j)
                   x(ij)=xread((i-1)*nvarread+nvarh+j)
                   dxold(ij)=dxread((i-1)*nvarread+nvarh+j)
                enddo
             enddo
          endif
       else
          x=xread
          dxold=dxread
       endif
    endif
    deallocate(xread,dxread)

    if (restart.eq.'chg') then
       dxold=0.
       dtold=dt_init
       dt=dt_init
    endif
    if (restart.eq.'n=0') then
       ! Reset model number to zero
       nstep_init=0
    endif
    if (restart.eq.'f=0') then
       ! Reset chemical gradients to zero
       dxold=0.
       where(variable.gt.nvarh.and.variable.le.nvarh+species)&
            x=0.
    endif
    if (restart.eq.'t=0') then
       time=0.
       dxold=0.
       dtold=dt_init
       dt=dt_init
       nstep_init=0
    endif
    if (restart.eq.'C+O'.or.restart.eq.'CST'.or.&
         restart.eq.'c+o'.or.restart.eq.'cst') then
       ! Reset a homogeneous composition profile ('cst'), or
       ! a typical C+O WD profile.
       if (restart.eq.'C+O'.or.restart.eq.'CST') then
          ! Reset times if upper case.
          time=0.
          dxold=0.
       endif
       ! C/O=1 in accreted matter
       if (C12.ne.0.and.O16.ne.0.) then                
          xsurf=0.5*(frac_init(C12)+frac_init(O16))
       endif
       dtold=dt_init
       dt=dt_init
       do i = 1,nz
          do j=1,species
             if ((restart=='C+O'.or.restart.eq.'c+o').and.(ispecie(j).eq.C12.or.ispecie(j).eq.O16)) then
                xm=EXP(x(nvar*(i-1)+ilnm))/EXP(x(ilnm))
                xm=min(1.,max(0.,(xm-0.5)/0.4))
                x(nvar*(i-1)+nvarh+species+j)=(1-xm)*frac_init(ispecie(j))+&
                     xm*xsurf                          
             else
                x(nvar*(i-1)+nvarh+species+j)=frac_init(ispecie(j))
             endif
             dxold(nvar*(i-1)+nvarh+species+j)=0.
             if (idx==2) then
                x(nvar*(i-1)+nvarh+j)=0.
                dxold(nvar*(i-1)+nvarh+j)=0.
             endif
          enddo
       enddo
    endif
    ! Finally these modes compute useful tables
    ! 'i=0' computes ignition curve
    ! 'rho' computes table in the rho-T 
    if (restart.eq.'i=0'.or.restart.eq.'rho') then
       ! 2D Table
       CALL ignition
    endif
    ! Get old time step
    xold=x-dxold
    return
  end SUBROUTINE init


  SUBROUTINE ignition
    use chemistry,only:ispecie
    use nuclear,only:nucrat,enx,ex
    use EOS,only:en,eCoulomb,eelectron,pelectron,pcoulomb
    use equations,only:Twant,rhowant,dxlnp
    use variables,only:lnT,lnf
    implicit none
    double precision,save::tmin,tmax=0.,rhomin,rhomax
    integer i,j,nt
    double precision::Ttable
    double precision::rtable
    double precision,dimension(nz)::fr,tl
    logical,save::first=.true.
    integer,save::ncount=0
    integer::istep
    if (verbose.ge.1) print*,' ignition'  
    if (first) then
       first=.false.
       dxlnp=1. ! To avoid a division by zero in printings..
       ! Open the ignition file
       print*,'Write ignition.dat..'
       OPEN(unit=13,file='ignition.dat',status='unknown')
       OPEN(unit=14,file='eos.dat',status='unknown')
       nt=199
       print*,'Number of densities    nz=',nz
       print*,'Number of temperatures nt=',nt
       WRITE(13,*)nz,nt
       WRITE(14,*)nz,nt
!!$       Tmin=3.*10.**(8.)
!!$       Tmax=1.e9
!!$       rhomin=1.e6
!!$       rhomax=1.e10
       Tmin=10.**7.d0
       Tmax=10.**9.d0
       rhomin=10.**(3.d0)
       rhomax=10.**10.d0    
       write(13,*)rhomin,rhomax,Tmin,Tmax
       write(14,*)rhomin,rhomax,Tmin,Tmax
       ! Table for T
       Ttable=Tmax
       if (restart.eq.'i=0') then
          Twant=1.e8
       else
          Twant=Tmax
       endif
       ! Table for rho
       rtable=rhomin       
       do j=1,nz
          rhowant(j)=rtable
          rtable=rtable*(rhomax/rhomin)**(1./float((nz-1)))
       enddo
       ! Initial guess 

       time=0.
       dxold=0.
       dtold=dt_init
       dt=dt_init
       do i = 1,nz!-1
          do j=1,species
             x(nvar*(i-1)+nvarh+species+j)=&
                  frac_init(ispecie(j))!x(nvar*(nz-1)+nvarh+j)
             xold(nvar*(i-1)+nvarh+species+j)=&
                  frac_init(ispecie(j))!x(nvar*(nz-1)+nvarh+j)
          enddo
          x(nvar*(i-1)+ilnt)=log(Twant(i))
          x(nvar*(i-1)+ilnf)=log(1.e8*rhowant(i)/Twant(i)**(1.5))
       enddo
    endif

    ncount=ncount+1
!!$    if (istep==0) then
!!$       print*,'istep=?'
!!$       read(*,*) istep
!!$    endif
    istep=1
    if (ncount.gt.1) then
       if (MOD(ncount,istep).eq.2.or.istep==1) then
          print*,'Ttable=',Ttable
          write(13,*)MAX(MIN(enx+en+ex,1.d30),-1.d30)
          write(14,*)eelectron
          write(14,*)eCoulomb
          write(14,*)pelectron
          write(14,*)pCoulomb
          write(14,*)lnf
!!$       print*,'ec/e',eCoulomb/eelectron
!!$       read(*,*)
       endif
       ttable=ttable*(Tmax/Tmin)**(-1./float(istep*(nt-1)))
       twant=ttable
    endif
    if (ncount.eq.nt*istep+1.or.&
         (ncount.eq.2.and.restart.eq.'i=0')) then
       close(13)
       print*,'...Completed'
       stop
    endif
  end SUBROUTINE ignition

  ! ================= Deals with the matrix =====================

  SUBROUTINE setequ
    ! Setup all equations
    use variables,only:primaries,setsecondaries
    use equations
    if (verbose.ge.0) print*,'setequ '
    ! Compute equations
    CALL primaries
    CALl setsecondaries
    CALL equns1
    CALL setequations
  end SUBROUTINE setequ

  SUBROUTINE setmatrix
    ! Compute matrix elements
    use variables,only:primaries,setsecondaries,frac,gfrac,dmold
    use secondaries,only:nsec,y
    use equations,only:dg,noflagcb,savemt,noprint
    use parameters,only:ich
    implicit none
    integer i,j,mm,ii,jj
    integer k,kk,ivar,jvar,iz,jz,jzz
    integer ideb,ifin
    integer plusun
    integer,dimension(nvar)::nskip,gskip,dskip
    double precision,dimension(neq)::xd,xg,xdd,xgg,dxd,dxg,dxdd,dxgg
    double precision,dimension(neq)::EQUD,EQUG,EQUC,arow
    double precision,dimension(nz,nsec)::ysave,yd,yg
    double precision::dscale
    double precision,dimension(nz)::savedg
    if (verbose.ge.0) print*,'setmatrix '

    noflagcb=.true.
    noprint=.true.
    savemt=.false.

    ! save primaries
    xsave=x
    dxsave=dx
    ! save secondaries
    ysave=y
    ! save equations
    equc=equ
    savedg=dg

    ! Band matrix inversion

    ! Comb definition, we skip
    ! - gskip zones on left
    ! - dskip zones on right
    ! - nskip zones in total
    gskip=pskip
    dskip=gskip
    nskip=1+dskip+gskip

    ! Clear matrix
    if (allocated(A)) deallocate (A)
    allocate(A(ndiag,neq))
    A=0.D0

    ! Loop on derivating variables
    do ivar=1,nvar
       ! Derivation interval is [dxg,dxd]
       where(variable.eq.ivar) 
          dxd=dxsave+xder
          dxg=dxsave-xder
       elsewhere
          dxd=dxsave
          dxg=dxsave
       endwhere
       ! Compute secondary variables deviation
       ! Right
       dx=dxd
       x=xold+dx
       CALL xdomain ! stay in the definition domain
       dxd=dx
       CALL primaries
       CALl setsecondaries
       yd=y

       ! Derivate equations (here the comb is necessary)
       ! Loop on the first zone-1 of the comb
       do kk=0,nskip(ivar)-1

          x=xsave
          y=ysave
          dx=dxsave

          ! Right
          ! Primaries are changed only on the zones of the comb
          do j = ivar+kk*nvar,neq,nvar*nskip(ivar)
             dx(j)=dxd(j)
          enddo
          dxdd=dx
          x=xold+dx
          CALL primaries
          ! Uncomment next line if you don't use pointers in module secondaries
          !POINTERS!          call setsecondaries  
          ! Secondaries are changed only on the zones of the comb
          do j=1+kk,nz,nskip(ivar)             
             y(j,1:nsec)=yd(j,1:nsec)
          enddo
          ! Compute deviations of the equations
          CALL equns1
          CALL setequations
          EQUD=EQU
          EQUG=EQUC
          dxgg=dxsave

          ! Compute derivatives
          do j = ivar+kk*nvar,neq,nvar*nskip(ivar)
             if (dxdd(j)-dxgg(j).eq.0.) then 
                print*,'dxdd=dxgg',nameofvar(variable(j)),'(',zone(j),')'
                print*,'dxdd=',dxdd(j),' xder=',xder(j)
                print*,'xold=',xold(j),' x=',x(j)
             endif
             ideb=max(1,(zone(j)-gskip(ivar)-1)*nvar+1)
             ifin=min(neq,(zone(j)+dskip(ivar))*nvar)
             ideb=max(ideb,j-m1)
             ifin=min(ifin,j+m2)
             do i = ideb,ifin
                mm=i-j+idiag
                A(mm,j)=xscale(j)*(EQUD(i)-EQUG(i))/(dxdd(j)-dxgg(j))
             end do
          enddo
       enddo
    enddo

1   format(a7,'(',i3,')/',a7,'(',i3,')=',1pe10.3)
    ! recover primaries, secondaries and equations
    y=ysave
    x=xsave
    dx=dxsave
    equ=equc
    noflagcb=.false.
    noprint=.false.
    savemt=.true.
  end SUBROUTINE setmatrix

  double precision function fA(i,j)
    ! Get a given matrix element of the band matrix from its indices
    ! in the square matrix it is embedded in
    use dimensions,only:idiag
    integer::i,j
    integer::mm
    mm=idiag+i-j
    if (mm.le.0.or.mm.gt.ndiag) then
       fA=0.
    else
       fA=A(mm,j)
    endif
  end function fA

  SUBROUTINE setequations
    ! Puts nicely named equations into global 'equ' vector.
    use equations
    integer i,j
    if (verbose.ge.0) print*,'setequations'
    ! Name to identify equations
    nameofequ=nameofvar
    do i=1,nz
       equ(nvar*(i-1)+ ilnm)  =equm(i)
       equ(nvar*(i-1)+ idm)   =equdm(i)
       equ(nvar*(i-1)+ ins)   =equsmooth(i)
       equ(nvar*(i-1)+ ilnr)  =equr(i)
       equ(nvar*(i-1)+ ilnf)  =equP(i)
       equ(nvar*(i-1)+ ilnT)  =equT(i)
       equ(nvar*(i-1)+ iL)    =equL(i)
       !       equ(nvar*(i-1)+ i)   =equ(i)
    enddo
    do j=1,species
       nameofequ(j+nvarh)=nameofvar(nvarh+j+species)
       nameofequ(j+nvarh+species)=nameofvar(nvarh+j)
       do i=1,nz
          equ(nvar*(i-1)+nvarh+j)=equX(i,j)
          if (idx==2) equ(nvar*(i-1)+nvarh+j+species)=equdX(i,j)
       enddo
    enddo
  end SUBROUTINE setequations


  !  ========= ajust Newton-Raphson shoot ==========
  SUBROUTINE ajust1(flbest,out)
    ! Simply bounds the correction 
    logical::out
    double precision::flbest
    double precision::maxvar

    maxvar=max(varrel,.99)
    if (errcontrol.gt.maxvar) then
       flbest=maxvar/errcontrol
       out=.true.
       return
    else
       flbest=1.
       return
    endif
  end SUBROUTINE ajust1

  SUBROUTINE ajust2(flbest,out)
    ! Computes equation along the line of shoot and finds minimum
    implicit none
    double precision::flbest
    logical::out
    double precision,dimension(neq) ::x0,equsave
    double precision ::              g1,g1min,g1minold,flbest1,flbest2
    double precision  ::             fl
    integer           :: test
    integer           :: jj,j

    out=.false.

    !  Ajuste le tir Newton-Raphson (cherche le minimum, le mi--ni-mi-nimum..)
    x0=x
    equsave=equ
    do j=1,10
       fl = float(j)/10.
       x=x0+fl*B
       CALL setequ
       g1=.5*SUM(equ**2)
       if (g1.lt.g1min.or.j.eq.1) then
          g1min   = g1
          flbest1 = fl
       endif
    enddo
    !  Decimale suivante:
    g1minold = g1min
    test = 0
    ii   = 1
    do while(test.eq.0  .and. ii.le.4)
       ii=ii+1
       flbest2 = flbest1
       do j= -9,9
          fl = flbest1 + float(j)*10.**(-ii)
          if (fl.ge.0.1.and.fl.le.1.)then
             x=x0+fl*B
             CALL setequ
             g1=.5*SUM(equ**2)
             if (g1.lt.g1min) then
                g1min   = g1
                flbest2 = fl
             endif
          endif
       enddo
       flbest1 = flbest2
       if ((1. -g1min/g1minold).le.0.1) test=1
       g1minold = g1min
    enddo

    flbest = flbest1
    x=x0
    equ=equsave
  end SUBROUTINE ajust2

  SUBROUTINE ajust3(flbest,check)
    ! Same as ajust2, but with gradient method (see numrec)
    ! Supposedly more clever, but hardly ever works... may be bugged !
    use maths,only:bande_multiplication
    implicit none
!!$    ! par critre du gradient  
!!$    double precision,parameter      ::alf=1.e-4
!!$    double precision                ::g0,g1,g2,gp0,fl1,fl2
!!$    logical             ::critere,ok
!!$    double precision,dimension(neq) ::x0,gradf

    double precision::flbest
    logical::check
    double precision,parameter::tolf=1.e-4,tolmin=1.d-6,tolx=1.d-7,stpmx=100.
    integer::i,j
    double precision::den,f,fold,stpmax,summ,temp,test
    double precision,dimension(neq)::g,p,xold,x0

    f=.5*SUM(equ*equ)
    summ=SUM(x*x)
    stpmax=stpmx*max(sqrt(summ),float(neq))
    ! Compute gradient
    call BANDE_MULTIPLICATION(Asave,neq,m1,m2,neq,m1+m2+1,equ,g)
    xold=x
    fold=f
    p=B
    x0=x
    call lnsrch(neq,xold,fold,g,p,x0,f,stpmax,check,flbest)
    if (check) then
       test=0.
       den=max(f,.5*neq)
       do i=1,neq
          temp=abs(g(i))*max(abs(x(i)),1.)/den
          if (temp.gt.test) test=temp
       enddo
       if (test.lt.TOLMIN) then
          check=.true.
          pause 'gradient too low : false convergence'
       else
          check=.false.
       endif
       return
    endif
    test=0.
  end SUBROUTINE ajust3

  SUBROUTINE lnsrch(n,xold,fold,g,p,x0,f,stpmax,check,flbest)
    ! For ajust3 (see numrec)
    implicit none
    integer::n
    logical::check
    double precision::f,fold,stpmax,func
    double precision,dimension(n)::g,p,x0,xold,xsave,equsave
    double precision,parameter      ::alf=1.e-4,tolx=1.d-7
    integer i
    double precision::a,alam,alam2,alamin,b,disc,f2,rhs1,rhs2,slope,summ,temp,test,tmplam,norm,flbest

    check=.false.
!!$          summ=sqrt(SUM(p*p))
!!$          if (summ.gt.stpmax) then
!!$             p=p*stpmax/summ
!!$             norm=stpmax/summ
!!$          else
!!$             norm=1.
!!$          endif
    norm=1.
    slope=SUM(g*p)
    if (slope.ge.0.) then
       print*,'roundoff problem in lnsrch'
       flbest=1.
       return
    endif
    test=0.
    do i=1,n
       temp=abs(p(i))/max(abs(xold(i)),1.)
       if (temp.gt.test) test=temp
    enddo
    alamin=MIN(1.,tolx/test)
    alam=1.
1   continue
    x0=xold+alam*p
    xsave=x
    equsave=equ
    x=x0
    CALL setequ
    f=.5*SUM(equ*equ)
    x=xsave
    equ=equsave
    if (alam.le.alamin) then
       x0=xold
       check=.true.
       alam=alamin
       flbest=alam*norm
       return
    else if (f.le.fold+alf*alam*slope) then
       flbest=alam*norm
       return
    else
       if (alam.eq.1.) then
          tmplam=-slope/(2.*(f-fold-slope))
       else
          rhs1=f-fold-alam*slope
          rhs2=f2-fold-alam2*slope
          a=(rhs1/alam**2-rhs2/alam2**2)/(alam-alam2)
          b=(-alam2*rhs1/alam**2+alam*rhs2/alam2**2)/(alam-alam2)
          if (a.eq.0.) then
             tmplam=-slope/(2.*b)
          else
             disc=b*b-3.*a*slope
             if (disc.lt.0.) then
                tmplam=.5*alam
             else if (b.le.0.) then
                tmplam=(-b+sqrt(disc))/(3.*a)
             else
                tmplam=-slope/(b+sqrt(disc))
             endif
          endif
          if (tmplam.gt..5*alam) tmplam=.5*alam
       endif
    endif
    alam2=alam
    f2=f
    alam=max(tmplam,.1*alam)
    goto 1
  end SUBROUTINE lnsrch


  ! =============== Gestion des pas de temps ===========

  SUBROUTINE TIMESTEP(toomuch,dt_want)
    ! Timestep control
    use variables
    use equations,only:conv,convold
    use secondaries,only:netheat
    use parameters,only:tcool,taccretion
    implicit none
    double precision dt_want
    logical toomuch

    double precision,dimension(neq)::var
    integer::icheck,ivartop
    double precision,dimension(nz+1,nvar)::varf,conf
    double precision,dimension(nvar)::varmax,conmax
    integer,dimension(nvar)::varloc,conloc,ixloc
    integer i,ivar,iz
    integer nselect,ivarmax,ichimmax
    character*9,dimension(nvar)::dnameofvar
    double precision::varcontrol,totalheat
    double precision,dimension(nz)::varres,varheat
    if (verbose.ge.0) print*,'timestep'

    ! Variations of the nuclear luminosities
    totalheat=SUM(abs(netheatold)*dmold,DIM=1)
!    sumvarheat=SUM(abs(netheat-netheatold)*dmold,DIM=1)
    if (totalheat.ne.0.) then
       varheat=abs(netheat-netheatold)*dmold/totalheat
    else
       varheat=0.
    endif
    ! Variations of the convective criterion
    varres=ABS(conv-convold)
    if (ALL(conv==0.))varres=.1*ABS(argsh(1.d1*lg)-argsh(1.d1*lgold))

    ! Relative variations recorded during last time step
    choicet=.true.
    if (idx==2.and.(CHECKG.ne.'tst'.and.CHECKG.ne.'yes')) &
         choicet(nvarh+1:nvarh+species)=.false.
    if (CHECKL.ne.'yes'.and.CHECKL.ne.'tst') choicet(iL)=.false.
    choicet(ins)=.false.
    choicet(idm)=.false.
    WHERE(choicet(variable))
       var=(x-xold)/xscale
    ELSEWHERE
       var=0.
    ENDWHERE

    ! Value to control
    ivartop=MAXLOC(ABS(var),DIM=1)  
    vartop=var(ivartop)
    varcontrol=asum*(SUM(ABS(var),DIM=1)+SUM(ABS(varres))+SUM(varheat))&
         +amax*MAX(MAXVAL(ABS(var)),MAXVAL(ABS(varres)),MAXVAL(varheat))
    ! Set up the format for nice printing
    varf=0.
    do ivar=1,nvar
       iz=0
       do i=ivar,neq,nvar
          iz=iz+1
          varf(iz,ivar)=var(i)
       enddo
       ixloc(ivar)=MAXLOC(ABS(var),MASK=variable.eq.ivar,dim=1)
       varloc(ivar)=MAXLOC(ABS(varf(:,ivar)),DIM=1)
       varmax(ivar)=varf(varloc(ivar),ivar)
    enddo

    ! Nice printing
    !!    print 102,'Variations relatives: ',nameofvar(ivartop),var2(ivartop)
    if (verbose.ge.-10) print *,'Collective variation:',varcontrol
    if (verbose.ge.-10) print 102,' Top individual variation : ',nameofvar(variable(ivartop)),zone(ivartop),var(ivartop)
    dnameofvar=' d'//nameofvar
    ivarmax=MAXLOC(ABS(varmax),DIM=1)
    dnameofvar(ivarmax)='*D'//nameofvar(ivarmax)
    ichimmax=MAXLOC(ABS(varmax(nvarh+1:nvar)),DIM=1)
    dnameofvar(nvarh+1)=dnameofvar(ichimmax+nvarh)
    varmax(nvarh+1)=varmax(ichimmax+nvarh)
    varloc(nvarh+1)=varloc(ichimmax+nvarh)
    ixloc(nvarh+1)=ixloc(ichimmax+nvarh)
    nselect=0
    do ivar=1,nvarh+1
       if (choicet(ivar).or.ivar==nvarh+1) then
          nselect=nselect+1
          dnameofvar(nselect)=dnameofvar(ivar)
          varloc(nselect)=varloc(ivar)
          varmax(nselect)=varmax(ivar)
          ixloc(nselect)=ixloc(ivar)
       endif
    enddo
    if (verbose.ge.-10) print 100,(dnameofvar(i),varloc(i),varmax(i),i=1,nselect)
    if (verbose.ge.-10) print 99,(dnameofvar(i),varloc(i),varmax(i),&
         xold(ixloc(i)),x(ixloc(i)),xscale(ixloc(i)),&
         i=1,nselect)
    ! New time step you want
!!!    dt_want=dt*vardt*varrel/MAX(vardt*varcontrol,varrel)
    dt_want=dt*varrel/varcontrol
    clear=.false.
    dt_want=MIN(dt_want,dt2*dt)
    ! Clear advanced solution dor next time-step if too big time-step
    if (dt_want.lt.(dt1*dt).and.dt1<dt2) clear=.true.
!!$    if (dt_want.lt.(dt1*dt).and.k.gt.1) then
!!$       toomuch=.true.
!!$       return
!!$    endif
    dt_want=MAX(dt_want,dt1*dt)
    dt_want=MIN(dtmax,dt_want)

    ! Test if accretion will be triggered, because we need to
    ! lower dt:
    if (time+dt<tcool.and.dt_want>1.d-4*taccretion.and.time+dt+dt_want>tcool) then
       clear=.true.
       dt_want=max(.5*(tcool-time-dt),1.d-4*taccretion)
    endif
    ! Reject time step if variation is too much :
!!$    if (varcontrol.gt.varrel/(.01+dt1).and.k.gt.1) then
!!$       toomuch=.true.
!!$       dt_want=dt*varrel/varcontrol
!!$    else
!!$       toomuch=.false.
!!$    endif
    toomuch=.false. ! toomuch is not used
99  FORMAT (10(1(a8,'(',i4,') dx=',1pe12.3,' xold=',1pe12.3,' x=',1pe12.3,' xscale=',1pe12.3),/))
100 FORMAT (10(3(a8,'(',i4,'):',1pe12.3,'  '),/))
101 FORMAT (a8,'(',i4,'):',1pe12.3)
    !!102 FORMAT (a,a7,':',1pe12.3)
102 FORMAT (a,a7,'(',i4,'):',1pe12.3)
103 FORMAT ('dres=',1pe12.3,'(',i4,')  Sum(dres)=',1pe12.3,&
         ' dheat=',1pe12.3,'(',i4,')  Sum(dheat)=',1pe12.3)
    if (verbose.ge.-10) print 103,varres(MAXLOC(ABS(varres))),MAXLOC(ABS(varres)),SUM((ABS(varres))),&
         varheat(MAXLOC(ABS(varheat))),MAXLOC(ABS(varheat)),SUM((ABS(varheat)))
  end SUBROUTINE TIMESTEP

  SUBROUTINE DEBUG
    ! Useful debugging stuff to check that the matrix is correct
    implicit none
    double precision,allocatable,dimension(:,:)::Ad
    integer::m1d,m2d,ndiagd,mmd
    double precision::comparaison=0.!1.e-14
!!$    !LOAD DEBUG
!!$    print*,'lecture en cours..'
!!$    OPEN(unit=10,status='unknown',file='matrice')
!!$    read(10,*)m1d
!!$    read(10,*)m2d
!!$    read(10,*)Bsave
!!$    allocate(Ad(m1d+m2d+1,neq))
!!$    read(10,*)Ad
!!$    CLOSE(10)
!!$    print*,'lecture termine..'
    print*,'lecture en cours.. matrix1'
    OPEN(unit=10,status='unknown',file='matrix1')
    read(10,*)m1d
    read(10,*)m2d
    read(10,*)Bsave
    allocate(Ad(m1d+m2d+1,neq))
    read(10,*)Ad
    CLOSE(10)
    print*,'lecture termine..'
    print*,'lecture en cours..matrix2'
    OPEN(unit=10,status='unknown',file='matrix2')
    read(10,*)m1d
    read(10,*)m2d
    read(10,*)B
    read(10,*)A
    CLOSE(10)
    print*,'lecture termine..'

!!$
!!$
!!$          !COMPUTE DEBUG
!!$          print*,'Compute the hard way...'
!!$          Bsave=B
!!$          m1d=m1
!!$          m2d=m2
!!$          allocate(Ad(m1d+m2d+1,neq))
!!$          Ad=A
!!$          toto=.true.
!!$          allnum=.true.
!!$          call setEQU  ! initialise les equations
!!$          B = - EQU 
!!$          pskip=200
!!$          CALL setmatrix
!!$          print*,'... true matrix computed.'
!!$
!!$          !SAVE DEBUG
!!$          print*,'sauvegarde en cours..'
!!$          OPEN(unit=10,status='unknown',file='matrice')
!!$          write(10,*)m1
!!$          write(10,*)m2
!!$          write(10,*)B
!!$          WRITE(10,*)A
!!$          CLOSE(10)
!!$          print*,'sauvegarde termine..'
!!$          return
    !DEBUG
    do while(comparaison.ge.0.) 
    print*,'Dtection fixe :'
    READ(*,*)comparaison
    WHERE(B.ne.0..or.Bsave.ne.0.)
       compareB=ABS(B-Bsave)/(ABS(B)+ABS(Bsave))
    ELSEWHERE
       compareB=0.
    ENDWHERE
    if (ANY(compareB.gt.comparaison)) then
       print*,'Il y a du grabuge pour B:'
       do i=1,neq
          if (compareB(i).gt.comparaison)then
             print 665,nameofequ(variable(i)),zone(i)&
                  ,Bsave(i),B(i),compareB(i)
          endif
       enddo
       read(*,*)
    else
       print*,'NO PROBLEM pour B !'
    endif

665 FORMAT(a7,'(',i3,') Bsave=',1pe10.3,' B=',1pe10.3,' delta=',1pe10.3)
!!$          WHERE(A.ne.0..or.Asave.ne.0.)
!!$             compare=ABS(A-Asave)/(ABS(A)+ABS(Asave))
!!$          ELSEWHERE
!!$             compare=0.
!!$          ENDWHERE
!!$          if (ANY(compare.gt.comparaison)) then
!!$             print*,'Il y a du grabuge pour A:'
    ndiagd=m1d+m2d+1
    do i=1,neq
       do mmd=1,ndiagd
          j=m1d+1+i-mmd
          if (j.ge.1.and.j.le.neq) then
             mm=m1+1+i-j
             if (mm.ge.1.and.mm.le.ndiag) then
                if (A(mm,j).ne.0..or.Ad(mmd,j).ne.0.) then
                   compare(mm,j)=ABS(A(mm,j)-Ad(mmd,j))/&
                        (ABS(A(mm,j))+ABS(Ad(mmd,j)))
                else
                   compare(mm,j)=0.
                endif
                if (compare(mm,j).gt.comparaison)then
                   print 666,nameofequ(variable(i)),zone(i),nameofvar(variable(j)),zone(j)&
                        ,Ad(mmd,j),A(mm,j),compare(mm,j)
                endif
             else
                if (Ad(mmd,j).ne.0.) then
                   print 666,nameofequ(variable(i)),zone(i),nameofvar(variable(j)),zone(j)&
                        ,Ad(mmd,j),0.,666.
                endif
             endif
          endif
       enddo
    enddo
    enddo
    STOP
!!$          else
!!$             print*,'NO PROBLEM pour A !'
!!$          endif
    deallocate (Ad)
    STOP
666 FORMAT(a7,'(',i3,')/',a7,'(',i3,') Asave=',1pe10.3,' A=',&
         1pe10.3,' delta=',1pe10.3)

  END SUBROUTINE DEBUG


  double precision FUNCTION norm(x,ivar,n)
    ! Get norm n of variable ivar in vector x
    double precision,dimension(neq)::x
    integer ivar,n
    integer i

    if (n==0) then
       norm=0.
       do i=ivar,neq,nvar
          norm=MAX(norm,ABS(x(i)))
       enddo
       return
    endif

    norm=0.
    do i=ivar,neq,nvar
       norm=norm+x(i)**n
    enddo
    norm=norm**(1./float(n))
  end FUNCTION norm

  double precision FUNCTION minv(x,ivar)
    ! Get minimum non zero value of variable ivar in x
    double precision,dimension(neq)::x
    integer ivar
    integer i
    minv=0.
    do i=ivar,neq,nvar
       if (minv==0..and.x(i).ne.0.) minv=ABS(x(i))
       if (x(i).ne.0.) minv=MIN(minv,ABS(x(i)))
    enddo
    return
  end FUNCTION minv

  subroutine vtypical
    ! Compute typical values of each variables:
    ! for derivating and control purposes
    use equations,only:Ltypold,Ftypold
    implicit none
    integer::i
    double precision,dimension(nz)::local
    if (verbose.ge.0) print*,'vtypical'
    ! Local typical values for control
    xscale=xtypical(variable)
!    return
    do i=1,nz
       xscale((i-1)*nvar+iL)=1.!&
       !             (Lold(i)-2.*MIN(MINVAL(Ltotold(1:nz)),-1.))/sqrt(1.+Ltotold(i)**2)
       !          xscale((i-1)*nvar+iL)=sqrt((Uold(i)*dmold(i)/dt)**2+x((i-1)*nvar+iL)**2) !&       
       local(i)=sqrt((Ltypold(i))**2+x((i-1)*nvar+iL)**2) 
    enddo
    do i=1,nz
       xscale((i-1)*nvar+iL)=sqrt(0*local(max(1,i-1))**2+&
            local(i)**2+0*local(min(i+1,nz))**2)
    enddo
!    return
!!$    if (checkl=='log') then
!!$       do i=1,nz
!!$          xscale((i-1)*nvar+iL)=ABS(xold((i-1)*nvar+iL))+xtypical(iL)   
!!$       enddo
!!$    endif
!!$
    ! Values for derivatives of fluxes and composition           
    do i=1,nz
       do j=1,species
          ij=nvar*(i-1)+j+nvarh
          ij=ij+species
          xscale(ij)=1.                 
          if (j==iprim(a23d).or.j==iprim(a23m)) then
             xscale(ij)=1.!e-8+x(ij)
          else
             xscale(ij)=1.
          endif
          !             xscale(ij-species)=sqrt((dmold(i)/dt*xscale(ij))**2+x(ij-species)**2)
          !          xscale(ij-species)=1.
          xscale(ij-species)=sqrt(xscale(ij)**2+x(ij-species)**2)
       enddo
    enddo
  end subroutine vtypical

  subroutine invertsys
    ! Invert matrix according to the selected method of matrix inversion
    if (verbose.ge.0) print*,'invertsys'
    B0=B
    b=0.
    singular=.false.
    if (JPC=='sqr') then
       CALL numrec_SOLVE(Asquare,B,B0,neq,INFO)    
       !           CALL SQUARE_SOLVE(Asquare,B(isqr:jsqr),B0(isqr:jsqr),nsqr,INFO,invert)    
       if (info>0.and.info<=neq) print*,'Corresponds to :',nameofvar(variable(info)),'(',zone(info),')'
    else
       if (JPC=='non') then
          CALL SOLVE(A,B,B0,neq,INFO,invert)    
          if (info.ne.0) then 
             print*,'Singular matrix'
             if (info.le.neq) then
                print*,'Corresponds to :',nameofvar(variable(info)),'(',zone(info),')'
             else 
                print*,'info=',info,' matrix is not invertible at machine precision ?'
             endif
             singular=.true.
             b=0.
          endif
       else
          if (JPC=='spa') then
             if (ANY(A(m1+1,:).eq.0.))then
                do i=1,neq
                   if (A(m1+1,i).eq.0.) then
                      print*,zone(i),nameofequ(i),'/',nameofvar(i),'=0'
                   endif
                enddo
             endif
             CALL solve_sparse(a,b,b0,neq,info)
          else
             CALL SOLVE_JPC(A,B,B0,neq,m1,m2,A_jpc,A0,A_L,index,SA,one,INFO,invert)
          endif
       endif
    endif
  end subroutine invertsys

  subroutine Bdomain
    ! Constrains N-R correction according to definition domain
    if (verbose.ge.0) print*,'Bdomain'
    ! Respect of (non zero) positive variables
    WHERE(defpos(variable)) 
       B=max(B,-.999*x/xscale)
    endWHERE
    do j=1,species
       do i=1,nz
          ii=(i-1)*nvar+nvarh+j+species
          if (x(ii)+xscale(ii)*B(ii).lt.tiny) then
             B(ii)=-x(ii)/xscale(ii)
             if (idx==2) B(ii-species)=-x(ii-species)/xscale(ii-species)
          endif
       enddo
    enddo

    ! Put flux and composition variables to 0 if they are less than tiny
    WHERE(variable.gt.nvarh & !.and.variable.le.nvarh+species&
         .and.ABS(x+xscale*B).lt.tiny) B=-x/xscale

    WHERE(varpos(variable)) 
       B=max(B,-x/xscale)
    endwhere

! If you really need to uncomment the following lines, 
! I suspect the code is really doing
! something wrong..
    ! Respect of <=1. variables
!!$    WHERE(variable.gt.nvarh+species.and.x+xscale*B.gt.1.) B=(1.-x)/xscale
    ! check the for sum of the mass fractions to be less than 1.
!!$    do i=1,nz
!!$       massfrac=0.
!!$       dmassfrac=0.
!!$       do j=1,species
!!$          ij=nvar*(i-1)+nvarh+j
!!$          massfrac=massfrac+x(ij)
!!$          dmassfrac=dmassfrac+xscale(ij)*B(ij)
!!$       enddo
!!$       if (massfrac+dmassfrac.gt.1.) then
!!$          do j=1,species
!!$             ij=nvar*(i-1)+nvarh+j
!!$             B(ij)=(1.-massfrac)/dmassfrac*B(ij)
!!$          enddo
!!$       endif
!!$    enddo

  end subroutine Bdomain

  subroutine xdomain
    implicit none
    ! Constrains vector x according to definition domain
    integer i
    if (verbose.ge.0) print*,'xdomain'
    xsave2=x
    where (defpos(variable).and.x.le.0.) x=.999*xold
    WHERE (varpos(variable).and.x<0.)   x=xold
!    where (variable.gt.nvarh+species.and.x.lt.tiny*xscale) x=0.
    ! (tiny to be improved, as flux must be put to zero as well...)
!!$    where (variable.gt.nvarh+species.and.x.gt.1.) x=1.
    ! SHOULD CHECK BETTER for x>1
    dx=x-xold
    if (verbose.ge.-3) then
       do i=1,neq
          if (x(i).ne.xsave2(i)) then
             print*,nameofvar(variable(i)),zone(i),'x=',xsave2(i),'xder=',xder(i)
          endif
       enddo
    endif
  end subroutine xdomain

  subroutine ajust
    ! Select between modes of adjusting N-R shoot
    ! (improves convergence ? Difficult to say yes in all cases...)
    if (verbose.ge.0) print*,'ajust'
    if (abs(itermax).gt.1) then
       select case(najust)
       case(0)
          coeff=1.
          out=.false.
       case(1)
          CALL ajust1(coeff,out)
       case(2)
          CALL ajust2(coeff,out)
       case(3)
          CALL ajust3(coeff,out)
       case(4)
          ! Same method as in STAR code
          if ((iiter.gt.3).and.(errcontrol.ge.(0.99*errpr))) reduce=.true.
          if (reduce) then
             if (errpr.gt.errcontrol) then
                coeff=min(1.,1.05*coeff)
             else
                coeff=max(0.3,0.3*coeff)
             endif
          else
             coeff=1.e-2/max(errcontrol,1.e-2)
          endif
       end select
       errcontrol=errcontrol
    else 
       coeff=1.
    endif
  end subroutine ajust

  subroutine stoptests
    ! stop tests
    use variables,only:lnT
    use equations,only:tburn,tburn2,tconv
    use maths,only:smax
    use constants,only:Msun,Rsun,cln10
    logical::tropvite,tropvieux,tropdense,tropdt,toocold,toohot,explode
    logical::toomass,toomassc

    if (verbose.ge.0) print*,'stoptests'
    toomass=(M(1)/Msun.gt.Masslim.and.Masslim.gt.0.).or.&
         (M(1)/Msun.lt.ABS(Masslim).and.Masslim.lt.0.)
    toomassc=(m(nz)/msun.gt.mcentre.and.mcentre.gt.0.).or.&
         (m(nz)/Msun.lt.abs(mcentre).and.mcentre.lt.0.)
    tropvieux=(time.gt.tstop.and.tstop.gt.0.) 
    toocold=(Tlim.lt.0.).and.MAXVAL(lnT).lt.abs(Tlim)*CLN10
    toohot=(Tlim.gt.0.).and.MAXVAL(lnT).gt.abs(Tlim)*CLN10
    tropdt=dt.ge.dtmax
    explode=ANY(tburn.lt.tconv.and.tburn.gt.0.)
    if (explode.or.tropvieux.or.tropvite.or.toocold.or.toohot&
         .or.tropdt.or.toomass.or.toomassc) then
       CALL IMP(0)
       CALL output
    endif
    if (explode) then
       print*,'Thermonuclear runaway !'
       print*,'log(Tc)=',lnT(nz)/CLN10
       print*,'log(rhoc)=',alog10(rho(nz))
       print*,'log(Te)=',lnT(1)/CLN10
       print*,'log(L_r)=',alog10(L(1)/Lsun)
       print*,'Position of the flame:'
       print*,'mflame=',smax(tburn2*(tconv-tburn),m)/Msun,' Msun'
       print*,'mflame=',smax(tburn2*(tconv-tburn),r)/Rsun,' Rsun'
       STOP
    endif
    if (toocold) then
       print*,'Central temperature too low, I stop there'
       print*,'log(Tc)=',lnT(nz)/CLN10,'  log(Tc limit)=',-Tlim
       print*,'log(rhoc)=',alog10(rho(nz))
       print*,'log(Te)=',lnT(1)/CLN10
       print*,'log(L_r)=',alog10(L(1)/Lsun)
       STOP
    endif
    if (toohot) then
       print*,'Central temperature too high, I stop there'
       print*,'log(Tc)=',lnT(nz)/CLN10,'  log(Tc limit)=',Tlim
       print*,'log(rhoc)=',alog10(rho(nz))
       print*,'log(Te)=',lnT(1)/CLN10
       print*,'log(L_r)=',alog10(L(1)/Lsun)
       STOP
    endif
    if (toomassc) then
       print*,'Too massive at centre: Istop here'
       print*,'Massc=',M(nz)/Msun,' Msun  Mass limit=',Mcentre
       STOP
    endif
    if (toomass) then
       print*,'Mass=',M(1)/Msun,' Msun  Mass limit=',Masslim
       print*,'Save in file=','MASS'//numero(nbmass)
       OPEN (80,FILE = 'MASS'//numero(nbmass),STATUS = 'unknown')
       CALL outputhydro
       CLOSE(80)
       nbmass=nbmass+1
       Masslim=Masslim+.1*sign(1.,Masslim)
       print*,'I carry on with Masslim=',Masslim
    endif
    if (tropvieux) then
       print*,'TEMPS=',time/csecyr,' year(s): I stop here...'
       print*,'tstop=',tstop/csecyr,' year(s) (=',tstop,' program units)'
       STOP
    endif
    if (tropdt) then
       print*,'DT=',dt/csecyr,' year(s): it should be in equilibrium now...'
       print*,'dtstop=',dtmax/csecyr,' year(s)'
       STOP
    endif
  end subroutine stoptests

  subroutine allprints
    ! Do all printings
    if (verbose.ge.0) print*,'allprints'    
    ! Time to print the survey
    if (FOLLOW.ne.'non') CALL TRACE

    ! Print if it is time to
    if (MOD(k,nimp)==0) then
       CALL IMP(0)
       CALL output
    endif

    ! Photograph time...
    nstep=k+nstep_init
    clicclac=.false.
    if (tphoto.eq.0.) then
       clicclac=nstep.gt.nphoto*photostep
    endif
    if (tphoto.ne.0.) then
       if (photostep==0) then
          clicclac=time.gt.float(nphoto)*tphoto
       else
          clicclac=time.gt.tphoto*10.**(float(nphoto-1)/float(photostep))
       endif
    endif
    if (clicclac) then
       ! do it only if it is not already done...
       if (MOD(k,nimp).ne.0) then
          CALL IMP(0)
          CALL output
       endif
       ! ajust counters for next picture
       nphoto=nphoto+1
       CALL photofilename
    endif
  end subroutine allprints

  subroutine sizeB
    ! Size N-R correction
    if (verbose.ge.0) print*,'sizeB'
    ! Convergence test is on non zero positive variables
    choice=.true.
    if (idx==2.and.(CHECKG.ne.'n-r'.and.CHECKG.ne.'yes')) &
         choice(nvarh+1:nvarh+species)=.false.
    if (CHECKL.ne.'yes'.and.CHECKL.ne.'n-r') choice(iL)=.false.
!!$        choice(idm)=.false.
!!$        choice(ins)=.false.

    ! Relative error for chosen variables
    WHERE(choice(variable))
       errel=B
    ELSEWHERE
       errel=0.
    ENDWHERE
    ident=MAXLOC(ABS(errel),DIM=1)
    checkcoeff=1./MAX(MAXVAL(ABS(errel)),1.d-10)
    errpr=errcontrol
    errcontrol=asum*SUM(ABS(errel),DIM=1)/float(neq)+&
         amax*MAXVAL(ABS(errel))
  end subroutine sizeB

  subroutine sizequ
    ! Size of the equation vector
    double precision,dimension(neq)::equn
    if (verbose.ge.0) print*,'sizequ'
    ! Size the equations
    ! Chose the variables you want to control the equations of
    choicequ=.true.
!!$    choicequ=.false.
!!$    choicequ(iL)=.true.
!!$    choicequ(nvarh+1:nvarh+species)=.true.
    !        choicequ(ins)=.false.
    !        choicequ(idm)=.false.
    ! Print info on equations:
    if (verbose.ge.-4) call top10(equ,choicequ,'ec')
    ! Size the equations
    maxequ=MAXVAL(ABS(equ),mask=choicequ(variable),dim=1)
    identequ=MAXLOC(ABS(equ),mask=choicequ(variable),dim=1)
    equpr=tolsqr
    tolsqr=MAXVAL(ABS(equ),mask=choicequ(variable).and.&
         (zone.lt.zone(identequ)-1.or.zone.gt.zone(identequ)+1),dim=1)
    tolsqr=amax*maxequ+asum*SUM(ABS(equ),mask=choicequ(variable))/float(neq)
    if (verbose.ge.-10)print*,'Maxequ=',maxequ,' tolsqr=',tolsqr
  end subroutine sizequ

  subroutine top10(x,mask,name)
    ! Print info on top 10 values fo vector x
    character(*)::name
    double precision,dimension(neq)::x,xn
    logical,dimension(nvar)::mask
    integer::ident
    xn=x
    do i=1,10
       ident=MAXLOC(ABS(xn),mask=mask(variable),dim=1)
       print 402,nameofequ(variable(ident)),nameofvar(variable(ident)),zone(ident),name,x(ident)
       xn(ident)=0.
    enddo
402 FORMAT(2x,'equ',a7,' var',a7,' (',i4,')',a7,'=',1pe10.3)
  end subroutine top10
end PROGRAM newton


