! routine nucrat.f from STAR program
MODULE nuclear
  use dimensions,only:nz,nreacmax,verbose
  implicit none

  double precision,dimension(nz,nreacmax)::rate
  double precision,dimension(nz)::EX, ENX
  double precision,dimension(nz)::Qcc,eUrca,eqUrca,dmuUrca
  double precision,dimension(nz)::XLurca,Curca,Hurca
contains
  SUBROUTINE NUCRAT(TL)        
    ! Compute rates of the list of nuclear reactions given in 'chemistry-data'
    ! (this is read in chemistry.f90) and their corresponding
    ! energy and neutrino release
    use parameters,only:INUC
    use EOS,only:zt,psi,muip,firstguess,ne!,T,mui,cp,delta
    use secondaries,only:rho,na,cp,delta,T,mui,mue,sni
    use abundances,only:neo,nio,nzz,avm,xa!,na
    use maths,only:cbrt
    use constants,only:CPI, PI4, CLN10,CPL, CMEVMU,CHE,CT,BOLTZM,EVOLT,CR,AMU
    use chemistry
    implicit none

    ! Detect the first call to initialise the screening parameters.
    logical,save::first=.true.

    ! input variables
    double precision,dimension(nz)::tl

    ! local variables
    integer i,j,k
    double precision,dimension(nz,nreacmax)::rrt
    double precision,dimension(nz)::rhb
    double precision,dimension(nz)::wc,wb,wa,xb,vl,za,zb,zc,zd,ab,ac,ad,zw
    double precision,dimension(nz)::tf,rn,tt,tu,rr
    integer,dimension(nz)::it
    logical,dimension(nz)::ok1,ok2
    double precision,dimension(nz)::scrn,strn,dstr
    double precision,dimension(nz)::t6r,fccg,f34,pp2,pp3
    double precision::fpng,qpp
    double precision,dimension(nz)::qnpp
    double precision,dimension(nz)::smu,mu

    ! screening
    double precision,dimension(nreacmax)::CZA, CZB, CZC, CZD
    double precision,parameter::cxa=5./3.,cxc=cxa-1.,cxb=2.*cxc,cxd=1.86
    double precision,dimension(ndim)::vz

    ! data 
    integer idum
    double precision,parameter::CSA=0.624, CSB=0.316, CSC=0.460, CSD=0.38, CSXD=0.86

    ! Reading variables
    character*7::readname
    character*1::readtype
    double precision::readQnt
    double precision::readb1,readb2,readb3,readb4
    integer::readn1,readn2
    integer::readndata
    double precision::readczw
    double precision,dimension(200)::readcrt
    double precision,dimension(:,:),allocatable::readt2,readt3

    ! Nuclear rates data
    character*1,dimension(nreacmax)::ratype
    integer::j1,j2
    integer,save,dimension(nreacmax)::jtable
    double precision,save,dimension(200,nreacmax)::crt
    integer,parameter::nurcamax=6,nx1max=200,nx2max=20
    double precision,save,dimension(nx1max,nx2max,nurcamax)::crt2,crt3
    double precision,save,dimension(nreacmax,4)::bound
    integer,save,dimension(nreacmax,2)::ndata=200
    logical,save,dimension(nreacmax)::found
    double precision,save,dimension(nreacmax)::czw
    ! Neutrino Qvalues
    double precision,save,dimension(nreacmax)::qnt

    ! Interpolation variables
    integer,dimension(nz)::i1,i2
    double precision,dimension(nz)::x1,x2
    double precision,dimension(nz,nurcamax)::UrcaQ

    if (verbose.ge.2) print*,'  nucrat'
    ! If it is the first call to this routine, then read the data
    if (first) then
       first=.false.
       ! initialize some screening parameters.
       DO J = 1, nreac
          CZA(J) = (Zreac(J,1)+Zreac(J,2))**CXA - Zreac(J,1)**CXA - ZREAC(J,2)**CXA
          CZB(J) = (ZREAC(J,1)+ZREAC(J,2))**CXB - ZREAC(J,1)**CXB - ZREAC(J,2)**CXB
          CZC(J) = -((ZREAC(J,1)+ZREAC(J,2))**CXC - ZREAC(J,1)**CXC - ZREAC(J,2)**CXC)
          CZD(J) = (ZREAC(J,1)+ZREAC(J,2))**CXD - ZREAC(J,1)**CXD - ZREAC(J,2)**CXD
       END DO
       do i=1,ndim
          vz(i)=iz(i)**(3*cxd-4)
       enddo

       ! Read the nuclear rates tables
       print*,'NUCRAT: Reading nuclear rates-data...'
       open(unit=17,file='rates-data',status='unknown')
       readname=''
       found=.false.
       j1=0
       j2=0
       do while (readname.ne.'END')
          read(17,'(a7,a1)')readname,readtype
          if (TRIM(readname).ne.'END') then
             if (readtype.eq.'u') then
                read(17,*)readb1,readb2,readb3,readb4,readczw
                read(17,*) readn1,readn2
                allocate(readt2(readn1,readn2),readt3(readn1,readn2))
                read(17,*)readt2
                read(17,*)readt3
                readt2=readt2
             else
                read(17,*)readb1,readb2,readQnt,readczw
                read(17,*) readndata
                read(17,*)readcrt(1:readndata)
             endif
             j=0
             do i=1,nreac
                if (rname(i)==readname) j=i
             enddo
             if (j.ne.0) then
                if (found(j)) then
                   print*,'WARNING ! I already found reaction ',readname,' before...'
                   print*,'The last found is the one I retain...'
                endif
                found(j)=.true.
                print*,'I found reaction ',readname,'. It will bear number ',j
                ratype(j)=readtype
                if (readtype.eq.'u') then
                   j2=j2+1
                   if (j2.gt.nurcamax) then
                      print*,'Make more space for urca reactions.'
                      print*,'At the present time, I cannot handle more than nurcamax=',nurcamax
                      STOP
                   endif
                   if (readn1.gt.nx1max) then
                      print*,'Make more space for urca reactions.'
                      print*,'At the present time, nx1max=',nx1max,'.'
                      print*,'You want at least :',readn1
                      STOP
                   endif
                   if (readn2.gt.nx2max) then
                      print*,'Make more space for urca reactions.'
                      print*,'At the present time, nx2max=',nx2max,'.'
                      print*,'You want at least :',readn2
                      STOP
                   endif
                   jtable(j)=j2
                   bound(j,1)=readb1  
                   bound(j,2)=readb2  
                   bound(j,3)=readb3  
                   bound(j,4)=readb4
                   Qnt(j)=0.
                   czw(j)=readczw
                   ndata(j,1)=readn1
                   ndata(j,2)=readn2
                   crt2(1:readn1,1:readn2,j2)=max(-99.,readt2)
                   crt3(1:readn1,1:readn2,j2)=readt3
                   deallocate(readt2,readt3)
                else
                   j1=j1+1
                   jtable(j)=j1
                   bound(j,1)=readb1  
                   bound(j,2)=readb2  
                   Qnt(j)=readQnt
                   czw(j)=readczw
                   ndata(j,1)=readndata
                   crt(1:ndata(j,1),j1)=readcrt(1:readndata)
                endif
             else
                if (readtype.eq.'u') then
                   deallocate(readt2,readt3)
                endif
             endif
          endif
       enddo
       if (ANY(.not.found(1:nreac))) then
          print*,'I miss a few reactions in file rate-data, they are :'
          do j=1,nreac
             if (.not.found(j)) print*,rname(j)
          enddo
          STOP
       endif
       print*,'... reading completed.'
    endif

    ! RHB is ''baryon density'': 1 amu * number of baryons per cm3
    RHB = RHO/AVM
    ! Electron screening theory from Graboske, DeWitt, Grossman & Cooper (1973),
    ! for strong (ZA, ZB, ZC) are intermediate screening (ZD). The reaction
    ! dependent charge parameters are stored in CZA ... CZD.
    WC = 0.0
    DO I = 1, ndim
       if (istatus(i).ne.' ') WC = WC + NA(:,I)*VZ(I)
    END DO
    WC = WC/NIO
    WB = NEO/NIO
    WA = ZT*ZT/(WB*WB)
    XB = CBRT(ABS(WB))
    VL = CPL*SQRT(ABS(NIO)*RHB*EXP(-3.0D0*TL))
    ZA = CSA*XB*EXP(LOG(VL)/1.5D0)
    ZB = CSB*XB*ZA
    ZC = CSC/(XB*XB)
    ZD = CSD*WC*WA*EXP(LOG(VL/(WA*ZT))*CSXD)
    ! weak screening
    zw = 0.5*zt*vl
    ! Reaction rates interpolated in T, mostly from Caughlan & Fowler (1988)
    ! Units of rrt are ? nb of reactions per nb of reactants ?
    TF = TL/CLN10
    urcaQ=0.
    DO J = 1, nreac
       if (j.ne.irpng) then ! We assume fpng=8.e-4 in the following !
          RN = 0.0D0
          if (ratype(j).eq.'u') then
!!$             print*,rname(j)
             ! urca rates
             j2=jtable(j)
             smu=0.
             do k=1,5
                if (rstatus(ireac(j,k)).ne.'c'.and.&
                     rstatus(ireac(j,k)).ne.'C') then
                   if (ireac(j,k).ne.0) smu=smu+stoech(j,k)*iz(ireac(j,k))
                   !mui(:,ireac(j,k))
                endif
             enddo
             smu=smu*mue
             !          x1=psi*exp(tl)*BOLTZM/ABS(Qvalue(j))/EVOLT/1e6
             !             x1=mue/cmevmu/abs(Qvalue(j))
             x1=ABS(smu)/cmevmu/abs(Qvalue(j))
             if (j.eq.iup23) dmuUrca=ABS(smu)/cmevmu-ABS(Qvalue(j))
             if (j.eq.iup23.and.verbose.eq.-2) then
                print*,'   rhb=',rhb(nz)
                print*,'   smu=',smu(nz)
                print*,'   x1 =',x1(nz)
                print*,'   x2 =',x2(nz)
             endif
             x2=1./CT/exp(tl)
             x1=float(ndata(j,1))*(x1-bound(j,1))/(bound(j,2)-bound(j,1))+1.
             x2=float(ndata(j,2))*(x2-bound(j,3))/(bound(j,4)-bound(j,3))+1.
!!$             x1=MIN(float(ndata(j,1)),MAX(1.,x1))
!!$             x2=MIN(float(ndata(j,2)),MAX(1.,x2))
             ok1=.true.!(x1.ge.1.).and.(x2.ge.1.)
             where(ok1)
                i1=max(1,min(ndata(j,1)-1,int(x1)))
                i2=max(1,min(ndata(j,2)-1,int(x2)))
                x1=x1-i1
                x2=x2-i2
             endwhere
             do i=1,nz
                if (ok1(i)) then
                   rr(i)=(1.-x1(i))*(1.-x2(i))*crt2(i1(i)  ,i2(i) ,j2)+&
                        x1(i)     *(1.-x2(i))*crt2(i1(i)+1,i2(i)  ,j2)+&  
                        (1.-x1(i))*x2(i)     *crt2(i1(i)  ,i2(i)+1,j2)+&  
                        x1(i)     *x2(i)     *crt2(i1(i)+1,i2(i)+1,j2)
                   UrcaQ(i,j2)=(1.-x1(i))*(1.-x2(i))*crt3(i1(i)  ,i2(i)  ,j2)+&
                        x1(i)     *(1.-x2(i))*crt3(i1(i)+1,i2(i)  ,j2)+&  
                        (1.-x1(i))*x2(i)     *crt3(i1(i)  ,i2(i)+1,j2)+&  
                        x1(i)     *x2(i)     *crt3(i1(i)+1,i2(i)+1,j2)
                endif
             enddo
             if (j.eq.iup23.and.verbose.eq.-2) print*,'rr=',rr(nz)
          else
             ! normal rates
             j1=jtable(j)
             if (ratype(j).eq.'p') then
                ! Argument is density for pycnonuclear rates
                x1=log(rho)
             else
                ! Argument is temperature for others
                x1=tf
             endif
             TT = float(ndata(j,1))*(x1-bound(j,1))/(bound(j,2)-bound(j,1))+1.
             ok1=( TT .GE. 1.0D0 ) 
             where (ok1)
                IT = MAX(1, MIN(ndata(j,1)-1, INT(TT)))
                TT = TT - IT
                TU = 1.0D0 - TT
                RR = TU*CRT(IT,j1) + TT*CRT(IT+1,j1)
             endwhere
          endif

          ! screening
          ok2=.true.!( RR .GE. -98.0D0 )!.true.
          where(ok1.and.ok2)
             SCRN = ZD*CZD(J)
             STRN = ZA*CZA(J) + ZB*CZB(J)
             DSTR = ZC*CZC(J)
          endwhere
          where(ok1.and.ok2.and. DSTR .LT. 0.29*STRN) 
             SCRN = MIN(SCRN, STRN - DSTR)
          endwhere

          ! weak screening
          if (inuc.ge.10) then
             where(ok1.and.ok2)
                scrn = zw*czw(j)
             end where
          end if

          ! Screening is comprised in pycnonuclear reactions ? See Langanke ?
          if (ratype(j)=='p') scrn=0.
          ! No screening for Urca either
          if (ratype(j)=='u') scrn=0

          ok2=( RR+SCRN .GE. -98.0D0 )
          where(ok1.and.ok2)
             RN = EXP(min(150.,CLN10*(RR + 20.0D0) + SCRN))*1.0D-20
          endwhere
          RRT(:,J) = RN
          if (j==iup23.and.verbose.eq.-2) print*,'rrt=',rrt(nz,j)
       endif
    END DO
!!$    read(*,*)
    if(mod(inuc,10).eq.1)then
       ! correct rates to simulate Bahcall (1992) cross sections
       rrT(:,irpp) = rrT(:,irpp)*0.9828
       rrT(:,ir33) = rrT(:,ir33)*0.971
       rrT(:,ir34) = rrT(:,ir34)*0.987
       t6r = 10.0**(0.5*tf - 3.0)
       rrT(:,irbe) = 5.54d-9/t6r*(0.936 + 0.004*t6r*t6r)
       rrT(:,irbp) = rrT(:,irbp)*0.933
    else if(mod(inuc,10).eq.2)then
       ! idem, for Bahcall (1995) cross sections
       rrT(:,irpp) = rrT(:,irpp)*0.9557
       rrT(:,ir33) = rrT(:,ir33)*0.969
       rrT(:,ir34) = rrT(:,ir34)*0.970
       t6r = 10.0**(0.5*tf - 3.0)
       rrT(:,irbe) = 5.54d-9/t6r*(0.936 + 0.004*t6r*t6r)
       rrT(:,irbp) = rrT(:,irbp)*0.933
       rrT(:,irpn) = rrT(:,irpn)*0.991
    end if
    ! Multiply with density and abundances to get rates per baryon per second,
    ! note that abundances of He3 and Be7 are not needed in equilibrium
    do j=1,nreac
       if (ratype(j).ne.'b') then
          if (j.eq.irbe)then !electron capture on Be7
             rate(:,j)=rrt(:,j)*rhb*neo
          else ! other reactions
             rate(:,j)=rrt(:,j)*na(:,ireac(j,1))
             if (j.eq.iup23.and.verbose.eq.-2) then
                print*,'na=',na(nz,ireac(j,1))
                print*,'nNa23=',na(nz,a23m)
                print*,'rate=',rate(nz,j)
             endif
             if (ratype(j).ne.'u') rate(:,j)=rate(:,j)*rhb!*1.d-3
             if (ratype(j).eq.'u') rate(:,j)=rate(:,j)
             if (kreac(j).eq.2) then
                rate(:,j)=rate(:,j)*na(:,ireac(j,2))
             endif
             if (j==irpp.or.j==ir33.or.j==ircc.or.j==iroo) then
                rate(:,j)=rate(:,j)*0.5
             endif
             if (j==ir3a) then
                rate(:,j)=rate(:,j)*rhb*na(:,ireac(j,1))/6.
             endif
          endif
       else ! branching rates
          rate(:,j)=rrt(:,j)
       endif
    enddo

    ! Urca equilbrium fraction 
    if (iup23.ne.0.and.ium23.ne.0) then
       eqUrca=(frac_init(A23M)+frac_init(A23D))&
            *rrt(:,iup23)/(rrt(:,iup23)+rrt(:,ium23))
    else
       eqUrca=0.
    endif
    ! Urca rates at equilibrium
    if (iup21.ne.0.and.ium21.ne.0.and.&
         rstatus(A21M)=='e'.and.rstatus(A21D)=='e') then
       where(rate(:,iup21)+rate(:,ium21)==0.) 
          rate(:,iup21)=0.
       elsewhere
          rate(:,iup21)=rate(:,iup21)*rate(:,ium21)/&
               (rate(:,iup21)+rate(:,ium21))&
               *(frac_init(A21M)+frac_init(A21D))/am(A21M)
          rate(:,ium21)=rate(:,iup21)
       endwhere
    endif
    if (iup23.ne.0.and.ium23.ne.0.and.&
         rstatus(A23M)=='e'.and.rstatus(A23D)=='e') then
       where(rate(:,iup23)+rate(:,ium23)==0.) 
          rate(:,iup23)=0.
       elsewhere
          rate(:,iup23)=rate(:,iup23)*rate(:,ium23)/&
               (rate(:,iup23)+rate(:,ium23))&
               *(frac_init(A23M)+frac_init(A23D))/am(A23M)
          rate(:,ium23)=rate(:,iup23)
       endwhere
    endif
    if (iup25.ne.0.and.ium25.ne.0.and.&
         rstatus(A25M)=='e'.and.rstatus(A25D)=='e') then
       where(rate(:,iup25)+rate(:,ium25)==0.) 
          rate(:,iup25)=0.
       elsewhere
          rate(:,iup25)=rate(:,iup25)*rate(:,ium25)/&
               (rate(:,iup25)+rate(:,ium25))&
               *(frac_init(A25M)+frac_init(A25D))/am(A25M)
          rate(:,ium25)=rate(:,iup25)
       endwhere
    endif

    ! Branching of pN and CC reactions
    if (irpng.ne.0.and.irpn.ne.0) then
       FPNG = 8.0D-4
       RATE(:,IRPNG) = FPNG*RATE(:,IRPN)
       RATE(:,IRPN) = (1.0 - FPNG)*RATE(:,IRPN)
    endif
    if (irccg.ne.0.and.ircc.ne.0) then
       FCCG = RATE(:,IRCCG)
       rate(:,iRCCG) = FCCG*rate(:,iRCC)
       RATE(:,IRCC) = (1.0 - FCCG)*rate(:,iRCC)
    endif
    ! PP chain in equilibrium, RPP becomes effective rate of 2 H1 -> 0.5 He4
    if (rstatus(He3)=='e'&
         .and.ir34.ne.0.and.irpp.ne.0.and.ir33.ne.0.and.irbe.ne.0.and.irbp.ne.0) then
       F34 = 0.0
       where (rate(:,iR34) .GT. 1.0D-20) &
            F34 = 2.0/(1.0 + SQRT(1.0 + 8.0*RATE(:,IRPP)*RATE(:,IR33)/(RATE(:,IR34)*RATE(:,IR34))))
       RATE(:,IRPP) = RATE(:,IRPP)*(1.0 + F34)
       PP2 = 1.0
       where (RATE(:,IRBE)+RATE(:,IRBP) .GT. 1.0D-20) PP2 = RATE(:,IRBE)/(RATE(:,IRBE) + RATE(:,IRBP))
          PP3 = 1.0 - PP2
          QPP = Qvalue(irpp) + 0.5*Qvalue(ir33)
          QNPP = (QNT(irpp) + F34*(QNT(irbe)*PP2 + QNT(irbp)*PP3))/(1.0 + F34)
          ! calculate energy release and neutrino losses, in erg/gram/sec
          EX = QPP*RATE(:,IRPP)
          ENX = QNPP*RATE(:,IRPP)
       else
          ex=0.
          enx=0.
       endif
       eUrca=0.
       sni=0.
       DO J = 1,nreac
          ! if He3 is in equilibrium, this means that Qvalues for PP chain
          ! have already been accounted for.
          if (.not.(rstatus(He3)=='e'.and.(&
               j==irpp.or.j==ir33.or.j==ir34.or.j==irbe.or.j==irbp))) then

             ! Compute "thermal" Q-value
             smu=0.
             do k=1,5
                if (ireac(j,k).ne.0) then
                   if (firstguess) then
                      mu=muip(:,ireac(j,k))
                   else
                      mu=mui(:,ireac(j,k))
                   endif
                   if (rstatus(ireac(j,k)).ne.'c') then !.and.&
                      !  rstatus(ireac(j,k)).ne.'C') then
                      smu=smu+stoech(j,k)*mu
                   endif
                endif
             enddo
             if (ratype(j).ne.'u')EX = EX + (Qvalue(j)+ 0.*smu/cmevmu)*RATE(:,J)             
             !EX = EX + (Qvalue(j)+1.*smu/cmevmu)*RATE(:,J)
             !                if (j.eq.ircc) qcc=(Qvalue(j)+smu/cmevmu)*RATE(:,J)*cmevmu
             if (j.eq.ircc) qcc=Qvalue(j)*Rate(:,J)*cmevmu
             if (ratype(j).eq.'u') then
                ! Urca losses
!                enx=enx+UrcaQ(:,jtable(j))*rate(:,j)
                k=ireac(j,1)
                ! Adjust mui instead
                sni(:,k)=(Qvalue(j)-UrcaQ(:,jtable(j)))*cmevmu/stoech(j,1)
                mui(:,k)=mui(:,k)+(Qvalue(j)-UrcaQ(:,jtable(j)))*cmevmu/stoech(j,1)
                eUrca=eUrca+(-UrcaQ(:,jtable(j))&
                     +(Qvalue(j)+0*smu/cmevmu))*RATE(:,j)
             else
                !print*,QNT(j)
                ! Other neutrino losses
                ENX = ENX + QNT(j)*RATE(:,J)
             endif
          endif
       END DO
       EX = CMEVMU*EX/AVM
       ENX = -CMEVMU*ENX/AVM

       eUrca=CMEVMU*eURCA/AVM

       ! Compute Urca functions
       if (iup23.ne.0.and.ium23.ne.0) then
          Hurca=-urcaQ(:,jtable(iup23))*rrt(:,iup23)&
               +urcaQ(:,jtable(ium23))*rrt(:,ium23)+&
               dmuUrca*(rrt(:,iup23)+rrt(:,ium23))
          Curca=-(urcaQ(:,jtable(iup23))+urcaQ(:,jtable(iup23)))*&
               rrt(:,iup23)*rrt(:,ium23)/(rrt(:,iup23)+rrt(:,ium23))
          XLurca=delta*qcc/(cp*T)/(&
               (rrt(:,iup23)+rrt(:,ium23))/Ne&
               -delta*Hurca/cP/T)*23.
       endif
     END SUBROUTINE NUCRAT
end MODULE nuclear
