MODULE parameters       
  ! PHYSICAL PARAMETERS
  implicit none

  ! STAR parameters
  integer::  ICL, ION, IAM, IOP, INUC, IBC, ICN,JP
  double precision:: ALPHAM, RML, CMG, CSI, CFE, CT(10), M1, &
       EC, BM, ANG, MTA, MTB, TM, T0, M0, OS, RCD,&
       RMG, RHL, XF, DRdum ,RMT,  IB,&
       TRB, ZS
  integer:: IX,IY,IZ
  double precision::ICH,ITH
  integer:: JIN
  integer:: NCH

  ! modular parameters

  ! method
  character*3::VANLEER,MIXING,MIXSCHEME,SEMICONV,WORKC
  double precision::alphagrid
  character*3::TRUEL
  character*3::PARTCONV,UDIFF,UCHEM,USEDGOLD
  character*3::STAGGERED,DOSAVE,NOFLUX
  integer::imuidn,imuphi

  ! boundary conditions
  ! accretion parameters
  double precision::taccretion,tcool
  double precision::tacc0,macc0
  ! Mass and central temperature limits
  double precision::Masslim,Tlim
  ! Controls for the mass of the central zone
  double precision::mcentre,tcentre

  ! mesh
  integer::ifix,nbs
  double precision::taugref 
  double precision::stiffness
  character*3::RESOLUTION,MESH

  ! implicitation parameter
  double precision::alpha
  ! Time integration scheme
  character*3::T2ndORDER

CONTAINS
  SUBROUTINE paramphysics(unit)
    use constants,only:Msun,Rsun,CSECYR
    implicit none
    ! Read the physical parameters
    integer unit


    ! dummies to read STAR input parameters
    integer:: idum
    double precision:: dum
    integer,dimension(110)::ID
    integer,dimension(45)::ISX

    ! Read STAR data, usually unchanged during one evol run         
    ! Most of these are unused in the code.
    ! The only used ones that are not overwritten by 'inlist' parameters
    ! are ICL, CT and JP, I believe..
99003 FORMAT (11I4,/,7I4,/,6I4,/,1P,5E8.1,0P,/,2(10I3,/,3(15I3,/)),3(15I&
         3,/), 9F5.2, 1P, 3E8.1,&
         /, E9.2, 0P, 9F6.3, /, 1P, 2(7E9.2, /), 0P)
    READ  (20,99003) idum,idum,idum,JIN,idum,NCH,JP,ITH,IX,IY,IZ,&
         ICL,ION,IAM,IOP,IBC,INUC,ICN,&
         idum,idum,idum,idum,idum,idum,& 
         dum,dum,dum,dum,dum,&
         ID,ISX,&
         dum,dum,CT,&
         ZS,ALPHAM,dum,dum,dum,dum,dum,CMG,CSI,CFE,&
         RCD,OS,RML,RMG,dum,XF,DRdum,&
         RMT,RHL,dum,dum,dum,dum,TRB

    read(unit,*)!'--- Physical parameters ---'

    ! modular parameters
    read(unit,*)VANLEER       ! Advection scheme
    read(unit,*)STAGGERED     ! Centering scheme
    read(unit,*)NOFLUX
    read(unit,*)DOSAVE
    read(unit,*)MIXING
    read(unit,*)MIXSCHEME
    read(unit,*)SEMICONV
    read(unit,*)alphagrid
    read(unit,*)imuidn
    read(unit,*)imuphi
    read(unit,*)WORKC
    read(unit,*)PARTCONV
    read(unit,*)UDIFF
    read(unit,*)USEDGOLD
    read(unit,*)UCHEM
    read(unit,*)TRUEL
    ! boundary conditions
    read(unit,*) RMG     ! Accretion rate
    RMG=RMG/CSECYR
    read(unit,*) taccretion  !Accretion rise time scale
    taccretion=taccretion*csecyr
    read(unit,*) tcool  ! Cooling time scale
    tcool=tcool*csecyr
    read(unit,*) tacc0   ! Decay time of accretion rate
    read(unit,*) macc0   ! critical mass at which accretion decays
    read(unit,*) Masslim ! Limit mass where to stop
    read(unit,*) Tlim    ! limiting central log-temperature under which to stop
    read(unit,*) tcentre ! rate of change of central zone
    read(unit,*) mcentre ! limiting mass of central zone
    IB=1
    TM=TM*Msun
    BM=2.*TM
    M0=0.
    ANG=0.
    MTA=0.
    MTB=0.
    T0=0.

    read(unit,*) EC     ! Constant energy generation

    ! alpham (Mixing length parameter)
    read(unit,*) alpham

    ! mesh
    read(unit,*) RESOLUTION
    reaD(unit,*) MESH
    read(unit,*) taugref 
    read(unit,*) stiffness 

    ! implicitation parameter
    read(unit,*) alpha

    ! Grid numbers
    read(unit,*) ifix
    read(unit,*) nbs

    ! 2nd order scheme in time
    read(unit,*) T2ndORDER

  end SUBROUTINE paramphysics

end MODULE parameters
