MODULE printings
  use dimensions
  use constants
  use parameters
  use abundances
  use chemistry
  use secondaries
  use EOS
  use variables
  use nuclear,only:rate,enx,ex,qcc,eqUrca,eUrca
  use equations
  use maths,only:comp
  implicit none

  ! Printing parameters:
  character*3::FOLLOW
  integer::nimp,zdeb,zfin
  integer::photostep
  double precision::tphoto
  character*10::fouriername
  character*8::photoname
  character*8::hydroname,fritzname
  character*9::chimiename,chorefname
  character*9::arnaudname
  integer ::nstep_init

  ! Printing variables :
  integer ::nphoto,nstep
  ! informations for variations during time steps
  double precision::check,vartop

CONTAINS

  SUBROUTINE paramprintings(unit)
    integer unit

    read(unit,*)!'--- Printing parameters ---'
    read(unit,*) FOLLOW
    read(unit,*) nimp
    read(unit,*) zdeb
    read(unit,*) zfin
    read(unit,*) tphoto
    tphoto=tphoto*csecyr
    read(unit,*) photostep
    read(unit,*) verbose
  end SUBROUTINE paramprintings

  ! --------------------
  subroutine IMP(mode)
    ! --------------------
    use chemistry,izat=>iz
    implicit none
    integer i,mode,j,ideb,ifin
    double precision,dimension(nz)::Lnuc,t,fudge
    double precision::dum
    double precision::tchem
    double precision::trho,tKH,mdotp,tshell,tX
    integer,dimension(2)::iloc
    double precision,dimension(nz)::dL,dupdv,tdsmudn,dtm,dfm
    double precision::gne,testf,testg


    t=exp(lnT)
    
    if (verbose.ge.-10) then
       print*
       print 100, time/csecyr,dt/csecyr,taugrille/csecyr
100    format (7x,'   time = ',1pe10.3,' year(s)  dt = ',e10.3,' year(s)  taugrid = ',e8.1,'year(s)  ',/)
       print 101,mdot/(Msun/csecyr),Lacc/Lsun,Pacc/P(1)
101    format (' Accretion: Mdot=',1pe10.3,' Msun/yr  Lacc=',1pe10.3,' Lsun Pacc/P(1)=',1pe10.3)
    endif

!!$    print 150
!!$    do i=1,nz
!!$       print 300,i,&
!!$            wt(i),&
!!$            vpk(i),&
!!$            vrk(i),&
!!$            vtk(i),&
!!$            lk(i),&
!!$            x1t(i),&
!!$            x4t(i),&
!!$            x12t(i),&
!!$            x16t(i),&
!!$            x20t(i)
!!$    enddo
!!$    print 150
!!$150 format (5x,'i ','WT         ',&
!!$         'vpk        ',&
!!$         'vrk        ',&
!!$         'vtk        ',&
!!$         'lk         ',&
!!$         'x1t        ',&
!!$         'x4t        ',&
!!$         'x12t       ',&
!!$         'x16t       ',&
!!$         'x20t       ')

    ! Compute characteristic times
    if (rho(nz)*rhoiold(nz).ne.1.&
         .and.rhoiold(nz).ne.0..and.rho(nz).ne.0..and.rmg.ne.0.) then
       trho=(lnmnew(1)-lnmold(1))/log(rho(nz)*rhoiold(nz))/rmg
    else
       trho=0.
    endif
    
    if (Lold(1)*Rold(1).ne.0.) then
       tKH=CG*Mold(1)**2/Lold(1)/Rold(1)*1.e22
    endif

    mdotp=0.
    do i=1,nz
       mdotp=mdotp+dm(i)*MAXVAL(ABS(xt2(i,:)))
    enddo
    iloc=MAXLOC(ABS(xt2))
    if (mdotp.ne.0.) then
       tshell=M(iloc(1))/mdotp
    else
       tshell=0.
    endif

    

    tchem=MAXVAL(ABS(rate))
    if (tchem.ne.0.) then
       tX=1./tchem
       iloc=MAXLOC(ABS(rate))
    else
       tX=0.
       iloc=1
    endif

    if (verbose.ge.-10) then
       print 25,MINLOC(r**2/ch),MINVAL(1.d22*r**2/ch)/csecyr,tdold/csecyr,tKHold/csecyr,trho/csecyr,tshell/csecyr,&
            tX/csecyr,rname(iloc(2)),iloc(1)
       print*
25     format('tdiff(',i4,')=',1pe10.3,' TDIFF=',1pe10.3,' tKH=',1pe10.3,' trho=',1pe10.3,&
            ' tshell=',1pe10.3,' tX=',1pe10.3,1x,a7,'(',i3,')')
    endif
    ! 

    Lnuc(nz)=ex(nz)*dm(nz)
    do i=nz-1,1,-1
       Lnuc(i)=Lnuc(i+1)+ex(i)*dm(i)
    enddo

    
    if (verbose.ge.-10) print 200,specie(1:ndim)!'g'//specie(ispecie),specie(ispecie)
    if (mode==0) then
       ideb=max(1,zdeb)
       ifin=min(nz,zfin)
    else
!!$       ideb=1
!!$       ifin=nz-1
       ideb=MAX(mode-3,1)
       ifin=MIN(mode+3,nz)
    endif
    do i = ideb,ifin
       if (i.lt.nz) then
          dL(i)=L(i)-L(i+1)
       else
          dL(i)=L(i)
       endif
       j=iprim(C12)
       gne=SUM(Ftot(i,:)*float(izat)/float(bn))
       if (i.lt.nz) then
          testf=r(i)-r(i+1)
       else 
          testf=r(i)
       endif
       testf=testf*1.d11
       testf=testf**2/dt/chid(i)
       testg=dm(i)*mt(i)/chid(i)/(surf(i)*rho(i)*1.d-11)**2
       if (verbose.ge.-10) print 300,&
            i,&
            r(i)/Rsun,&
            m(i)/Msun,&
            dm(i)/Msun,&
            mt(i)/Msun*csecyr,&
            psi(i),&
            rho(i),&
            T(i),&
            L(i)/Lsun,&
            grads(i),&
            gradc(i),&
            gradl(i),&
            vc(i),&
            tconv(i)/csecyr,&!eUrca(i),&
            tburn(i)/csecyr,&
            tburn2(i)/csecyr,&
            (Tdts(i)-Tads(i)/dmold(i)),&
            !(Tdts(i)),&
            muidN(i)/dmold(i),&
            Tdts(i)-Tads(i)/dmold(i)+muidN(i)/dmold(i),&
            !Wconv(i),&
            netheat(i),&
!            (arg(i,j),j=1,4),aimplicit(i),&
            rname(MAXLOC(ABS(rate(i,:)),MASK=rtype.eq.' ',DIM=1)),&
            1./(1.d-50+rate(i,MAXLOC(ABS(rate(i,:)),MASK=rtype.eq.' ',DIM=1)))/csecyr,&
            gNe,&
            (xa(i,j),j=1,ndim)

    end do
    if (verbose.ge.-10) print 200,specie(1:ndim)!'g'//specie(ispecie),specie(ispecie)
!!$    print *
!!$    print*,'i       en      ex      enx'
!!$    do i=1,nz
!!$       print 300,i,en(i),ex(i),enx(i)
!!$    enddo
!!$    print*,'i       en      ex      enx'
!!$    print *


!!$    print 202,specie(1:ndim)
!!$    do i=1,nz
!!$       print 203,i,Tdts(i),mudtn(i),-dmold(i)*muxt(i),muedtn(i),(muHedtn(i,j),j=1,3),(mujdtn(i,j),j=1,ndim)
!!$    enddo
!!$    print 202,specie(1:ndim)
    202 format(' zone       ',&
             ' Tdts       ',&
             ' mudtn      ',&
             ' muxt       ',&
             ' muedtn     ',&
             'He+         ',&
             'He++        ',&
             'He          ',&
             20(a7,4x))
    203 format(2x,i5,5x,41(1pe10.3,1x))

    ! ------------------------------------------------------------
    ! Write the HYDRO file
    ! ------------------------------------------------------------
    OPEN(60,file=fritzname,status='unknown')
    write(60,*)'radius(cm)      density(g.cm-3) temperature(K)  pressure(cgs)   C               O'
    do i=nz,1,-1
	write(60,'(10(1pe15.8,1x))')&
	r(i)*1e11,&
        rho(i),&
	T(i),&
        P(i),&
	xa(i,C12),& 
	xa(i,O16)
    enddo
    CLOSE(60)
	
    OPEN  (60,FILE = hydroname,       STATUS = 'unknown')
    write (60,201)  specie(ispecie),'g'//specie(ispecie) 
    do i = 1,nz
       write (60,360)&
            i,&
            r(i),&
            fk(i),&
            m(i)/Msun,&
            dm(i)/Msun,&
            T(i),&
            Lrad(i)/Lsun,&
            Lconv(i)/Lsun,&
            Ltot(i)/Lsun,&
            psi(i),&
            lnf(i),&
            rho(i),&
            p(i),&
            s(i),&
            mt(i)/Msun*csecyr,&
            grads(i),&
            gradc(i),&
            gradl(i),&
            vc(i),&
            wconv(i),&
            tconv(i)/csecyr,&
            tburn(i)/csecyr,&
            tburn2(i)/csecyr,&
            (Tdts(i)-Tads(i)/dmold(i)),&
            muidN(i)/dmold(i),&
            muedne(i)/dmold(i),&
            -DtN(i),&
            diffN(i),&
            diffN(i)+RN(i)+DtN(i),&
            !advn(i),&
            RN(i),&
            FluxN(i),&
            Tdts(i)-Tads(i)/dmold(i)+muidN(i)/dmold(i),&
            (dxenergy(i))/dxlnp(i),&
            -(Ltot(i)-Ltot(i+1))/dmold(i),&
            ex(i),qcc(i),enx(i),en(i),netheat(i),&
            time/csecyr,&
            (frac(i,j),j=1,species),&
            (gfrac(i,j),j=1,species)
    end do
    CLOSE (60)
360 format (i5,1x,1pe22.15,1x,90(1pe10.3,1x))

200 format (5x,'i ','r(Rsun)   ',&
         'm(Msun)   ',&
         'dm(Msun)  ',&
         'mt(Ms/y)  ',&
         'psi       ',&
         'rho       ',&
         'T(K)      ',&
         'Ltot(sun) ',&
         'DG        ',&
         'LG        ',&
         'DGLG      ',&
         'vc        ',&
         'tc        ',&
         'tb        ',&
         'tb2       ',&
         'TdS(cpT)  ',&
         'muidN     ',&
         'Eg        ',&
         'Et        ',&
!!$         'arg1      ',&
!!$         'arg2      ',&
!!$         'arg3      ',&
!!$         'arg4      ',&
!!$         'aimplicit ',&
         'reaction  ',&
         'treac     ',&
         'gNe       ',&         
        11(a7,3x))


201 format (5x,'i ','r(Rsun)                ',&
         'Kap        ',&
         'm(Msun)    ',&
         'dm(Msun)   ',&
         'T(K)       ',&
         'Lrad(Lsun) ',&
         'Lcon(Lsun) ',&
         'Ltot(Lsun) ',&
         'psi        ',&
         'lnf        ',&
         'rho        ',&
         'p(cgs)     ',&
         's(cP)      ',&
         'mt         ',&
         'DG         ',&
         'LG         ',&
         'DGLG       ',&
         'vc         ',&
         'Wc         ',&
         'tconv      ',&
         'tburn1     ',&
         'tburn2     ',&
         'TDS(cpT)   ',&
         'muidN      ',&
         'muedNe     ',&
         'DtN        ',&
         'diffN      ',&
         'advN       ',&
         'RN         ',&
         'FluxN      ',&
         'Eg         ',&
         'ds(cp)     ',&
         'dLm        ',&
         'ex         ',&
         'qcc        ',&
         'enx        ',&
         'en         ',&
         'Et         ',&
         'time(yr)   ',&
         20(a7,4x))
300 format (1x,i5, 19(1pD9.2,1x),1x,(a7,2x,1pD9.2,1x),60(1pD9.2,1x))
!!$300 format (1x,i5, 29(1pD11.4,1x),1x,2(a7,2x,1pD11.4,1x),60(1pD11.4,1x))
!!300 format (1x,i5, 29(1pD11.4),1x,2(a7,1x,1pD11.4),60(1pD11.4,1x))
301 format (1x,i5,1pe10.3,1x,2(11x),40(1pe10.3,1x))


    if (verbose.ge.2)print*,'End print'
  end subroutine IMP


  SUBROUTINE output
    ! ------------------------------------------------------------
    ! Write the general save file
    ! ------------------------------------------------------------
    OPEN (80,FILE = 'output',STATUS = 'unknown')
    CALL outputhydro
    CLOSE(80)
    OPEN (80,FILE = photoname,STATUS = 'unknown')
    CALL outputhydro
    CLOSE(80)
  end SUBROUTINE output

  SUBROUTINE outputhydro
    implicit none
    integer i,ij
!!$    ! Convert dm to dr
!!$    do i=1,nz-1
!!$       ij=(i-1)*nvar+idm
!!$       x(ij)=LOG(3.*dm(i)/(pi4*rho(i)*(r(i)*r(i)+r(i)*r(i+1)+r(i+1)*r(i+1))))
!!$       dxold(ij)=0.
!!$    enddo
!!$    ij=ij+nvar
!!$    x(ij)=lnr(nz)
!!$    dxold(ij)=0.

!!$    do i=1,nz
!!$       ij=(i-1)*nvar+ins
!!$       x(ij)=LOG(nsmooth(i)*pi4*rho(i)*r(i)*r(i))
!!$       dxold(ij)=0.
!!$    enddo
    
    write(80,*) nvar
    write(80,*) nz
    write(80,*) x
    write(80,*) dxold
    write(80,*) time
    write(80,*) dtold
    write(80,*) dt
    write(80,*) nstep
    write(80,*) nphoto
    write(80,*) threshold
  end SUBROUTINE outputhydro


  SUBROUTINE initphoto
    double precision c_init
    nphoto=0
    ! pas de mode ?
    if (tphoto.eq.0. .and. photostep.eq.0) then
       print*,'INITPHOTO:Choisir un mode d''impression pour les photos.'
       STOP
    endif
    ! mode: tous les photostep pas de temps.
    if (tphoto.eq.0.) then
       nphoto=nstep_init/(photostep)
    endif
    ! mode: ds que temps=nphoto*tphoto
    if (tphoto.ne.0..and.photostep==0) then
       nphoto=INT(time/tphoto)
    endif
    ! mode: ds que temps=tphoto*10.^((nphoto-1)/photostep)
    if (tphoto.ne.0.and.photostep.ne.0.) then
       if (time.lt.tphoto) then
          nphoto=0
       else
          nphoto=INT(alog10(time/tphoto)*float(photostep))+1
       endif
    endif
    nphoto=nphoto+1
    CALL photofilename
  end SUBROUTINE initphoto

  SUBROUTINE photofilename
    fouriername='FOURIER'//numero(nphoto)
    photoname='PHOTO'//numero(nphoto)
    hydroname='HYDRO'//numero(nphoto)
    fritzname='FRITZ'//numero(nphoto)
    chimiename='CHIMIE'//numero(nphoto)
    chorefname='CHOREF'//numero(nphoto)
    arnaudname='ARNAUD'//numero(nphoto)
    print*,'----------------------------------------------'
    print*,'PHOTO number',nphoto,'in filess:'
    print*,photoname,',',hydroname,',',chimiename,',',chorefname
    print*,'----------------------------------------------'
    if (photostep.ne.0.and.tphoto.eq.0.) then
       print 333,(nphoto)*photostep
333    format('Next snapshot nstep=',i5) 
    endif
    if (tphoto.ne.0.) then
       if (photostep==0) then
          print 334,float(nphoto)*tphoto/csecyr
       else
          print 334,tphoto*10.**(float(nphoto-1)/float(photostep))/csecyr
       endif
334    format('Next snapshot at time=',1pe13.3,' year(s)')
    endif
  end SUBROUTINE photofilename

  character*3 FUNCTION numero(nombre)
    integer nombre
    integer n1,n2,n3
    n1=nombre/100
    n2=(nombre-100*n1)/10
    n3=nombre-100*n1-10*n2
    numero=CHAR(n1+48)//CHAR(n2+48)//CHAR(n3+48)
  end FUNCTION numero

  SUBROUTINE TRACE_init

    OPEN(90,status='Replace',file='TRACE',position='Rewind') 
    write(90,'(a120,40(1x,a8))')&
 'nstep time dt R M L Te muec Yec Tc Tmax rhoc izcb mf mg mg0 dmf mft mf0 mft0 mt tconv tb tb2 tkh td ',&         
 'c'//specie(1:ndim),'s'//specie(1:ndim),'M'//specie(ispecie)
    CLOSE(90)
  end SUBROUTINE TRACE_init

  SUBROUTINE TRACE
    use abundances
    use equations,only:cb,ncb
    implicit none
    integer i
    integer,dimension(ndim)::is,ic
    integer,dimension(species)::im
    integer::gcb,incb
    double precision::mf,mg,mg0,dmf,mft,mf0,mft0
    integer::izmf
    do i=1,species
       im(i)=icross(frac(:,i),nz,-1,.1d0)
    enddo
 
    
    if (ncb.gt.0) then
       if (cb(ncb)%top) then
          gcb=nz
          mg=m(nz)/Msun
          mg0=m(nz)/Msun
          incb=ncb
       else
          mg=cb(ncb)%m/Msun
          mg0=cb(ncb)%m0/Msun
          gcb=cb(ncb)%iz-1
          incb=max(1,ncb-1)
       endif
       izmf=cb(incb)%iz
       mf=cb(incb)%m/Msun
       mf0=cb(incb)%m0/Msun
       dmf=cb(incb)%dm/dm(izmf)
       mft=cb(incb)%mt/(Msun/csecyr)
       mft0=cb(incb)%mt0/(Msun/csecyr)
    else
       mf=0.
       dmf=0.
       mft=0.
       izmf=1
       gcb=nz
    endif
    OPEN(90,status='unknown',file='TRACE',position='Append')
    write(90,1)nstep,&
         time/csecyr,&
         dtold/csecyr,&
         r(1)/Rsun,&
         M(1)/Msun,&
         L(1)/Lsun,&
         T(1),&
         mue(nz)/cmevmu,&
         ye(nz),&
         T(nz),&
         MAXVAL(T),&
         rho(nz),&
         float(izmf),mf,mg,mg0,dmf,mft,mf0,mft0,mt(1)/(Msun/csecyr),&
         tconv(gcb)/csecyr,tburn(gcb)/csecyr,tburn2(gcb)/csecyr,&
         tkhold/csecyr,tdold/csecyr,&
         xa(gcb,:),&
         xa(1,:),&
         m(im)/Msun


1   FORMAT(i10,1x,80(1pe10.3,1x))
    CLOSE(90)
  end SUBROUTINE TRACE

  function icross(f,istart,is,const) Result(ic)
    double precision,dimension(nz)::f
    double precision const
    integer istart,is,ic
    integer i
    if (is.gt.0) then
       do i=istart,nz-1,is
          if ((f(i)-const)*(f(i+1)-const).lt.0.) then
             ic=i
             return
          endif
       enddo
       ic=nz
    else
       do i=istart,2,is
          if ((f(i)-const)*(f(i-1)-const).lt.0.) then
             ic=i
             return
          endif
       enddo
       ic=1
    endif
    return
  end function icross

!!$  function new(f) Result(fnew) 
!!$    double precision,dimension(nz)::f,fnew
!!$    integer,dimension(nz)::i
!!$    integer j
!!$    logical,dimension(nz)::ok
!!$    double precision,dimension(nz)::a
!!$    i=1
!!$    do j=1,nz
!!$       where(m.lt.mold(j)) i=j
!!$    enddo
!!$
!!$    ok=(i.lt.nz)
!!$    where(ok) ok=ok.and.mold(i+1).ne.mold(i)
!!$    ! check
!    do j=1,nz
!       if (ok(j)) then
!          if (m(j).gt.mold(i(j))) then
 !            print*,'Pb for j=',j,' m(j)>mold(i) i=',i(j)
!          endif
!          if (m(j).le.mold(i(j)+1)) then
!             print*,'Pb for j=',j,' m(j)<=mold(i+1) i=',i(j)
!          endif
!       endif
!    enddo
!    fnew=f(I)
!!$
!!$    where(ok)
!!$       a=(mold(i)-m)/(mold(i)-mold(i+1))
!!$       fnew=f(i+1)*a+f(i)*(1.-a)
!!$    elsewhere
!!$       fnew=f(i)
!!$    endwhere
!!$
!!$    return
!!$  end function new
!!$  function new(k,f) Result(fnew) 
!!$    double precision::f,fnew
!!$    integer::k
!!$    integer::i
!!$    integer j
!!$    i=1
!!$    do j=1,nz
!!$       if (m(k).le.mold(j)) i=j
!!$    enddo
!!$
!!$    if (i.lt.nz.and.mold(i).ne.m(k)) then
!!$       fnew=f(i)*(m(k)-mold(i))+f(i+1)*(mold(i+1)-m(k))
!!$       if (mold(i+1)==mold(i)) then
!!$          fnew=f(i)
!!$       else
!!$          fnew=fnew/(mold(i+1)-mold(i))
!!$       endif
!!$    else
!!$       fnew=f(i)
!!$    endif
!!$    return
!!$  end function new
END MODULE printings
