! routine statef from the STAR program
! I added the computation of the chemical potential,
! corrected the computation of internal energy along with a few bugs.
MODULE EOS
  ! Computes state variables, their derivatives, opacities, neutrino losses.
  use constants,only:cevb,cd,cen,ca,cb,cl,cr,cln10,ct,boltzm,amu,pi4,cmevmu
  use dimensions,only:nz,species,ndim,verbose
  use parameters,only:icl,iop,ion,UCHEM
  use abundances
  use secondaries,only:h,rho,lnp,p,T,cP,pt,pf,ch,grada,delta,sni,rni,mui,pni,na,sf,st,chid,mue,ne
  use chemistry,only:BN,IZ,AM,H1,He4,istatus,specie,a23d,ium23,qvalue
  implicit none

  double precision,dimension(nz)::gam
  ! PARAMETER
  logical,save::firstguess=.false.
  logical::toolow,toohigh

  ! IN 
  ! lnf, lnT, and abundances na.

  ! OUT 
  ! State variables (commented variables are in module secondaries)
!  double precision,dimension(nz)::T            ! Temperature
  double precision,dimension(nz)::arho,rt,rf   ! mass density, its log and derivatives
  double precision,dimension(nz)::pr,pg!,pf,pt!p,lnp ! pressure 
  !        (radiative, gas, total, derivatives, log)
  double precision,dimension(nz)::s!,sf,st    ! entropy and its derivatives
  double precision,dimension(nz)::energy     ! internal energy
  !energy Coulomb corr, and pressure of electron gas
  double precision,dimension(nz)::eelectron,eCoulomb,pelectron,pcoulomb 

  double precision,dimension(nz)::zt         ! ?
!  double precision,dimension(nz)::grada      ! d ln T / d ln P | adiabatic
  double precision,dimension(nz)::ftp        ! d lnf / d lnT | P,N
!  double precision,dimension(nz)::cp         ! heat capacity at cst pressure&abundances
  double precision,dimension(nz)::gamma      ! adiabatic coefficient
!  double precision,dimension(nz)::delta      ! d ln rho / d ln P | P,N
  
  double precision,dimension(nz,ndim)::muip!,mui   ! Chemical potentials
!  double precision,dimension(nz,ndim)::pni   ! d ln p / d Ni | T,P,Nj
!  double precision,dimension(nz,ndim)::sni   ! d s / d Ni | T,P,Nj
!  double precision,dimension(nz,ndim)::rni   ! d ln rho / d Ni | T,P,Nj

  double precision,dimension(nz)::psi        ! Fermi potential of the electrons
!  double precision,dimension(nz)::Ne         ! Electron density (per unit mass) 
!  double precision,dimension(nz)::mue        ! Electron chemical potential
  double precision,dimension(nz)::en         ! standard neutrino losses
  double precision,dimension(nz)::fk!,ch      ! opacity and heat diffusion coefficient

  ! LOCALS 
  ! tst must be (close to) zero, but does not mean all is correct if zero...
  private tst
  double precision,dimension(nz)::tst

  ! Tables for opacity and neutrino losses
  private cs
  double precision,dimension(90,127,10)::cs

  ! Fermi-Dirac variables
  private RE, PE, QE, RET, PET, QET, REF, PEF, QEF
  private DTT, DFT, DFF, DTTT, DFTT, DFFT, DFFF
  double precision,dimension(nz)::RE, PE, QE, RET, PET, QET, REF, PEF, QEF
  double precision,dimension(nz)::DTT, DFT, DFF, DTTT, DFTT, DFFT, DFFF

  ! Derivatives of Coulomb and ionisation pressure corrections.
  private DFCn01,DFCn12,n01,n12,dmui,DPn01,DPn12,DSn01,DSn12
  double precision,dimension(nz),save::DFCn01,DFCn12,n01,n12,dmui
  double precision,dimension(nz),save::DPn01,DPn12,DSn01,DSn12

 contains  
  SUBROUTINE STATEF(FL, TL)
    use maths,only:fxp ! truncated exponential to avoid overflow 
    implicit none

    ! IN
    ! LN(f), LN(T)
    double precision,dimension(nz)::fl,tl

    ! local variables
    integer::i,j,idum
    double precision,dimension(nz)::f,uf,wf,g
    double precision,dimension(nz)::se,sef,set,ue
    ! interface with pressi
    double precision,dimension(nz)::DC, DVT, DVF, DPA, DPAT, DPAF, DSA, DSAT, DSAF, DUA
    double precision,dimension(nz)::ti,de,dv
    double precision,dimension(nz)::nef,net,si,sif,sit,ui

    double precision::vm
    double precision,dimension(nz)::sha,sjha,scha
    double precision,dimension(nz,26)::va,ha
    double precision,dimension(nz)::vx,nx,nxf,nxt

    double precision,dimension(nz)::dh2ti,d1ti,d2ti,d3ti,zet,dzh2t,dzh2tt
    double precision,dimension(nz)::zh2,zh2t,zh2tt,zh2s
    double precision,dimension(nz)::h2a,h2bt,h2at
    double precision,dimension(nz)::aa,ab,ac

    double precision,dimension(nz)::invne
    double precision,dimension(nz)::ni

    double precision,dimension(nz)::ad,af,at,ae,bf,bt,hgf,hgt,hif,hit,h2f,h2t
    double precision,dimension(nz)::dl
    double precision,dimension(nz)::DPB, DPBT, DPBF, DSB, DSBT, DSBF, DUB
    double precision,dimension(nz)::tcr,p0,pi,t4,b,dsf,dst
    double precision,dimension(nz)::q
    double precision,dimension(nz)::fr,tf
    double precision,dimension(nz)::h2,hg,hi

    ! Parameter to avoid log(zero)
    double precision,parameter::eps=1.d-30

    ! Correction for chemical potentials:
    double precision,dimension(nz)::n01Ni,n12Ni

    ! Data for ionisation potentials
    ! detection of first call to initialise data levels
    logical,save::first=.true.
    ! Occupation number
    integer,parameter,dimension(27)::&
         OMG=(/1,2,1,2,1,6,9,4,9,6,1,2,1,6,9,4,9,6,1,10,21,28,25,6,25,28,21/)
    ! Levels
    double precision,save,dimension(26,9)::&
         CHI=RESHAPE((/13.595, (0.0,idum=1,25),&
         24.580, 54.403, (0.0,idum=1,24),&
         11.26,24.38,47.86,64.48,391.99,489.84,(0.0,idum=1,20),&
         14.54,29.60,47.43,77.45,97.86,551.92,666.83,(0.0,idum=1,19),&
         13.61,35.15,54.93,77.39,113.87,138.08,739.11,871.12,(0.0,idum=1,18),&
         21.56,41.07,63.5,97.16,126.4,157.91,207.3,239.,1196.,1360.,&
         (0.0,idum=1,16),&
         7.64,15.03,80.12,109.29,141.23,186.86,225.31,265.96,327.90,&
         367.36,1761.2,2085.,(0.0,idum=1,14),     &
         8.15,16.34,33.46,45.13,166.73,205.11,246.41,303.87,&
         351.83,401.3,476.0,523.2,2436.,2666.,(0.0,idum=1,12),&
         7.90,16.18,30.64,56.,79.,105.,133.,151.,235.,262.,290.,321.,&
         355.,390.,457.,489.,1266.,1358.,1456.,1582.,1689.,1799.,&
         1950.,2045.,8828.,9278./),(/26,9/))
    ! Z for each of the above row of levels.
    integer,parameter,dimension(9)::zpot=(/1,2,6,7,8,10,12,14,26/)
    ! Points to the correct row of levels
    integer,dimension(ndim)::ipot
    ! H2 partition function
    double precision,parameter::DH2=4.477, D1=-0.448, D2=0.1562, D3=-0.0851
    
    if (verbose.ge.2) print*,'  statef'

    ! Initalise ionisation potentials
    if (first) then
       first=.false.
       ! sum up ionisation potentials
       DO I = 1, 9
          DO J = 2, I
             CHI(J,I) = CHI(J,I) + CHI(J-1,I)
          END DO
       END DO
       ! attribute the potential to each specie
       do i=1,ndim
          ipot(i)=0
          do j=1,9
             if (iz(i)==zpot(j)) ipot(i)=j
          enddo
          if (ipot(i)==0.and.istatus(i)=='p') then
             print*,'  WARNING ! I don''t know the ionisation potentials for ',specie(i)
             print*,'so I switch its ionisation status to "f", fully ionised.'
             istatus(i)='f'
          endif
       enddo
    endif

    ! Recover F and T from their exponentials
    F = FXP(FL)
    T = FXP(TL)

    ! Evaluate Fermi-Dirac integrals according to Eggleton, Faulkner &
    ! Flannery (1973): RE, PE, SE and UE correspond to rho*, P*, S* and U*. 
    ! PSI is the usual degeneracy parameter
    UF = F/(1.0+F)
    WF = SQRT(1.0+F)
    PSI = FL + 2.0*(WF-LOG(1.0+WF))
    G = CT*T*WF
    ! Fdirac
    CALL FDIRAC(F, G)
    PE = G*PE
    PET = PET + 1.0
    PEF = PEF + 0.5*UF
    QE = QE/(RE*WF)
    SE = QE + 2.0*WF - PSI
    SEF = QE*(QEF-REF-0.5*UF) - 1.0/WF
    SET = QE*(QET-RET)
    UE = G/WF*RE*(SE+PSI)-PE

    ! PRESSI gives a crude model for pressure ionization and (if ICL=1 or 2) 
    ! a model for Coulomb interactions, returning corrections to the electron
    ! chemical potential, pressure, entropy and internal energy.
    ! TI is 1eV/kT, DE is 1 amu * number of electrons/cm3
    TI = CEVB/T
    DE = RE*CD
    n01=NEO/NIO
    n12=NZZ/NEO
    DFCn01=0.
    DFCn12=0.
    CALL PRESSI(ICL, TI, DE, REF, RET, F, DC, DVT, DVF, DPA, DPAT,&
         DPAF, DSA, DSAT, DSAF, DUA)
    DV = DC - PSI
    DVF = DVF - WF

!!$    ! ionisation pressure correction to the ionisation equilibrium
!!$    if (Icl.le.1) dv=dv-dmui

    ! chemical potential of the electrons
    mue=psi-dc

    ! Contributions of the completely ionised species
    NE = 0.0
    NEF = 0.0
    NET = 0.0
    SI = 0.0
    SIF = 0.0
    SIT = 0.0
    UI = 0.0
    mui=0.
    sni=0.
    pni=0.
    rni=0.
    DO I = 1,ndim
       if (istatus(i)=='f') then
          ! Contribution to
          ! the electron number :
          NE = NE + IZ(I)*NA(:,I)
          ! the entropy
          VM = AM(I)*SQRT(AM(I))
          SI = SI - NA(:,I)*LOG(NA(:,I)/VM + EPS)
          ! the chemical potential
          MUI(:,I)=LOG(NA(:,I)/VM + EPS)+mue*IZ(I)
          ! Contribution to energy is 0 since zero point for
          ! energies is taken at fully ionised state.
          
          ! Derivatives with respect to Ni :
          PNI(:,i)=1.
          SNI(:,i)=CR*(1.5-LOG(NA(:,I)/VM + EPS))
          RNI(:,i)=IZ(i)
       endif
    END DO

    ! Calculate ionisation of the fractionaly ionised elements.
    ! WARNING ! Hydrogen has to be computed the last. So H1=1 is 
    ! is enforced.
    if (H1.ne.1) then
       print*,'Please make sure that Hydrogen is the first specie in chemistry-data,'
       print*,'otherwise its ionisation/molecular state is not correctly computed...'
       STOP
    endif

    DO I = ndim,1,-1
       if (istatus(i)=='p') then
          SHA = 1.0
          SJHA = 0.0
          SCHA = 0.0
          ! compute potentials -VA and number ratios HA of ionization state J
          ! relative to the ground state
          DO J = 1, IZ(I)
             VA(:,J) = -CHI(J,ipot(i))*TI + J*DV
             IF(J.EQ.1) THEN
                HA(:,J) = FXP(VA(:,J))*OMG(IZ(I))/OMG(IZ(I)+1)
             ELSE
                HA(:,J) = HA(:,J-1)*FXP(VA(:,J) - VA(:,J-1))*OMG(IZ(I)+1-J)&
                     /OMG(IZ(I)+2-J)
             END IF
             SHA = SHA + HA(:,J)
             SJHA = SJHA + J*HA(:,J)
             SCHA = SCHA + CHI(J,ipot(I))*TI*HA(:,J)
          END DO
          VM = AM(I)*SQRT(AM(I))

          ! Entropy
          SI = SI + NA(:,I)*LOG(VM)

          IF(i.gt.1) THEN
             ! contributions to electron number density NE, entropy SI and
             ! internal energy UI for Helium and heavier

             ! Ground state
             VX = NA(:,I)/SHA
             ! Entropy
             SI = SI - VX*LOG(VX/OMG(IZ(I)+1) + EPS)
             ! Chemical potential
             MUI(:,I)=LOG(VX/OMG(IZ(I)+1) + EPS)&
                  -TI*CHI(IZ(I),ipot(i))-LOG(VM)
             ! Energy
             UI = UI-VX*CHI(IZ(i),ipot(i))
             ! Derivatives with respect to Ni :
             PNI(:,i)=1./SHA
             SNI(:,i)=CR/SHA*(1.5-LOG(VX/OMG(IZ(I)+1) + EPS))
             RNI(:,i)=0.

             ! Ionised states
             DO J = 1, IZ(I)
                NX = HA(:,J)*VX
                NXF = NX*DVF*(J - SJHA/SHA)
                NXT = NXF*DVT/DVF + NX*(CHI(J,ipot(I))*TI - SCHA/SHA)
                ! Electron number
                NE = NE + J*NX
                NEF = NEF + J*NXF
                NET = NET + J*NXT
                ! Entropy
                SIF = SIF - VA(:,J)*NXF
                SIT = SIT - VA(:,J)*NXT
                SI = SI - NX*LOG(NX/OMG(IZ(I)+1-J) + EPS)
                ! Energy
                UI = UI + (CHI(J,ipot(I))-CHI(IZ(i),ipot(I)))*NX
                ! Derivatives with respect to Ni :
                PNI(:,i)=PNI(:,i)+HA(:,j)/SHA
                SNI(:,i)=SNI(:,i)+CR*HA(:,j)/SHA*(1.5-LOG(NX/OMG(IZ(I)+1-J) + EPS))
                RNI(:,i)=RNI(:,i)+J*HA(:,j)/SHA
            END DO
          END IF
       endif
    END DO

    if (istatus(H1).eq.'p') then
    ! Ionisation and molecular dissociation of Hydrogen.
    ! partition function for H2 from Vardya (1960), Webbink (1975)
    DH2TI = DH2*TI
    D1TI = D1*TI
    D2TI = (D2*TI)**2
    D3TI = (D3*TI)**3
    ZET = 1.0 - (1.0 + DH2TI)*FXP(-DH2TI)
    DZH2T = -DH2TI**2*FXP(-DH2TI)/ZET
    DZH2TT = (DH2TI - 2.0 - DZH2T)*DZH2T
    ZH2 = 6608.8*ZET*DH2TI**(-2.5)*FXP(-D1TI - D2TI - D3TI)
    ZH2T = 2.5 + D1TI + 2.0*D2TI + 3.0*D3TI + DZH2T
    ZH2TT = -D1TI - 4.0*D2TI - 9.0*D3TI + DZH2TT
    ZH2S = ZH2*SQRT(8D0)/VM
    H2A = CEN*(ZH2S/4.0)*DE/(T*SQRT(T))*FXP(DH2TI)
    H2BT = DH2TI + 1.5 - ZH2T
    H2AT = RET - H2BT
    ! solve for densities of H+, H, and H2
    AA = 2*H2A + HA(:,1)*(1.0 + HA(:,1))
    AB = NE + HA(:,1)*(NE - NA(:,1))
    AC = NA(:,1)*NE
    HG = 2*AC/(SQRT(AB*AB + 4*AA*AC) + AB)
    HI = HA(:,1)*HG

    ! Total electron number density per unit mass 
    NE = NE + HI
    invNE = 1.0/NE

    ! H2 number 
    H2 = H2A*HG*HG*invNE

    ! Total number of heavy particles
    NI = NIO - H2
 
    ! derivatives w.r.t. F and T
    AA = NE + 4*HG*H2A
    AB = HA(:,1)*(NE - 2*H2)
    AD = 1/(AA + AB)
    AC = 2*H2*AD
    AF = (NEF - NE*REF)*AC
    AT = (NET - NE*H2AT)*AC
    AE = HG*AD
    BF = DVF*AE
    BT = (CHI(1,1)*TI + DVT)*AE
    HGF = AF - AB*BF
    HGT = AT - AB*BT
    HIF = HA(:,1)*(AF + AA*BF)
    HIT = HA(:,1)*(AT + AA*BT)
    NEF = NEF + HIF
    NET = NET + HIT
    H2F = H2*REF + invNE*(2*H2A*HG*HGF - H2*NEF)
    H2T = H2*H2AT + invNE*(2*H2A*HG*HGT - H2*NET)
    ! hydrogen contribution to entropy, internal energy
    SIF = SIF - VA(:,1)*HIF - H2BT*H2F
    SIT = SIT - VA(:,1)*HIT - H2BT*H2T + H2*(ZH2T + ZH2TT)
    SI = SI - HI*LOG(HI/OMG(1) + EPS) - HG*LOG(HG/OMG(2) + EPS)&
         + H2*(-LOG(H2/ZH2S + EPS) + ZH2T)
    ! Chemical potential
    mui(:,H1)=LOG(HG/OMG(2) + EPS)-LOG(VM)-TI*CHI(1,1)
    ! Total energy (zero reference is fully ionised state)
    UI = UI - CHI(1,1)*(HG+2.*H2)-DH2*H2
    ! Derivatives with respect to NI
!    print*,'Derivatives with respect to NI need to be implemented for H'
!    STOP
   else
       H2F=0.
       H2T=0.
       H2=0.
       HG=0.
       HI=na(:,H1)
       NI=NIO
       invNe=1./Ne
    endif

    ! RHO is the mass density in g/cm3
    RHO = invNE*DE
    DL = LOG(RHO)
    arho = LOG(RHO)
    RT = RET - invNE*NET
    RF = REF - invNE*NEF
    do i=1,ndim
       RNI(:,i)=-invNE*RNI(:,i)
    enddo

    ! second call to PRESSI compensates for spurious pressure & entropy terms
    DE = RHO*NEO
    if (icl.le.1) then
       CALL PRESSI(0, TI, DE, RF, RT, F, DC, DVT, DVF, DPB, DPBT, DPBF, &
            DSB, DSBT, DSBF, DUB)
    else
       CALL PRESSI(3, TI, DE, RF, RT, F, DC, DVT, DVF, DPB, DPBT, DPBF, &
            DSB, DSBT, DSBF, DUB)
    endif

    ! pressure terms
    ! electrons (Fermi)
    PE = CB*PE
    ! photons and atoms&molecules
    TCR = T*CR
    P0 = TCR*RHO
    PI = NI*P0
    T4 = T*T*T*T
    PR = CA*T4/3.0
    B = 4.0*PR/P0
    ! gas (electrons(Fermi+Coulomb&ionisation)+atoms&molecules
    !    PG = PE + TCR*(DPA-DPB) + PI
    ! a limit is enforced on Coulomb corrections to avoid pe<0.
    PG = max(PE + TCR*(DPA-DPB),0.5*pe) + PI
    ! total
    P = PG + PR
    pelectron=PE+ TCR*(DPA-DPB)+PR
    pcoulomb= TCR*(DPA-DPB)

    if (ANY(P.le.0.)) then
       print*,'Negative pressure...'
       do i=1,nz
          if (P(i).lt.0.) then 
             print*,'  Zone ',i
             print*,'pe,pi,pr=',PE(i)/tcr(i),PI(i)/tcr(i),PR(i)/tcr(i)
             print*,'dpa=',DPA(i),'dpb=',DPB(i)
             print*,'gam=',gam(i)
             print*,'psi=',psi(i),' dgdlne=~',nio(i)/neo(i)*.9*gam(i)/3.
          endif
       enddo
       stop
    endif

    PF = (PE*PEF      + PI*RF - H2F*P0 + TCR*(DPAF-DPBF))/P
    PT = (PE*PET + PI + PI*RT - H2T*P0 + TCR*(DPAT-DPBT) + PR*4.0)/P

    ! Correction to the ion chemical potentials due to ionisation pressure
    ! and Coulomb corrections (they have already be taken into
    ! account in mue and in the difference DV=mui-mue)
    ! Add the Coulomb correction part to entropy and pressure derivatives
    ! with respect to NI.
    do i=1,ndim
       n01Ni=(IZ(i)-n01)/NIO
       n12Ni=(IZ(i)*(IZ(i)-n12))/NEO
       mui(:,i)=mui(:,i)&!+DmuI*IZ(i)& ! ion. press. not yet correct, hence 0
               +NE*(DFCn01*n01Ni+DFCn12*n12Ni) ! Coulomb corrections 
       SNI(:,i)=SNI(:,i)+CR*NE*(DSn01*n01Ni+DSn12*n12Ni)&
            -CR*(NI+B+NE*(SE+DSA))*RNI(:,i)&
            +PNI(:,i)*CR*(1.5*TL-DL-LOG(CEN))!PNI=1. at this stage
       PNI(:,i)=TCR*(RHO+NI*RNI(:,i)*RHO&
            +(DPn01*n01Ni+DPn12*n12Ni))/P
    enddo
    ftp=-pt/pf
    lnp = LOG(P)

 
    ! Derivative d ln rho /d ln Ni | P,T,Nj
    do i=1,ndim
        RNI(:,i)=RNI(:,i)-PNI(:,i)*RF/PF
    enddo
     
    ! entropy, in erg/g/K
    DSF = NEF*DSA + NE*DSAF - NEO*DSBF - RF*B
    DST = NET*DSA + NE*DSAT - NEO*DSBT - (RT - 3.0)*B
    SF = CR*(-NI*RF       + NEF*SE + NE*SEF + SIF + DSF)
    ST = CR*( NI*(1.5-RT) + NET*SE + NE*SET + SIT + DST)
    do i=1,ndim
       SNI(:,i)=SNI(:,i)-PNI(:,i)*SF/PF
    enddo
    S = CR*(SE*NE+DSA*NE-DSB*NEO+B+(1.5*TL-DL+2.5-LOG(CEN))*NI+SI)

    ! chemical potentials
    mue=CR*T*mue
    do i=1,ndim
       if (istatus(i).eq.'f'.or.istatus(i).eq.'p')&
            mui(:,i)=CR*T*(mui(:,i)+DL-1.5*TL+LOG(CEN))
    enddo    
    ! Primed chemical potentials :
    do i=1,ndim
       mui(:,i)=mui(:,i)+T*SNI(:,i)
    enddo
    
    ! internal energy, in erg/g 
    energy = max(CB*UE/RHO+TCR*(DUA*NE-DUB*NEO),.5*cb*ue/rho)+TCR*(1.5*NI+0.75*B+TI*UI+ZH2T*H2)
!!$    energy = CB*UE/RHO+TCR*(DUA*NE-DUB*NEO+1.5*NI+0.75*B+TI*UI+ZH2T*H2)
    eelectron=CB*UE+rho*TCR*(DUA*NE-DUB*NEO+0.75*B)
    eCoulomb=rho*TCR*(DUA*NE-DUB*NEO)

    ! enthalpy
    h =energy+p/rho

    ! other thermodynamic quantities
    Q = PT*SF - PF*ST
    CP = -Q/PF 
    GRADA = SF/Q
    DELTA = -( RT - PT/PF*RF)
    GAMMA = Q/(RT*SF-RF*ST)
    ZT = SQRT(ABS((NE*REF/WF + NZZ)/NI))

    ! TST ought to be zero, if all the above programming is correct
    TST = SF/CR - P*(RT*PF-RF*PT)/(TCR*RHO)
!!$    print*,'TST',tst

! Diffusion coefficients
    CALL diffusion

!!! END OF THERMODYNAMIC CALCULATION. BEGINNING OF TABLE-BASED CALCULATIONS
    FR = arho/CLN10
    TF = TL/CLN10
    CALL tables(fr,tf)
  END SUBROUTINE STATEF
  
  subroutine diffusion
    ! Compute diffusion coefficients
    double precision,dimension(nz)::lam,sig,b,z,z2,a,ni,chim,om,gam
    integer::i
    if (verbose.ge.3) print*,'    diffusion'

    z=neo/nio
    z2=nz/nio
    b=0.
    do i=1,ndim
       b=b+bn(i)*na(:,i)
    enddo
    b=b/nio
    sig=1.d-15*b**(2./3.)
    ni=rho/(amu*b)
    lam=1./(sig*ni)
    ! hard spheres diffusion
    chim=lam*sqrt(max(0.,gamma*pg/rho))

    gam=57.7*(rho*1.d-6*16./b)**(1./3.)*1.d7/T*(z/8.)**2
    a=(3./pi4/ni)**(1./3.)
    om=sqrt(pi4*ni*gam*a*BOLTZM*T/b/amu)
    ! Hansen&McDonald (86) + Deloye&Bildsten(2004) diffusion coeff.
    chid=3.*om*a**2/(1.+gam**(4./3.))
    chid=chim+chid
  end subroutine diffusion

  SUBROUTINE tables(fr,tf)
    use abundances,only:xa,nzz,avm
    use chemistry,only:H1,He4,am,rname,nreac
    use constants,only:cl
    ! tables
    ! inputs
    double precision,dimension(nz)::fr,tf
    ! locals
    integer i,j,idum
    double precision,dimension(nz)::xxh,xxhe,xf
    integer,dimension(nz)::jx=2
    logical,dimension(nz)::toolow,toohigh
    double precision,dimension(nz)::xu,xt,tt,tu,rt,ru
    integer,dimension(nz)::jt,jr,it,ir,k
    double precision,dimension(nz)::fkl,fkh,enp,enb
    ! opacity data
    integer jcsx,iir,iit
    logical,save::first2=.true.
    double precision,dimension(10),save::csx

    double precision,dimension(60,41,2),save::cniu
    double precision,allocatable,dimension(:)::hat
    double precision,dimension(200,20)::crt
    integer,dimension(20)::&
         czw=(/2,8,8,0,8,12,14,16,16,24,28,32,40,72,96,128,0,0,0,0/)
    ! Neutrino Qvalues
    double precision,dimension(20)::&
       QNT=(/0.265, 0.0, 0.0, 0.814, 6.710, 0.707, 0.997, 0.999,(0.0,idum=1,11), 0.997/)
    if (verbose.ge.2) print*,'  tables'
      
      ! Read the data if this is the first time you come...
      if (first2) then
         first2=.false.

         ! Read the opacity tables
         print*,'Reading opacity-data...'
         open(unit=17,file='opacity-data',status='unknown')
         READ(17,'(I4)') JCSX
99004    FORMAT (1X, 10F7.3)
         READ (17,99004) CSX
         DO  I = 1, JCSX
            READ (17,99004) ((CS(iiR,iiT,i),iiR=1,90),iiT=1,127)
         enddo
         close(17)
         print*,'...reading completed.'

         ! Read the neutrino losses tables
         print*,'Reading neutrino-data...'
         allocate(hat(4920))
         open(unit=17,file='neutrino-data',status='unknown')
         READ (17,99004) HAT
         ! Standard neutrino losses'
         cniu=RESHAPE(hat,(/60,41,2/))
         deallocate(hat)
         close(17)
         print*,'...reading completed.'
      endif

      ! Opacity tables from Alexander & Ferguson (1994;molecular),Itoh (1983;
      ! electron conduction) and Iglesias & Rogers (1992; the rest)

      !      XF = 2.0*X + Y + 1.0
      XF = 2.0*xa(:,H1) + xa(:,He4) + 1.0

      ! Find the right jx
      toolow=.true.
      toohigh=.true.
      do while(ANY(toolow.or.toohigh))
         toolow=(csx(jx+1)>=xf).and.(jx<jcsx-1)
         toohigh=(csx(jx)<xf).and.(jx>1)
         where(toolow) jx=jx+1
         where(toohigh) jx=jx-1
      enddo

      XT = (XF-CSX(JX))/(CSX(JX+1)-CSX(JX))
      XU = 1.0 - XT
      IF (IOP.EQ.0) THEN
         ! linear interpolation in opacity table
         TT = 20.0*(TF - 2.95)
         JT = MAX(1, MIN(126, INT(TT)))
         TT = TT - JT
         TU = 1.0 - TT
         RT = 4.0*(FR + 12.25)
         JR = MAX(1, MIN(89, INT(RT)))
         RT = MAX(0.0, RT - JR)
         RU = 1.0 - RT
         DO J = 1, 2
            K = JX + J-1
            IF (J .EQ. 1) then
               do i=1,nz
                  FKL(i) = TT(i)*(RT(i)*CS(JR(i)+1,JT(i)+1,K(i))&
                       + RU(i)*CS(JR(i),JT(i)+1,K(i)))&
                       + TU(i)*(RT(i)*CS(JR(i)+1,JT(i),K(i)) &
                       + RU(i)*CS(JR(i),JT(i),K(i)))
               enddo
            endif
            IF (J .EQ. 2) then
               do i=1,nz
                  FKH(i) = TT(i)*(RT(i)*CS(JR(i)+1,JT(i)+1,K(i))&
                       + RU(i)*CS(JR(i),JT(i)+1,K(i)))&
                       + TU(i)*(RT(i)*CS(JR(i)+1,JT(i),K(i))&
                       + RU(i)*CS(JR(i),JT(i),K(i)))
               enddo
            endif
         END DO
    ELSE
       ! bicubic spline interpolation
       do i=1,nz
          CALL OPACTY(JX(i),TF(i),FR(i),FKL(i),FKH(i))
       enddo
    END IF
    FK = XT*10D0**FKH + XU*10D0**FKL
    if (ANY(fk.lt.0.)) then
       do i=1,nz
          if (fk(i).le.0.) then
             print*,'i',i,'fk',fk(i),'xt',xt(i),'xu',xu(i),'jx',jx(i)
          endif
       enddo
       print*,'xf(1)',xf(1)
       do i=1,jcsx
          print*,'csx(',i,')=',csx(i)
       enddo
    endif
    CH = 4D0*CL*PR/(FK*RHO*RHO*CP*T)
    ! Neutrino loss rates from Itoh et al (1983-1992)'
    EN = 0.0
    where ( TF .GE. 7.0 .AND. FR .GE. 0.0 ) 
       TT = 20.0*(TF - 6.95)
       IT = MAX(1, MIN(59, INT(TT)))
       TT = TT - IT
       TU = 1.0 - TT
       RT = 4.0*(FR + 0.25)
       IR = MAX(1, MIN(40, INT(RT)))
       RT = RT - IR
       RU = 1.0 - RT
    END where
    do i=1,nz
       if ( TF(i) .GE. 7.0 .AND. FR(i) .GE. 0.0 ) then
          ENP(i) = TT(i)*(RT(i)*CNIU(IT(i)+1,IR(i)+1,1) +&
               RU(i)*CNIU(IT(i)+1,IR(i),1)) +&
               TU(i)*(RT(i)*CNIU(IT(i),IR(i)+1,1) +&
               RU(i)*CNIU(IT(i),IR(i),1))
          ENB(i) = TT(i)*(RT(i)*CNIU(IT(i)+1,IR(i)+1,2) +&
               RU(i)*CNIU(IT(i)+1,IR(i),2)) +&
               TU(i)*(RT(i)*CNIU(IT(i),IR(i)+1,2) +&
               RU(i)*CNIU(IT(i),IR(i),2))
       endif
    enddo
    where ( TF .GE. 7.0 .AND. FR .GE. 0.0 )
       EN = -10.0D0**ENP - NZZ*10.0D0**ENB
    END where
    end subroutine tables

  SUBROUTINE FDIRAC (F, G)
    ! input variables
    double precision,dimension(nz)::f,g

    ! local variables
    double precision,dimension(nz)::vf,vg,uf,ug,fdf,wv
    double precision,dimension(nz,4)::ff,gg
    double precision,dimension(nz,4,4)::vw,vx
    double precision,dimension(nz,3,3)::D
    integer nd,i,j,l,k,m

    ! data variables
    integer,parameter::nc=4
    double precision,dimension(4,4,3)::&
         c=RESHAPE((/2.315472, 7.128660, 7.504998, 2.665350, 7.837752, &
              23.507934, 23.311317, 7.987465, 9.215560, 26.834068, &
              25.082745, 8.020509, 3.693280, 10.333176, 9.168960, 2.668248, &
              2.315472, 6.748104, 6.564912, 2.132280, 7.837752, 21.439740, &
              19.080088, 5.478100, 9.215560, 23.551504, 19.015888, &
              4.679944, 3.693280, 8.859868, 6.500712, 1.334124, 1.157736, &
              3.770676, 4.015224, 1.402284, 8.283420, 26.184486, 28.211372, &
              10.310306, 14.755480, 45.031658, 46.909420, 16.633242, &
              7.386560, 22.159680, 22.438048, 7.664928/),(/4,4,3/))
    if (verbose.ge.3) print*,'   dirac'

    ! Evaluate Fermi-Dirac integrals (Eggleton, Faulkner & Flannery 1973).'
    ! 1st row of matrix D contains rho*, P* and Q*; the 2nd row the logarithmic
    ! derivatives w.r.t. T; the 3rd row logarithmic derivatives w.r.t. f
    VF = 1.0/(1.0+F)
    VG = 1.0/(1.0+G)
    UF = F*VF
    UG = G*VG
    FDF = G*G + G
    FDF = UF*FDF*SQRT(FDF)
    FF(:,1) = 1.0
    GG(:,1) = 1.0
    DO I = 2, NC
       FF(:,I) = F*FF(:,I-1)
       GG(:,I) = G*GG(:,I-1)
       FDF = FDF*VF*VG
    END DO
    ND = 4
    DO I = 1, 3
       IF(I.GT.1) ND = 2
       DO L = 1, ND
          DO M = 1, ND + 1 - L
             VX(:,L,M) = 0.0
          END DO
       END DO
       DO J = 1, NC
          DO K = 1, NC
             VW(:,1,1) = C(K,J,I)*GG(:,J)*FF(:,K)
             DO L = 1, ND - 1
                VW(:,L+1,1) = (J-1)*VW(:,L,1)
                DO M = 1, ND - L
                   VW(:,L,M+1) = (K-1)*VW(:,L,M)
                END DO
             END DO
             DO L = 1, ND
                DO M = 1, ND + 1 - L
                   VX(:,L,M) = VX(:,L,M) + VW(:,L,M)
                END DO
             END DO
          END DO
       END DO
       WV = 1.0D0/VX(:,1,1)
       DO L = 1, ND
          DO M = 1, ND + 1 - L
             IF(L.NE.1 .OR. M.NE.1) VX(:,L,M) = VX(:,L,M)*WV
          END DO
       END DO
       D(:,I,1) = FDF*VX(:,1,1)
       D(:,I,2) = VX(:,2,1) + 1.5 + (2.5 - float(NC))*UG
       D(:,I,3) = VX(:,1,2) + 1.0 + (0.5*D(:,I,2) - float(NC))*UF
       IF(I.EQ.1) THEN
          ! second and third order derivatives of density, needed in PRESSI
          ! !!! Programming should be checked!!!
          DTT = VX(:,3,1) - VX(:,2,1)**2 + (2.5 - float(NC))*UG*VG
          DFT = VX(:,2,2) - VX(:,2,1)*VX(:,1,2) + 0.5*UF*DTT
          DFF = VX(:,1,3) - VX(:,1,2)**2 + UF*(DFT + (0.5*D(:,I,2) - float(NC))*VF&
               - 0.25*DTT*UF)
          DTTT = VX(:,4,1) + VX(:,2,1)*(2.0*VX(:,2,1)**2 - 3.0*VX(:,3,1)) +&
               (2.5 - float(NC))*(1.0-G)*UG*VG**2
          DFTT = VX(:,3,2) - 2.0*VX(:,2,1)*VX(:,2,2) - VX(:,1,2)*(VX(:,3,1) -&
               2.0*VX(:,2,1)**2) + 0.5*UF*DTTT
          DFFT = VX(:,2,3) - 2.0*VX(:,2,2)*VX(:,1,2) - VX(:,2,1)*(VX(:,1,3) -&
               2.0*VX(:,1,2)**2) + UF*(DFTT + 0.5*VF*DTT - 0.25*UF*DTTT)
          DFFF = VX(:,1,4) + VX(:,1,2)*(2.0*VX(:,1,2)**2 - 3.0*VX(:,1,3)) +&
               UF*(1.5*(DFFT + VF*DFT) - UF*(0.75*(DFTT + VF*DTT) - &
               0.125*UF*DTTT) + (0.5*D(:,1,2) - float(NC))*(1.0-F)*VF**2)
       END IF
    END DO

    ! Attribute D to the correct statef variables
    re =D(:,1,1)
    pe =D(:,2,1)
    qe =D(:,3,1)
    ret=D(:,1,2)
    pet=D(:,2,2)
    qet=D(:,3,2)
    ref=D(:,1,3)
    pef=D(:,2,3)
    qef=D(:,3,3)
  END SUBROUTINE FDIRAC

  ! subroutine pressi.f from STAR program
  SUBROUTINE PRESSI(icl, TI, XI, XF, XT, F, DC, DVT, DVF, DP, DPT,&
       DPF, DS, DST, DSF, DU)
    use maths,only:cbrt,fxp
    use abundances,only:neo,nio,nzz 
    use constants,only:cevb,cpl
    implicit none

    ! input variables
    integer::icl,i  ! Coulomb interaction switch
    double precision,dimension(nz)::ti,xi,xf,xt,f

    ! output variables
    double precision,dimension(nz)::DC, DVT, DVF, DP, DPT,DPF, DS, DST, DSF, DU

    ! local variables
    ! pressure ionisation
    double precision,dimension(nz)::yi,zi,dd,dd1,dd2,ff,fx,cc,cx,cxx
    double precision,dimension(nz)::ee,eex,bb,bx,by,bxx,bxy,byy,wx
    double precision,dimension(nz)::aa,ax,axx,gi,dgdx,dgdy,dgdxx,dgdxy,dgdyy
    ! coulomb interaction
    double precision,dimension(nz)::ff1,wf,ff2,th,tht,thf,thx,thy,thtt,thft,thff,thxx,thxy,thyy
    double precision,dimension(nz)::thc,be,eeg,beg,begg,ge,geg
    double precision,dimension(nz)::dgdg,dgdgg,gamx,gamy,gamxx,gamxy,gamyy!,gam
    ! Chemical potential correction
    double precision,dimension(nz)::GAMn01,GAMn12,FC,GAMXn12,GAMYn12,DGDGn01,DGDGn12
    ! data
    double precision,parameter::CA1=0.89752, CA2=0.768, CA3=0.208
    double precision,parameter::CF1=1.07654, CF2=0.56956
    double precision,parameter::CP1=3.,CP2=0.25,CP3=2.,CP4=3.0D-2
    if (verbose.ge.3) print*,'   pressi'
    !Computes effects on the EOS of a free energy contribution 
    !delta(F) = -k*T*N_e*G(XI,TI), where XI = n_e*m_u and TI = 1eV/(k*T).
    ! Pressure ionisation. 
    ! FF and CC are approximation of the degeneracy parameters F and PSI
    DmuI=0.
    if (icl.gt.2 ) then
       DC=0.;DVT=0.;DVF=0.; DP=0.; DPT=0.; DPF=0.; DS=0.; DST=0.; DSF=0.;DU=0. 
       return
    endif
    YI = 13.6*TI
    ZI = CF1*XI*YI*SQRT(YI)
    DD = CF2*ZI
    DD1 = 1.0 + DD
    DD2 = DD1 + DD/3.0
    FF = ZI*CBRT(DD1)
    FX = DD2/DD1
    CC = 2*SQRT(FF)
    CX = CC*0.5*FX
    CXX = CX*(0.5*FX + DD/(3.0*DD2*DD1))
    EE = 1.0/(1.0 + XI/CP4)
    EEX = CP3*(XI/CP4)*EE
    BB = YI + CC - CP3*LOG(EE)
    BX = CX + EEX
    BY = YI + 1.5*CX
    BXX = CXX + EE*EEX
    BXY = 1.5*CXX
    BYY = YI + 1.5*BXY
    WX = (CP1/XI)**CP2
    AA = FXP(-WX)
    AX = CP2*WX*AA
    AXX = CP2*AX*(WX - 1.0)
    if (icl.le.1) then
       GI = AA*BB
       DGDX = AA*BX + AX*BB
       DGDY = AA*BY
       DGDXX = AA*BXX + 2.0*AX*BX + AXX*BB
       DGDXY = AA*BXY + AX*BY
       DGDYY = AA*BYY
       dmui=1.*gi ! IN FACT, this is not the only contribution to d gi / d Ni,
       ! since x=RHO*NEO depends on Ni as well...
    else
       GI = 0.
       DGDX = 0.
       DGDY = 0.
       DGDXX = 0.
       DGDXY = 0.
       DGDYY = 0.
    endif
    IF (ICL .ge. 1) THEN
       ! Coulomb interaction.
       ! the input variable F is used, rather than the above approximation, 
       ! resulting in quite a bit of programming to get the derivatives of 
       ! TH = d log(n_e)/d(psi). DFF, DFFF etc are 2nd and 3rd derivatives of n_e.
       FF1 = F + 1.0
       WF = SQRT(FF1)
       FF2 = -0.5*F/FF1
       TH = XF/WF
       THT = DFT/WF
       THF = DFF/WF + FF2*TH
       THX = THF/XF
       THY = THX*XT - THT
       THTT = DFTT/WF
       THFT = DFFT/WF + FF2*THT
       THFF = DFFF/WF + FF2*(2.0*THF - (2.0 + F)*FF2*TH/F)
       THXX = (THFF - THX*DFF)/XF**2
       THXY = THXX*XT - (THFT - THX*DFT)/XF
       THYY = THXX*XT**2 + 2*XT/XF*(THX*DFT - THFT) + THTT - THX*DTT
       ! GAM is the plasma interaction parameter. Note that THC = ZT**2*NIO/NEO
       THC = TH + ABS(NZZ/NEO)
       GAM = CBRT(XI*(CPL*NEO/NIO)**2/3.0)*TI/CEVB*THC
       BB = (CA1*SQRT(3.0/GAM))**(1.0/CA2)
       EE = (GAM/(GAM + CA3))**(1.0/CA2)
       BE = EE + BB
       EEG = CA3/(GAM + CA3)
       BEG = (EEG*EE - 0.5*BB)/BE
       BEGG = (EEG*EEG*(1.0 - GAM*CA2/CA3)*EE + 0.25*BB)/BE
       GE = (NIO/NEO)*CA1*GAM/(BE**CA2)
       GEG = 1.0 - BEG!PIERRE?!*CA2 !PIERRE!
       DGDG = GE*GEG
       DGDGG = GE*(GEG*GEG + (BEG*BEG - BEGG)/CA2)
       GAMX = 1.0/3.0 + THX/THC
       GAMY = 1.0 + THY/THC
       GAMXn12=-THX/(THC*THC)
       GAMYn12=-THY/(THC*THC)
       GAMXX = THXX/THC - (THX/THC)**2
       GAMXY = THXY/THC - THX*THY/THC**2
       GAMYY = THYY/THC - (THY/THC)**2
       GI = GI + GE
       DGDX = DGDX + DGDG*GAMX
       DGDY = DGDY + DGDG*GAMY
       DGDXX = DGDXX + DGDGG*GAMX**2 + DGDG*GAMXX
       DGDXY = DGDXY + DGDGG*GAMX*GAMY + DGDG*GAMXY
       DGDYY = DGDYY + DGDGG*GAMY**2 + DGDG*GAMYY
       ! Derivatives of chemical potential muC with respect to :
       ! n01=NIO/NEO
       ! n12=NEO/NZZ
       GAMn01=2./(3.*n01)
       GAMn12=1./THC
       DGDGn01=DGDGG*GAMn01-DGDG/n01
       DGDGn12=DGDGG*GAMn12
       FC=-GE
       DFCn01=FC*(-1./n01+GEG*GAMn01)
       DFCn12=FC*GEG*GAMn12
    END IF
    ! evaluate changes to ionisation potential (DC), pressure (DP), and
    ! entropy (DS), and their derivatives w.r.t. log(f) and log(T)
    DC = DGDX + GI
    DVT = DGDXX + DGDX
    DVF = DVT*XF
    DVT = DVT*XT
    DP = -XI*DGDX
    if (icl.ge.1.and.icl.le.2) then
       ! In case of Coulomb corrections
       DPn01=-XI*DGDGn01*GAMX
       DPn12=-XI*(DGDGn12*GAMX+DGDG*GAMXn12)
    endif
    DPF = -XI*DVF
    DPT = XI*(DGDXY - DVT) + DP
    DVT = DVT - DGDXY - DGDY
    DS = GI - DGDY
    if (icl.ge.1.and.icl.le.2) then
       ! In case of Coulomb corrections
       DSn01=-DFCn01-DGDGn01*GAMY
       DSn12=-DFCn12-DGDGn12*GAMY-DGDG*GAMYn12
    endif
    DST = DGDX - DGDXY
    DSF = DST*XF
    DST = DST*XT - DGDY + DGDYY
    DU = -DGDY
  END SUBROUTINE PRESSI

  SUBROUTINE OPACTY (JX,TF,FR,FKL,FKH)
    use maths,only:spline
    implicit none
    integer::jx
    double precision::tf,fr,fkl,fkh
    integer,parameter::mt=127,mr=90
    double precision,save,dimension(4,4,mt,mr,10)::F
    double precision,dimension(mt),save::tfm
    double precision,dimension(mr),save::frm
    double precision,dimension(6)::fklm,fkhm
    double precision,save::fklo,fkho
    ! locals
    integer::i,j,jr,ic,kx
    integer::jt,ifail,iq,jq,ir,it
    double precision::dt,dr
    double precision,save,dimension(4,mt)::mat

    logical,save::first=.true.

    if (verbose.ge.10) print*,'    opacty'

    if (first) then
       first=.false.
       
!
! Calculate a bicubic spline interpolation fit for the temperature
! and density opacity fit.  This is based on routines from Princeton
! Splines, D. McCune found by L. Dray.
!
       DO 10 JT = 1,MT
          TFM(JT) = 2.95D0 + 0.05D0*DFLOAT(JT)
10        CONTINUE
          DO 20 JR = 1,MR
             FRM(JR) = 0.25D0*DFLOAT(JR) - 12.25D0
20           CONTINUE
             IFAIL = 0
             DO 30 kX = 1,10
                DO 26 JQ = 1,MT
                   DO 23 IQ = 1,MR
                      F(1,1,JQ,IQ,KX) = CS(IQ,JQ,KX)
23                    CONTINUE        
26                    CONTINUE
!
! Construct splines in the T direction.
!
      DO IR=1,MR
         DO IT = 1,MT
            MAT(1,IT) = F(1,1,IT,IR,KX)
         ENDDO
         CALL SPLINE(MT,TFM,MAT)
         DO IT = 1,MT-1
            F(2,1,IT,IR,KX) = MAT(2,IT)
            F(3,1,IT,IR,KX) = MAT(3,IT)
            F(4,1,IT,IR,KX) = MAT(4,IT)
         ENDDO
      ENDDO
!
! Construct splines in the rho direction.
!
      DO IT = 1,MT-1
!
! Construct splines for each T coeff
!
         DO IC = 1,4
            DO IR = 1,MR
               MAT(1,IR) = F(IC,1,IT,IR,KX)
            ENDDO
            MAT(2,1) = 0D0
            MAT(3,1) = 0D0
            MAT(2,MR)= 0D0
            MAT(3,MR)= 0D0
            CALL SPLINE(MR,FRM,MAT)
            DO IR = 1,MR-1
               F(IC,2,IT,IR,KX) = MAT(2,IR)
               F(IC,3,IT,IR,KX) = MAT(3,IR)
               F(IC,4,IT,IR,KX) = MAT(4,IR)
            ENDDO
         ENDDO
      ENDDO
30    CONTINUE
    endif

    ! Calculate a bicubic spline interpolation fit for the temperature
    ! and density opacity fit.  Do not stop if the input lies outside the
    ! array but rather use the previous result.

    ! Check that we are in the table.
    IF((TF.LT.TFM(1)).OR.(TF.GE.TFM(MT)).OR.&
           (FR.LT.FRM(1)).OR.(FR.GE.FRM(MR))) THEN
       FKL = FKLO
       FKH = FKHO
!!!       WRITE (6,100) TF,FR,FKL,FKH
    ELSE
       FKLO = FKL
       FKHO = FKH
       ! Find interval in which target point lies.
       I = 1 + (MT-1)*(TF-TFM(1))/(TFM(MT)-TFM(1))
       J = 1 + (MR-1)*(FR-FRM(1))/(FRM(MR)-FRM(1))
       DT = TF-TFM(I)
       DR = FR-FRM(J)

       ! Evaluate the splines.

       FKL = F(1,1,I,J,JX) + DR*(F(1,2,I,J,JX)&
            + DR*(F(1,3,I,J,JX) + DR*F(1,4,I,J,JX)))&
            + DT*(F(2,1,I,J,JX) + DR*(F(2,2,I,J,JX)&
            + DR*(F(2,3,I,J,JX) + DR*F(2,4,I,J,JX)))&
            + DT*(F(3,1,I,J,JX) + DR*(F(3,2,I,J,JX)&
            + DR*(F(3,3,I,J,JX) + DR*F(3,4,I,J,JX)))&
            + DT*(F(4,1,I,J,JX) + DR*(F(4,2,I,J,JX)&
            + DR*(F(4,3,I,J,JX) + DR*F(4,4,I,J,JX))))))

       FKH = F(1,1,I,J,JX+1) + DR*(F(1,2,I,J,JX+1)&
            + DR*(F(1,3,I,J,JX+1) + DR*F(1,4,I,J,JX+1)))&
            + DT*(F(2,1,I,J,JX+1) + DR*(F(2,2,I,J,JX+1)&
            + DR*(F(2,3,I,J,JX+1) + DR*F(2,4,I,J,JX+1)))&
            + DT*(F(3,1,I,J,JX+1) + DR*(F(3,2,I,J,JX+1)&
            + DR*(F(3,3,I,J,JX+1) + DR*F(3,4,I,J,JX+1)))&
            + DT*(F(4,1,I,J,JX+1) + DR*(F(4,2,I,J,JX+1)&
            + DR*(F(4,3,I,J,JX+1) + DR*F(4,4,I,J,JX+1))))))
    ENDIF
100 FORMAT ('OPACITY OUT OF RANGE',' TF FR FKL FKH ',4F9.4)
    RETURN
  END SUBROUTINE OPACTY
 
END MODULE EOS





