!
! This module recovers primary variables from x, dx and xold vectors
! and compute secondary variables (these are local variable, ie that
! depend only on other variables on the same interface or zone).
!
MODULE variables 
  use dimensions,only:nz,species,ndim,idx,verbose
  use secondaries  ! outputs of funcs1 used by equns1
  implicit none

  ! inverse of time-step
  double precision::invdt
  ! 2nd order time derivative
  double precision::a2nd
  logical,save::t2nd=.false.

  ! PRIMARY AND SECONDARY VARIABLES
  ! (old variables are with indices old, new ones with 'new' and 
  ! the time centered ones do not have an index)

  ! PRIMARY VARIABLES 
  double precision,dimension(nz)::lnfold,lnf,lnfnew,dtlnf
  double precision,dimension(nz)::lnTold,lnT,lnTnew,dtlnt
  double precision,dimension(nz)::lnmold,lnm,lnmnew,dtlnm
  double precision,dimension(nz)::lnrold,lnr,lnrnew,dtlnr
  double precision,dimension(nz)::Lold,L,Lnew,dtL
  double precision,dimension(nz)::dmold,dm,dmnew,dtdm
  double precision,dimension(nz)::nsmoothold,nsmooth,nsmoothnew,dtnsmooth
  double precision,dimension(nz,species)::fracold,frac,fracnew,dtfrac
  double precision,dimension(nz,species)::gfracold,gfrac,gfracnew,dtgfrac

  ! SECONDARY VARIABLES
  ! Save for previous old secondaries
  double precision,dimension(nz)::TdtSsave,TdtSold
  double precision,dimension(nz+1)::Ltotold,Ltotsave
  double precision,dimension(nz)::chold,chsave
  double precision,dimension(nz)::rhoisave,usave,ssave
  double precision,dimension(nz)::rhoiold,uold,sold,lnpold,lnpsave
  double precision,dimension(nz)::rhoold,rhosave
  double precision,dimension(nz)::netheatold,netheatsave
  double precision,dimension(nz)::nhsave,nhold
  double precision,dimension(nz)::neold,nesave,mueold,muesave
  double precision,dimension(nz,3)::xHold,xHsave,xHeold,xHesave
  double precision,dimension(nz,3)::muHold,muHsave,muHeold,muHesave
  double precision,dimension(nz,ndim)::xnold,xnsave,muiold,muisave,naold,nasave
  double precision::tgsave,tgold,TKHsave,TKHold,tdold,tdsave
  double precision,dimension(nz)::tcold,tcsave
  double precision,dimension(nz)::resold,ressave
  double precision,dimension(nz,species)::diffold,diffsave
  double precision,dimension(nz)::mtold,mtsave
  double precision,dimension(nz)::convold,convsave
  double precision,dimension(nz)::dgold,dgsave
  double precision,dimension(nz)::dglgold,dglgsave
  double precision,dimension(nz)::dglgcold,dglgcsave,lgold,lgsave
  ! geometry
  double precision,dimension(nz)::rold,rnew
  double precision,dimension(nz)::mold,mnew

  ! EQUATIONS IN
  ! Secondaries
!  double precision,dimension(nz)::netheat
  double precision,dimension(nz)::r2
  ! To printings
  double precision,dimension(nz)::Tdts,fudgecc
  double precision,dimension(nz,species)::ratet,dtx,dtxold,dtxsave
  double precision::mdot,lacc,pacc

  
CONTAINS

  ! ================= Primary variables ===================

  SUBROUTINE primaries
    ! xold,dx and x ---> 'nicely named' primary variables
    ! using pointers would be more beautiful, but would make
    ! debugging harder
    call setvarold
    call setvarnew
    call setdvar
    call setvart
  end SUBROUTINE primaries

  SUBROUTINE setvarold
    use dimensions
    use maths,only:expc,plexp,plsinh
    use chemistry,only:threshold,threflux
    implicit none
    ! xold --> variables_old (new variables)
    integer i,j
    if (verbose.ge.1) print*,' setvarold'

    do i=1,nz
       lnfold(i)=xold(nvar*(i-1)+ilnf)
       lnTold(i)=xold(nvar*(i-1)+ilnT)
       lnmold(i)=xold(nvar*(i-1)+ilnm)       
       lnrold(i)=xold(nvar*(i-1)+ilnr)
       Lold(i)  =xold(nvar*(i-1)+iL)!plSINH(xold(nvar*(i-1)+iL))
       dmold(i) =plEXP(xold(nvar*(i-1)+idm))
       nsmoothold(i)=plEXP(xold(nvar*(i-1)+ins))
       do j=1,species
          if (threshold.ne.0.) then
             fracold(i,j)=xold(nvar*(i-1)+species+nvarh+j)!EXPC(xold(nvar*(i-1)+species+nvarh+j))*threshold
          else
             fracold(i,j)=xold(nvar*(i-1)+species+nvarh+j)
          endif
          if (idx==2) gfracold(i,j)=xold(nvar*(i-1)+nvarh+j)
!!$          if (idx==2) gfracold(i,j)=threflux*1.d-18*plSINH(xold(nvar*(i-1)+nvarh+j))
       enddo
    enddo
    if (idx==1) then
       do j=1,species
          gfracold(1:nz-1,j)=fracold(2:nz,j)-fracold(1:nz-1,j)
       enddo
       gfracold(nz,:)=0.
    endif
    rold=exp(lnrold)
    mold=exp(lnmold)
  end SUBROUTINE setvarold

  SUBROUTINE setvarnew
    use dimensions
    use maths,only:expc,plexp,plsinh
    use chemistry,only:threshold,threflux,specie,ispecie,C12
    implicit none
    ! x --> variables_new (new variables)
    integer i,j
    if (verbose.ge.1) print*,' setvarnew'

    do i=1,nz
       lnfnew(i)=x(nvar*(i-1)+ilnf)
       lnTnew(i)=x(nvar*(i-1)+ilnT)
       lnmnew(i)=x(nvar*(i-1)+ilnm)
       lnrnew(i)=x(nvar*(i-1)+ilnr)
       Lnew(i)  =x(nvar*(i-1)+iL)!plSINH(x(nvar*(i-1)+iL))
       dmnew(i) =plEXP(x(nvar*(i-1)+idm))
       nsmoothnew(i) =plEXP(x(nvar*(i-1)+ins))
       do j=1,species
          if (threshold.ne.0.) then
             fracnew(i,j)=x(nvar*(i-1)+species+nvarh+j)!threshold*EXPC(x(nvar*(i-1)+species+nvarh+j))
          else
             fracnew(i,j)=x(nvar*(i-1)+nvarh+species+j)
          endif
!!$          if (idx==2) gfracnew(i,j)=threflux*1.d-18*plSINH(x(nvar*(i-1)+nvarh+j))
          if (idx==2) gfracnew(i,j)=x(nvar*(i-1)+nvarh+j)
       enddo
    enddo
  end SUBROUTINE setvarnew

  SUBROUTINE setdvar
    use dimensions
    use maths,only:expc,plexp,plsinh
    use chemistry,only:threshold,threflux,specie,ispecie,C12
    implicit none
    ! dx --> dtvariables (variation of variables, new-old)
    integer i,j
    if (verbose.ge.1) print*,' setdvar'

    do i=1,nz
       dtlnf(i)=dx(nvar*(i-1)+ilnf)
       dtlnT(i)=dx(nvar*(i-1)+ilnT)
       dtlnm(i)=dx(nvar*(i-1)+ilnm)
       dtlnr(i)=dx(nvar*(i-1)+ilnr)
       dtL(i)  =dx(nvar*(i-1)+iL)
       dtdm(i) =plEXP(x(nvar*(i-1)+idm))-plEXP(xold(nvar*(i-1)+idm))
       dtnsmooth(i) =plEXP(x(nvar*(i-1)+ins))-plEXP(xold(nvar*(i-1)+ins))
       do j=1,species
          if (threshold.ne.0.) then
             dtfrac(i,j)=dx(nvar*(i-1)+species+nvarh+j)
          else
             dtfrac(i,j)=dx(nvar*(i-1)+nvarh+species+j)
          endif
          if (idx==2) dtgfrac(i,j)=dx(nvar*(i-1)+nvarh+j)
       enddo
    enddo
  end SUBROUTINE setdvar

  SUBROUTINE setvart
    use parameters,only:alpha
    integer i
    if (verbose.ge.1) print*,' setvart'

    ! Update the centered primary variables
    ! In this code, alpha=1., so this routine is kind of useless
    ! but one day, I'll make this 2nd order accurate in time..
    lnf  = (1.-alpha)*lnfold  + alpha*lnfnew
    lnT  = (1.-alpha)*lnTold  + alpha*lnTnew
    lnm  = (1.-alpha)*lnmold  + alpha*lnmnew
    lnr  = (1.-alpha)*lnrold  + alpha*lnrnew
    L    = (1.-alpha)*Lold    + alpha*Lnew  
    dm   = (1.-alpha)*dmold   + alpha*dmnew  
    nsmooth   = (1.-alpha)*nsmoothold   + alpha*nsmoothnew  
    frac = (1.-alpha)*fracold + alpha*fracnew
    if (idx==2) gfrac = (1.-alpha)*gfracold + alpha*gfracnew
  end SUBROUTINE setvart

  ! ===============  Secondary variables ========================


  SUBROUTINE setsecondaries 
    ! Knowing the primaries, setup all secondary variables  
    ! Set the pointers
     CALL pointers
    ! compute the local secondaries
    CALL funcs1
  end SUBROUTINE setsecondaries

  ! Routine funcs1.f from STAR program
  ! convective stuff has been moved up in equations.f90, because
  ! it involves coupling between interfaces and zones variables
  SUBROUTINE FUNCS1
    ! this certainly needs some cleaning...
    use chemistry,only:stoech,kprod,kreac,nreac,ipreac,ireac,iprim,rstatus,am,bn,&
         H1,He4,iRPP,ir33,ir34,irbe,irbp,specie,ispecie,rtype,istatus&
         ,izat=>iz,zreac,a23d,a23m,C12,ircc,irccg
    use dimensions,only:nz,time,dt,dtold
    use constants,only:CPI, PI4, CLN10, CA, CB, CL, CD, CG, CEVB, &
         CEN, CPL, CMEVMU, CSECYR, LSUN, MSUN, RSUN, TSUNYR,CR,BNEL
    use parameters
    use EOS,only:statef,pr,pg,fk,en
    use maths,only:cbrt,ps,rlobe,lcbrt,lps
    use nuclear,only:nucrat,rate,ex,enx,qcc
    implicit none

    ! local variables 
    integer i,j,k
    ! surface boundary conditions
    double precision::prb
    ! Dmu for each reaction
    double precision,dimension(nz)::smu
    !fudge for cC reactions
    double precision,dimension(nz)::rk,fudge
    double precision::fudcc
    if (verbose.ge.1) print*,' funcs'

    ! inverse of the time step
    if (dt.eq.0.) then
       invDT=0.
    else
       invDT=1./dt
    endif

    ! mass
    m = EXP(lnmnew)
    ! mass flow
!    mt = dtlnm*m*invdt
    mt=(m-mold)*INVDT

    ! Low temperature warning'
    IF (ANY(lnT .LE. 0.000001)) then
       print*,'WARNING ! Low temperature !'
       do i=1,nz
          if (lnT(i).le.0.000001) then
             write (6,*) 'zone',i,' : AT = ',lnT(i),' AF = ',lnf(i)
          endif
       enddo
    endif

    ! set up the composition variables
    CALL initchem

    ! Equation of state
    CALL STATEF(lnf,lnT)
    CALL NUCRAT(lnT)
    
    ! radius and geometry
    r = EXP(lnr)
    r2 = r*r
    surf=pi4*r2

    ! pressure gradient equation
    grav = 1.0D11*CG*m/r2
    pm = -1.d11*grav/surf

    ! energy equation
    netheat=ex+en+enx+ec
    TdtS=invdt*T*cp*(dtlnt-grada*(lnp-lnpold))
    if (t2nd.and.T2ndORDER=='yes') then
       ! I am not sure this is still up to date,
       ! but it is where you want to implement
       ! Leon Lucy's 2nd order scheme in time
       a2nd=dt/(dt+dtold)
       Tst1=((1.+a2nd)*TdtS-a2nd*TdtSold)
       Tst2=-netheat       
    else
       Tst1=tdTs
       Tst2=-netheat
    endif

    ! composition equations
    xt2=0.
    do i=1,nreac
       if (rtype(i).ne.'q') then
          if (i==irpp) then
             ! Equivalent PP reaction is 2H1-->0.5He4
             if (iprim(H1).ne.0) xt2(:,iprim(H1))=xt2(:,iprim(H1))+2.*rate(:,i)
             if (iprim(He4).ne.0)xt2(:,iprim(He4))=xt2(:,iprim(He4))-.5*rate(:,i)
          else
             do j=1,5
                if (ipreac(i,j).ne.0)&
                     xt2(:,ipreac(i,j))=xt2(:,ipreac(i,j))+rate(:,i)*stoech(i,j)
             enddo
          endif
       endif
    enddo


    do i=1,species
       ratet(:,i)=xt2(:,i)*bn(ispecie(i))
       dtx(:,i)=+ich*invdt*dtfrac(:,i)
       if (t2ndorder=='yes'.and.t2nd) then
          xt2(:,i)=xt2(:,i)*bn(ispecie(i))
          XT1(:,I)=(1.+a2nd)*dtx(:,i)-a2nd*dtxold(:,i)
       else
          xt2(:,i)=xt2(:,i)*bn(ispecie(i))
          XT1(:,I)=dtx(:,i)! Eggleton like (conserves mass)
       endif
       !   xt(:,i)=xt(:,i)*am(ispecie(i)) ! does not conserve mass.
    enddo

    ! surface boundary conditions(iz=1)
    ! mass
    mdot=rmg*Msun
    if (rmg.eq.0..and.taccretion.ne.0.) then
       ! Han&Podsiadlowski(2005), equ. (1), plus decay at m=macc0*msun, tacc0 yrs
       mdot=min(mcr(m(1)/Msun),mcr(macc0)-(m(1)/msun-macc0)/tacc0)
!!$       ! equ (3)
!!$       if (mdot.lt.1./8.*mcr(m(1)/Msun)) mdot=0.
       ! equ. (4) 
       if (mdot.gt.0.) then
          if (alog10(mdot).lt.-7.3) then
             mdot=mdot*(-0.175*(-7.3+5.35)**2+1.05)
          else if (alog10(mdot).lt.-5.9) then
             mdot=mdot*(-0.175*(alog10(mdot)+5.35)**2+1.05)
          endif
       endif
       mdot=mdot*Msun/csecyr
    endif
    if (time+taccretion-tcool.ne.0.) then
       mdot=mdot*MAX(0.,time-tcool)/(taccretion+time-tcool)
    endif
    bc1=m(1)-mold(1)-dt*mdot
    bc1=bc1/Mold(1)
    if (dt==0.) bc1=dtlnm(1)
    if (MESH.eq.'lag') then
       mt=0.
       mdot=0.
    endif

    ! Pressure surface boundary condition
    PRB = CA*TRB**4/3D0
    if (TRB.ne.0.) Then
       print*,'trb=',trb
       print*,' WARNING ! I set it up to zero.'
       trb=0.
       prb=0.
       read(*,*)
    endif
    LACC = 0*1D22*CG*max(0.,MDOT)*M(1)/R(1)
    PACC = 0*1D22*max(0.,MDOT)/(PI4*R2(1))*SQRT(2D0*CG*M(1)/R(1))
    BC2 = FK(1)*(1.5D0*(PG(1)-PACC) + 0.75D0*PR(1))

    ! Temperature surface boundary condition
    bc3=1.0D11/(0.75D0*PI4*CL*(PR(1)-PRB))*(L(1)+LACC)/L(1)

    ! STAR mesh spacing function
    call setmeshspacing
  contains
      double precision function mcr(m)
        double precision::m
        mcr=5.3d-7*(1.7-0.7)/0.7*max(0.,m-0.4)
        return
      end function mcr
  END SUBROUTINE FUNCS1

  SUBROUTINE setmeshspacing
    ! mesh spacing equation, with modified pressure, temperature gradient equations'
    use secondaries
    use parameters
    use maths,only:lcbrt,cbrt
    use chemistry,only:a23m,a23d,iprim
    implicit none
    double precision::CT1,CT10
    double precision,dimension(nz)::vm,vm3,vr,vx
    integer::i
    ! Q is the quantity that the meshpoints should be at equal intervals of'
    CT1 = 1.0D1**(1.0D1*CT(1)) !!! should change input format !!!
    q= CT(4)*lnP + CT(5)*LOG((P+CT(9))/(P+CT1)) + &
         CT(2)*LOG((P+CT(10))/(P+CT1))
    VPP = CT(4) + P/(P+CT1)*(CT(5)*(CT1-CT(9))/(P+CT(9))&
         + CT(2)*(CT1-CT(10))/(P+CT(10)))
    CT10 = 2.0D4              !!! fixed, should change input format !!!
    q=q+ CT(7)*DLOG(T/(T+CT10))
    VTT = CT(7)*CT10/(T+CT10)
    ! VR and M must go to zero at centre like r**2, m**(2/3)'
    VM = CT(6)*lCBRT(mold(1)*mold(1))
    VM3=cbrt(M)
    VMM = VM + VM3*VM3
    vm=DLOG(VM/VMM)
    q=q+vm
    VMM = -0.66666667D0/(VMM*VM3)
    vr=-CT(3)*DLOG(R2/CT(8)+1.0D0)
    VRR = -CT(3)/(R2+CT(8))
    q=q+vr
    bc4=vm(nz)+vr(nz)

    vxx=0.
!!$    q=q-1.d8*frac(:,iprim(a23m))
!!$    vxx(:,iprim(a23m))=-1.d8
!!$    q=q+1.d8*frac(:,iprim(a23d))
!!$    vxx(:,iprim(a23d))=+1.d8
!!$    q=q-10.*log(frac(:,iprim(a23m))+1.d-8)
!!$    vxx(:,iprim(a23m))=-1./(frac(:,iprim(a23m))+1.d-8)
!!$    q=q+10.*log(frac(:,iprim(a23d))+1.d-8)
!!$    vxx(:,iprim(a23d))=+1./(frac(:,iprim(a23d))+1.d-8)
!!$    do i=1,species-1
!!$       vx=.1*frac(:,i)
!!$       vxx=.1
!!$       q=q+vx
!!$    enddo
!    QM = VPP*pm/p + VTT*ATM + VMM + VRR*R2M
    
  end SUBROUTINE setmeshspacing


  SUBROUTINE initchem
    ! Compute secondary chemical variables
    use chemistry,only:iz,bn,am,rstatus,istatus,iprim,frac_init,DEDUCE&
         ,specie,nat,ispecie
    use abundances
    implicit none

    ! local variables
    integer i,j
    ! Deduced abundance
    integer ideduce
    double precision,dimension(nz)::xdeduce
    double precision,parameter::eps=1.d-30
    if (verbose.ge.2) print*,'  initchem'
    ! Some quantities that do not depend on the state of ionization:'
    ! the NA are the element number densities (per baryon); AVM is the average'
    ! mass per baryon (in amu); NEO and NIO are the numbers of electrons and'
    ! ions assuming complete ionization.'
    ! XA is the mass fraction.
    ! Ftot contains chemical gradients
    xdeduce=1.
    ideduce=0
    DO I = 1, ndim
       if (rstatus(i).eq.'e') then
          xa(:,i)=0.
          na(:,i)=1.
          Ftot(:,i)=0.
       endif
       if (rstatus(i).eq.'E'.or.rstatus(i).eq.'n'.or.&
            rstatus(i).eq.'C'.or.rstatus(i).eq.'D') then
          xa(:,i)=frac(:,iprim(i))
          na(:,i)=frac(:,iprim(i))/bn(i)! Eggleton like! /am(i)
          Ftot(:,i)=gfrac(:,iprim(i))
       endif
       if (rstatus(i).eq.'c') then
          xa(:,i)=frac_init(i)
          na(:,i)=frac_init(i)/bn(i)! Eggleton like! /am(i)
          Ftot(:,i)=0.
       endif
       if (rstatus(i).eq.'u') then
          xa(:,i)=MAX(0.,frac_init(i)+frac_init(i-1)-xa(:,i-1))
          na(:,i)=xa(:,i)/bn(i)
          Ftot(:,i)=-Ftot(:,i-1)
       endif
       if (rstatus(i).eq.'d') then
          ideduce=i
          xa(:,i)=0.
          na(:,i)=0. 
          Ftot(:,i)=0.
       endif
       xdeduce=xdeduce-xa(:,i)
    enddo    
    if (DEDUCE.eq.'yes') then
       xa(:,ideduce)=MAX(0.,xdeduce)
       na(:,ideduce)=xa(:,ideduce)/bn(ideduce)! Eggleton like! /am(ideduce)
       Ftot(:,ideduce)=0.
       do i=1,ndim
          if (i.ne.ideduce)&
               Ftot(:,ideduce)=Ftot(:,ideduce)-Ftot(:,i)
       enddo
    endif

    nbar=0.
    NEO = 0.0
    NIO = 0.0
    NZZ = 0.0
    AVM = 0.0
    ye=0.
    FNe=0.
    do i=1,ndim
       if (istatus(i).ne.' ') then
          nbar=nbar+bn(i)*na(:,i)
          AVM = AVM + float(bn(I))*NA(:,I) ! Eggleton like
          ye=ye-.5*float(bn(i)-2*iz(i))*na(:,i)
!          AVM = AVM + AM(I)*NA(:,I)
          NIO = NIO + NA(:,I)
          FNe = FNe + .5*float(2*IZ(i)-bn(i))*Ftot(:,i)/float(bn(i))
          NEO = NEO + float(IZ(I))*NA(:,I)
          NZZ = NZZ + IZ(I)*IZ(I)*NA(:,I)
       endif
    END DO    
    avm=1.
  end SUBROUTINE initchem


end MODULE variables



